/* $Novell: debug.h,v 1.10 2002/09/18 21:39:55 $ 
 ****************************************************************************
 * Copyright (c) 2006 Novell, Inc.
 * All Rights Reserved.
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; version 2.1 of the license.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, contact Novell, Inc.
 * 
 * To contact Novell about this file by physical or electronic mail,
 * you may find current contact information at www.novell.com
 * 
 *****************************************************************************
 * Copyright (C) 1999, 2000 Novell, Inc. All Rights Reserved.
 * 
 * THIS WORK IS SUBJECT TO U.S. AND INTERNATIONAL COPYRIGHT LAWS AND
 * TREATIES. USE, MODIFICATION, AND REDISTRIBUTION OF THIS WORK IS SUBJECT
 * TO VERSION 2.0.1 OF THE OPENLDAP PUBLIC LICENSE, A COPY OF WHICH IS
 * AVAILABLE AT HTTP://WWW.OPENLDAP.ORG/LICENSE.HTML OR IN THE FILE "LICENSE"
 * IN THE TOP-LEVEL DIRECTORY OF THE DISTRIBUTION. ANY USE OR EXPLOITATION
 * OF THIS WORK OTHER THAN AS AUTHORIZED IN VERSION 2.0.1 OF THE OPENLDAP
 * PUBLIC LICENSE, OR OTHER PRIOR WRITTEN CONSENT FROM NOVELL, COULD SUBJECT
 * THE PERPETRATOR TO CRIMINAL AND CIVIL LIABILITY. 
 *****************************************************************************/


/****************************************************************************
*
* Description:   
*
*  DBG tracing macros.  
*
*  NOTE:  If using any macros which referrence "API_Name" the macro 
*     DBG_FUNC must be used first.
*
****************************************************************************/

#if !defined(DEBUG_H)
#define DEBUG_H

#ifdef __cplusplus
extern "C" {
#endif

#if defined (DEBUG)

   #define DBG_FUNC(API_Name) \
         char func_name[50] = API_Name; \
         DbgOutput("\n");  \
         DbgOutput(">** %s\n", func_name);

   #define DBG_RETURN(err) \
       {   \
         DbgOutput("<** %s - returning: %d, 0x%x\n", func_name, \
                         err, err ); \
         DbgOutput("\n"); \
         return(err); \
       }

   #define DBG_STRING(string) \
         DbgOutput("%s\n", string);

   #define DBG_OPENSSL_ERROR(err) \
         DbgOpenSSLError(err);

   #define DBG_PARM(FormatStr, parm) \
         DbgOutput(FormatStr, parm );
         
   #define DBG_LINE() \
         DbgOutput("\n");

   #define DBG_OUTPUT(FormatStr, parm) \
         DbgOutput(FormatStr, parm );
         
   #define DBG_DUMP_BUF(len, buf) \
         DbgDumpBuf(len, buf);
         
   #define DBG_DUMP_FDSET(string, fdset) \
         DbgDumpFDSet(string, fdset);
         
   #define DBG_RETURN_PARM(ccode, operation) \
      if (0==ccode)\
      {            \
         operation \
      }
#else

   #define DBG_STRING(string)
   #define DBG_RETURN(err) \
         {return(err);}
   #define DBG_FUNC(API_Name)
   #define DBG_PARM(FormatStr, parm)
   #define DBG_LINE()
   #define DBG_OUTPUT(FormatStr, parm)
   #define DBG_DUMP_BUF(len, buf)
   #define DBG_DUMP_FDSET(string, fdset)
   #define DBG_RETURN_PARM(ccode, operation)
   #define DBG_OPENSSL_ERROR(err)

#endif



extern void DbgOutput(char *szFormat, ...);

extern void DbgOpenSSLError(int err);

extern void DbgReturn(int err, char* szAPI_Name);

extern void DbgDumpBuf(int len, const unsigned char* buf);
extern void DbgDumpFDSet(char* name, fd_set *fdset);
extern char *DbgAddressFamily(int af);
extern char *DbgSockType(int type);
extern char *DbgProtocol(int protocol);
extern char *DbgCertType(int type);
extern char *DbgEncodingType(int type);
extern char *DbgVerifyMode(int mode);
extern char *DbgCertAttr(int attrID);

#ifdef __cplusplus
}
#endif

#endif /* !defined(DEBUG_H) */


