 /*
 * $Novell: events.c,v 1.22 2003/07/01 16:03:46 $
 ******************************************************************************
 * Copyright (c) 2006 Novell, Inc.
 * All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; version 2.1 of the license.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, contact Novell, Inc.
 *
 * To contact Novell about this file by physical or electronic mail,
 * you may find current contact information at www.novell.com
 *
 ****************************************************************************
 *
 * Copyright (C) 2001 Novell, Inc. All Rights Reserved.
 * 
 * THIS WORK IS SUBJECT TO U.S. AND INTERNATIONAL COPYRIGHT LAWS AND
 * TREATIES. USE, MODIFICATION, AND REDISTRIBUTION OF THIS WORK IS SUBJECT
 * TO VERSION 2.0.7 OF THE OPENLDAP PUBLIC LICENSE, A COPY OF WHICH IS
 * AVAILABLE AT HTTP://WWW.OPENLDAP.ORG/LICENSE.HTML OR IN THE FILE "LICENSE"
 * IN THE TOP-LEVEL DIRECTORY OF THE DISTRIBUTION. ANY USE OR EXPLOITATION
 * OF THIS WORK OTHER THAN AS AUTHORIZED IN VERSION 2.0.7 OF THE OPENLDAP
 * PUBLIC LICENSE, OR OTHER PRIOR WRITTEN CONSENT FROM NOVELL, COULD SUBJECT
 * THE PERPETRATOR TO CRIMINAL AND CIVIL LIABILITY. 
 ******************************************************************************
 */

//#include "portable.h"
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>

#ifdef HAVE_STRING_H
#include <string.h>
#endif

//#include "ldap-int.h"
#include "ldap.h"  
#include <ldapx.h>
#include "ldap_events.h"

/* ---------------------------------------------------------------------------
	defines for tag values for the ber encoding of the control info
   ---------------------------------------------------------------------------*/
#define LDAP_TAG_OCTET_STRING	((ber_tag_t) 0x04U)	/* octect string */
#define LDAP_TAG_SET_OF         ((ber_tag_t) 0x31U) /* set of */

#define MAX_DN_CHARS 256 

/* Tags for the various event data sequences */
#define EVT_TAG_MASK      		0X60U /*application and constructed  */
#define EVT_TAG_ENTRY_EVENT_DATA	  ((ber_tag_t)(EVT_TAG_MASK | 1U))
#define EVT_TAG_VALUE_EVENT_DATA      ((ber_tag_t)(EVT_TAG_MASK | 2U))
#define EVT_TAG_GENERAL_EVENT_DATA    ((ber_tag_t)(EVT_TAG_MASK | 3U))
#define	EVT_TAG_SKULK_DATA  		  ((ber_tag_t)(EVT_TAG_MASK | 4U))
#define EVT_TAG_BINDERY_EVENT_DATA    ((ber_tag_t)(EVT_TAG_MASK | 5U))
#define EVT_TAG_DSESEV_INFO 	      ((ber_tag_t)(EVT_TAG_MASK | 6U))
#define EVT_TAG_MODULE_STATE_DATA  	  ((ber_tag_t)(EVT_TAG_MASK | 7U))
#define EVT_TAG_NETWORK_ADDRESS       ((ber_tag_t)(EVT_TAG_MASK | 8U))
#define EVT_TAG_CONNECTION_STATE      ((ber_tag_t)(EVT_TAG_MASK | 9U))
#define EVT_TAG_CHANGE_SERVER_ADDRESS ((ber_tag_t)(EVT_TAG_MASK | 12U))
#define EVT_TAG_CHANGE_CONFIG_PARAM   ((ber_tag_t)(EVT_TAG_MASK | 13U))
#define EVT_TAG_NO_DATA				  ((ber_tag_t)(EVT_TAG_MASK | 14U))
#define EVT_TAG_STATUS_LOG            ((ber_tag_t)(EVT_TAG_MASK | 15U))
#define EVT_TAG_DEBUG_EVENT_DATA      ((ber_tag_t)(EVT_TAG_MASK | 16U))


/*tags for the elements of the general event data*/
#define EVT_TAG_CONTEXT_MASK          0X80U
#define EVT_TAG_CONTEXT_CONSTR_MASK   0XA0U

#define EVT_TAG_GEN_DSTIME   ((long)(EVT_TAG_CONTEXT_MASK | 1U))
#define EVT_TAG_GEN_MILLISEC   ((long)(EVT_TAG_CONTEXT_MASK | 2U))
#define EVT_TAG_GEN_VERB ((long)(EVT_TAG_CONTEXT_MASK | 3U))
#define EVT_TAG_GEN_CURRPROC  ((long)(EVT_TAG_CONTEXT_MASK | 4U))
#define EVT_TAG_GEN_PERP ((long)(EVT_TAG_CONTEXT_MASK | 5U))
#define EVT_TAG_GEN_INTEGERS  ((long)(EVT_TAG_CONTEXT_MASK | 6U))
#define EVT_TAG_GEN_STRINGS  ((long)(EVT_TAG_CONTEXT_CONSTR_MASK | 7U))

/*tags for the elements of the config param event data*/
#define EVT_TAG_CFG_BOOLEAN   ((ber_tag_t)(EVT_TAG_CONTEXT_MASK | 1U))
#define EVT_TAG_CFG_INTEGER   ((ber_tag_t)(EVT_TAG_CONTEXT_MASK | 2U))
#define EVT_TAG_CFG_BINARY    ((ber_tag_t)(EVT_TAG_CONTEXT_MASK | 3U))
#define EVT_TAG_CFG_UTF8      ((ber_tag_t)(EVT_TAG_CONTEXT_MASK | 4U))
#define EVT_TAG_CFG_NULL      ((ber_tag_t)(EVT_TAG_CONTEXT_MASK | 5U))

/*tags for the elements of the debug info event data*/
#define EVT_TAG_DB_ENTRYID      ((long)(EVT_TAG_CONTEXT_MASK | 1U))
#define EVT_TAG_DB_STRING       ((long)(EVT_TAG_CONTEXT_MASK | 2U))
#define EVT_TAG_DB_RAWDATA      ((long)(EVT_TAG_CONTEXT_MASK | 3U))
#define EVT_TAG_DB_INTEGER      ((long)(EVT_TAG_CONTEXT_MASK | 4U))
#define EVT_TAG_DB_ADDRESS      ((long)(EVT_TAG_CONTEXT_MASK | 5U))
#define EVT_TAG_DB_TIMESTAMP    ((long)(EVT_TAG_CONTEXT_MASK | 6U))
#define EVT_TAG_DB_TIMEVECTOR   ((long)(EVT_TAG_CONTEXT_MASK | 7U))

/* ---------------------------------------------------------------------------
    ldap_monitor_events

    Registers with the server to monitor one or more events. The function 
    sends an EventMonitorRequest extended operation to the server. The request
    is sent asynchronously; it does not wait for a response from the server. 
    The function returns the constant LDAP_SUCCESS if the request was 
    successfully sent, or another LDAP result code if not. Responses are 
    received by calling the ldap_result function. The response is interpreted
    by calling the ldap_parse_ds_event function.

    ld       (IN) An LDAP session handle, as obtained from a call to
				  ldap_init().

    eventCount (IN) The number of event specifiers in the array specified
	              by the events parameter.

    events	 (IN) An array of EVT_EventSpecifier structures describing the 
                  events the application wishes to monitor. The eventType 
                  field of the last entry in the array should be equal to 
                  EVT_INVALID to indicate the end of the array.

    msgID	(OUT) This output parameter will be set to the message id of the
                  request if the ldap_monitor_event() call succeeds. The value
                  is undefined if the function returns a value other than 
                  LDAP_SUCCESS.

    The LBER encoding of the ASN.1 definition of EventMonitorRequest is 
    shown below:
 
     EventMonitorRequest ::= SEQUENCE {
                   eventCount  INTEGER,
                   events      SET OF {eventSpecifier }

     eventSpecifier ::= SEQUENCE {
          eventType    INTEGER,
          eventStatus  ENUMERATED {
                         AllEvents          (0),
                         SuccessfulEvents   (1)
                         FailedEvents       (2)} }


   ---------------------------------------------------------------------------*/
LDAP_F( int )
ldap_monitor_events(
    LDAP 				 *ld,
    int                  eventCount,
    EVT_EventSpecifier    *events,
    int 				 *msgID)
{
	BerElement     *requestBer = NULL;
	int            err = LDAP_SUCCESS;
    struct berval  *requestValue = NULL;
	int            i = 0 ;
 
    /* Must specify an array of events */
    if ( (0 == eventCount) ||
         (NULL == events) )
	    return LDAP_PARAM_ERROR;

	/* Allocate a BerElement for stuffing in the function IN parameters */
	requestBer = ber_alloc();
	if(requestBer == NULL)
		return LDAP_NO_MEMORY;
 
    /* Start sequence and write the eventCount */
   	err = ber_printf(requestBer, "{i", eventCount);
    if (LBER_ERROR == err)
	{
	    err = LDAP_ENCODING_ERROR;
		goto cleanup;
	}

    /* Start the set of EventSpecifiers*/
   	err = ber_printf(requestBer, "[");
    if (LBER_ERROR == err)
	{
	    err = LDAP_ENCODING_ERROR;
		goto cleanup;
	}
   
   	/* Add the individual EventSpecifiers */
    for (i=0; i<eventCount; i++)
	{
		err = ber_printf(requestBer, "{ie}", 
		                 events[i].eventType, 
		                 events[i].eventStatus);
		if (err == LBER_ERROR)
		{
			err = LDAP_ENCODING_ERROR;
			goto cleanup;
		}
	}

    /* End the set of EventSpecifiers and the enclosing sequence*/
   	err = ber_printf(requestBer, "]}");
    if (LBER_ERROR == err)
	{
	    err = LDAP_ENCODING_ERROR;
		goto cleanup;
	}

	/* Converts the BER we just built to a berval that we will send out
	  in the extended request */
	if (ber_flatten(requestBer, &requestValue) == LBER_ERROR)
	{
		err = LDAP_NO_MEMORY;
		goto cleanup;
	}

    err = ldap_extended_operation(
	    ld,
	    NLDAP_MONITOR_EVENTS_REQUEST,
	    requestValue,
	    NULL,	/* no server controls */
	    NULL,  /* no client controls */
	    msgID);
 
cleanup:
	/* Free memory that we allocated when building the request ber and berval */
    if (NULL != requestValue)
	    ber_bvfree(requestValue);

    if (NULL != requestBer)
	    ber_free(requestBer, 1);

    //ld->ld_errno = err;

    return err;
}

/* ---------------------------------------------------------------------------
    ldap_monitor_events_filtered

    Registers with the server to monitor one or more events and specifies 
    filter expressions to be used by the server to filter the event 
    notifications that are sent by the server. The function sends a
    FilteredEventMonitorRequest extended operation to the server. The request
    is sent asynchronously; it does not wait for a response from the server. 
    The function returns the constant LDAP_SUCCESS if the request was 
    successfully sent, or another LDAP result code if not. Responses are 
    received by calling the ldap_result function. The response is interpreted
    by calling the ldap_parse_ds_event or ldap_parse_monitor_events_response 
    functions.

    ld       (IN) An LDAP session handle, as obtained from a call to
				  ldap_init().

    eventCount (IN) The number of event specifiers in the array specified
	              by the events parameter.

    events	 (IN) An array of EVT_FilteredEventSpecifier structures describing
                  the events the application wishes to monitor. The array must
				  contain at least eventCount elements.

    msgID	(OUT) This output parameter will be set to the message id of the
                  request if the ldap_monitor_event() call succeeds. The value
                  is undefined if the function returns a value other than 
                  LDAP_SUCCESS.

    The LBER encoding of the ASN.1 definition of EventMonitorRequest is 
    shown below:
 
    FilteredEventMonitorRequest ::= SEQUENCE {
                   eventCount  INTEGER,
                   events      SET OF FilteredEventSpecifier }

    FilteredEventSpecifier ::= SEQUENCE {
	               eventType			INTEGER,
	               eventStatus			ENUMERATED {
				                          AllEvents 	       (0),
								          SuccessfulEvents   (1),
								          FailedEvents       (2)},
	               eventFilter       OPTIONAL FilterString }
   ---------------------------------------------------------------------------*/
LDAP_F( int)
ldap_monitor_events_filtered(
	LDAP                        *ld,
	int	                        eventCount,
	EVT_FilteredEventSpecifier  *events,
	int                         *msgID)
{
	BerElement     *requestBer = NULL;
	int            err = LDAP_SUCCESS;
    struct berval  *requestValue = NULL;
	int            i = 0 ;
 
    /* Must specify an array of events */
    if ( (0 == eventCount) ||
         (NULL == events) )
	    return LDAP_PARAM_ERROR;

	/* Allocate a BerElement for stuffing in the function IN parameters */
	requestBer = ber_alloc();
	if(requestBer == NULL)
		return LDAP_NO_MEMORY;
 
    /* Start sequence and write the eventCount */
   	err = ber_printf(requestBer, "{i", eventCount);
    if (LBER_ERROR == err)
	{
	    err = LDAP_ENCODING_ERROR;
		goto cleanup;
	}

    /* Start the set of EventSpecifiers*/
   	err = ber_printf(requestBer, "[");
    if (LBER_ERROR == err)
	{
	    err = LDAP_ENCODING_ERROR;
		goto cleanup;
	}
   
   	/* Add the individual EventSpecifiers */
    for (i=0; i<eventCount; i++)
	{
		err = ber_printf(requestBer, "{ies}", 
		                 events[i].eventType, 
		                 events[i].eventStatus,
		                 events[i].filter);
		if (err == LBER_ERROR)
		{
			err = LDAP_ENCODING_ERROR;
			goto cleanup;
		}
	}

    /* End the set of EventSpecifiers and the enclosing sequence*/
   	err = ber_printf(requestBer, "]}");
    if (LBER_ERROR == err)
	{
	    err = LDAP_ENCODING_ERROR;
		goto cleanup;
	}

	/* Converts the BER we just built to a berval that we will send out
	  in the extended request */
	if (ber_flatten(requestBer, &requestValue) == LBER_ERROR)
	{
		err = LDAP_NO_MEMORY;
		goto cleanup;
	}

    err = ldap_extended_operation(
	    ld,
	    NLDAP_FILTERED_MONITOR_EVENTS_REQUEST,
	    requestValue,
	    NULL,	/* no server controls */
	    NULL,  /* no client controls */
	    msgID);
 
cleanup:
	/* Free memory that we allocated when building the request ber and berval */
    if (NULL != requestValue)
	    ber_bvfree(requestValue);

    if (NULL != requestBer)
	    ber_free(requestBer, 1);

    return err;
}

/* ---------------------------------------------------------------------------
    ldap_parse_monitor_events_response

    After a successful call to ldap_monitor event(), server responses to the 
    EventMonitorRequest are retrieved by calling the ldap_result function. If 
    the return value of ldap_result is equal to LDAP_RES_EXTENDED, the result
    is parsed by calling the ldap_parse_ds_event_response.

    ld       (IN) An LDAP session handle, as obtained from a call to
				  ldap_init().

    eventMessage  (IN)	a pointer to the LDAPMessage returned by ldap_result().

    resultCode (OUT)   This output parameter receives the resultCode field from 
                       the LDAP result. Possible values are:
                        
                       LDAP_OPERATIONS_ERROR - indicates that an error occurred
                           on the server while trying to perform the operation.
                           The errorMessage may give an indication of the error.

                       LDAP_PROTOCOL_ERROR - an invalid event type or status 
                           was specified in the request. The invalid invents
                           will be contained in the badEvents array.

                       LDAP_ADMINLIMIT_EXCEEDED - the maximum number of active
                           Monitor Events are already active on this server.

                       LDAP_INSUFFICIENT_ACCESS - the bound object associated
                           with the connection does not have rights to perform
                           a Monitor Events request.

                       LDAP_UNAVAILABLE - although the extension is generally
                           supported by the server, the event service is not 
                           currently available.
                       
    badEvents (OUT) If the value of resultCode is LDAP_PROTOCOL_ERROR, 
                    this output parameter receives an array of 
                    EVT_EventSpecifier structures identifying any unrecognized 
                    events. Otherwise, the parameter is set to NULL.

    The response field will contain the LBER encoding of the 
    EventExtendedResponse whose ASN.1 definition is given below.

     EventExtendedResponse ::= SET OF eventSpecifier

     eventSpecifier ::= SEQUENCE {
          eventType    INTEGER,
          eventStatus  ENUMERATED {
                         AllEvents          (0),
                         SuccessfulEvents   (1)
                         FailedEvents       (2)} }

   ---------------------------------------------------------------------------*/
LDAP_F( int )
ldap_parse_monitor_events_response(
    LDAP         	   *ld, 
 	LDAPMessage        *eventMessage,
    int          	   *resultCode,
    char               **errorMsg,
	int                *badEventCount,
    EVT_EventSpecifier  **badEvents,
    int                freeIt)
{
	BerElement        *responseBer = NULL;
    struct berval     	  *extRespData = NULL;
	ber_len_t         temp = 0;
	char              *retOID = NULL ;
	EVT_EventSpecifier *currEvent = NULL;
	int      	  i = 0;
	int     	  count = 0;
	int     	  err = LDAP_SUCCESS;

   /* responseCode must be non-null */
    if ( NULL == resultCode )
	   return err = LDAP_PARAM_ERROR;

    /* initialize the badEvents array to NULL */
    if (badEvents != NULL)
	    *badEvents = NULL;

    if (badEventCount != NULL)
		*badEventCount = 0;

    if (errorMsg != NULL)
       *errorMsg = NULL;

    err = ldap_parse_result(ld, eventMessage, resultCode, NULL,
                                     errorMsg, NULL, NULL, 0);

    if (err != LDAP_SUCCESS)
	    goto cleanup;

    err = ldap_parse_extended_result(ld, eventMessage,
                                              &retOID, &extRespData, freeIt);

    if ( (err != LDAP_SUCCESS) ||
         (NULL == retOID) )
	    goto cleanup;

    /* verifiy the OID of the result */
    if ( 0 != strcmp(retOID, NLDAP_MONITOR_EVENTS_RESPONSE) )
	{
	    err = LDAP_DECODING_ERROR;
		goto cleanup;
	}

    /* if we have bad events and caller wants them, then get 'em  */
	if ( (extRespData != NULL) && 
	     (badEvents != NULL) && 
	     (badEventCount != NULL) )
	{
	    /* Create a BerElement from the berval returned  */
        responseBer = ber_init(extRespData);
        if (responseBer == NULL)
        {
            err = LDAP_NO_MEMORY;
            goto cleanup;
        }
    
		/* Check for a sequence and get the number of elements */
		if (LBER_ERROR == ber_scanf(responseBer, "{i", &count)) 
		{
			goto cleanup;
		}

	    if (LDAP_TAG_SET_OF == ber_peek_tag(responseBer, &temp))
		{
            ber_scanf(responseBer, "[");
		    /* allocate memory for bad events array */
		    *badEvents = (EVT_EventSpecifier*)malloc(
		      sizeof(EVT_EventSpecifier) * count);
			if (NULL == *badEvents)
    		{
        		err = LDAP_NO_MEMORY;
        		goto cleanup;
    		}
			else
			    currEvent = *badEvents;

		    for (i=0; i<count; i++)
			{
			    if (LBER_ERROR == ber_scanf(responseBer, "{ie}", 
			                                &(currEvent->eventType), 
			                                &(currEvent->eventStatus)) )
				{
				    err = LDAP_DECODING_ERROR;
					goto cleanup;
				}
				currEvent ++;
			}
			*badEventCount = count;
		}	
	}
    
cleanup:
	/* Free memory that we allocated when building the request ber and berval */
	if (extRespData)
		ber_bvfree(extRespData);

	if (responseBer)
		ber_free(responseBer, 1); 

    if ( (err != LDAP_SUCCESS) && 
         (badEvents != NULL) && (*badEvents != NULL ) )
	{
	    free(*badEvents);
		*badEvents = NULL;
	}

    return err;

}
/* ---------------------------------------------------------------------------
    allocEventData

    This allocates memory for a data structure to hold that events datafunction completes the parsing of a berelement that contains a 
	response for an Entry event.

    size   (IN) the number of bytes needed.

    tag    (IN) the type of the event data to be stored in the memory
   ---------------------------------------------------------------------------*/
void * allocEventData(
    size_t   size,
    ber_tag_t tag)
{
	ber_tag_t    *pTag = NULL ;
	void         *data = NULL;
	
    //allocate memory for the event data plus an initial tag to identify
	//the type of data for the evt_free() function.
    pTag = (ber_tag_t*)malloc(size + sizeof(ber_tag_t));

    if (NULL != pTag)
	{
        *pTag = tag;
	    pTag++;  /* point to the start of the real data */
	}

    return pTag;
}

/* ---------------------------------------------------------------------------
    parseOctetString

    responseBer  (IN) the ber element containing the response to be parsed.

    pString   (OUT) a pointer to the pointer to the string to be allocated
    
    size      (OUT) a pointer to the size of the octet string
    
    expectedTag (IN) the tag expected on the octet string. If zero, then
                     LBER_OCTETSTRING is used.
   ---------------------------------------------------------------------------*/
int parseOctetString(
    BerElement  *responseBer, 
    char		**pString,
    ber_len_t   *size,
    ber_tag_t   expectedTag)
{
    ber_tag_t  tag = 0;
    int        result = LDAP_SUCCESS;
    ber_len_t  sizePlusNull = 0 ;

    if (0 == expectedTag)
        expectedTag = LBER_OCTETSTRING;
        
    /* make sure it is an Octet String or the specified tag,
     * the size of the data is set here. */
    tag = ber_peek_tag(responseBer, size);
	assert(expectedTag == tag);
	if (expectedTag != tag)
	{
	    *pString = NULL;
        *size = 0;
	    result = LDAP_DECODING_ERROR;
	}
	else if (0 != *size)
	{
        /*add one for the null character added by ber_scanf*/
	    sizePlusNull = (*size) + 1; 
	    *pString = (char*)malloc(sizePlusNull);
		if (NULL == *pString)
		{
            *size =  0;
		    result = LDAP_NO_MEMORY;
		}
		else
		{
		    if (LBER_ERROR == ber_scanf(responseBer, "s", 
		                            *pString, &sizePlusNull))
		    { 
                free(*pString);
                *pString = NULL;
                *size = 0;
		        result = LDAP_DECODING_ERROR;
		    } 
        }
	}
	else
	{  /* skip a zero length string */
        ber_skip_tag(responseBer, size);
	    *pString = NULL;
        *size = 0;
	}

    return result;
}

/* ---------------------------------------------------------------------------
    parseReferralList
   
    responseBer  (IN) the ber element containing the response to be parsed.

    pList   (OUT) a pointer to a pointer to the array of EVT_ReferralAddress 
                  pointers allocated by this function.
                  
   NOTE: Memory allocated by this function must be freed by the caller even if an
   error result is returned.

   Data format:
      
   referralList := SEQUENCE {
			count		INTEGER,
			referrals	SEQUENCE OF ReferralAddress }

	ReferralAddress := SEQUENCE {
		addressType		INTEGER,
		address			OCTET STRING}

   ---------------------------------------------------------------------------*/
int parseReferralList(
    BerElement           *responseBer, 
    int                  *count,
    EVT_ReferralAddress  **pList)
{
    int            result = LDAP_SUCCESS;
    int            i = 0;
    ber_len_t      length = 0 ;

    *pList = NULL;
    *count = 0;

    if (LBER_ERROR == ber_scanf(responseBer, "{"))
    {
        goto exit;  /*no error, assume no referral list present*/
    }

    /* get the count and start of SEQUENCE OF */
    if (LBER_ERROR == ber_scanf(responseBer, "i{", count))
    {
        result = LDAP_DECODING_ERROR;
        goto exit;
    }

    *pList = calloc(*count, sizeof(EVT_ReferralAddress));

    if (NULL == *pList)
    {
        result = LDAP_NO_MEMORY;
        goto exit;
    }

    for (i=0; i<*count; i++)
    {
        if (LBER_ERROR == ber_scanf(responseBer, "{i", &(*pList)[i].type))
        {
            result = LDAP_DECODING_ERROR;
            goto exit;
        }
        result = parseOctetString(responseBer, 
                                  &(*pList)[i].address, 
                                  &length, 0);
        if (result != LDAP_SUCCESS)
        {
            (*pList)[i].length = 0;
            goto exit;
        }
        (*pList)[i].length = (int)length;
     }

 exit:
    return result;
}   

/* ---------------------------------------------------------------------------
    parseLDAPString


    responseBer  (IN) the ber element containing the response to be parsed.

    pString   () a pointer to the pointer to the string to be allocated
   ---------------------------------------------------------------------------*/
int parseLDAPString(
    BerElement  *responseBer, 
    char		**pString,
    ber_tag_t   expectedTag)
{
    ber_tag_t  tag = 0 ;
    ber_len_t  size = 0;
    int        result = LDAP_SUCCESS;

    if (0 == expectedTag)
        expectedTag = LBER_OCTETSTRING;
        
    /* make sure it is an Octet String */
    tag = ber_peek_tag(responseBer, &size);
	assert(expectedTag == tag);
	if (expectedTag != tag)
	{
	    *pString = NULL;
	    result = LDAP_DECODING_ERROR;
	}
	else if (0 != size)
	{
	    size++; /*for the null character*/
	    *pString = (char*)malloc(size);
		if (NULL == *pString)
		{
		    result = LDAP_NO_MEMORY;
		}
		else
		{
		    if (LBER_ERROR == ber_scanf(responseBer, "s", 
		                            *pString, &size))
		    { 
                free(*pString);
                *pString = NULL;
		        result = LDAP_DECODING_ERROR;
		    } 
        }
	}
	else
	{  /* skip a zero length string */
        ber_skip_tag(responseBer, &size);
	    *pString = NULL;
	}

    return result;
}

/* ---------------------------------------------------------------------------
    parseDebugEventData

    This function completes the parsing of a berelement that contains a 
	response Debug event.

    responseBer  (IN) the ber element containing the response to be parsed.

    eventData   (OUT) a pointer to the EVT_DebugInfo structure 
                      containing the parsed data.

	DebugEventData ::= [APPLICATION 13] SEQUENCE {
		 DsTime            INTEGER,
		 Milliseconds      INTEGER,
		 Perpetrator       LDAPDN,
		 FormatString      LDAPSTRING,
		 Verb              INTEGER,
		 ParameterCount    INTEGER,
		 Parameters        SEQUENCE OF Parameter OPTIONAL }

	ReferralAddress ::=  SEQUENCE {
		 addressType      INTEGER,
		 address          OCTET STRING }

	DSETimeStamp ::= SEQUENCE {
		 seconds          INTEGER,
		 replicaNumber    INTEGER,
		 event            INTEGER }

	Parameter ::= CHOICE {
		 entryDN          [1] LDAPDN,
		 stringValue      [2] OCTET STRING,
		 rawData          [3] OCTET STRING,
		 integerValue     [4] INTEGER,
		 address          [5] ReferralAddress,
		 timeStamp        [6] DSETimeStamp,
		 timeVector       [7] SEQUENCE { count INTEGER, SEQUENCE OF DSETimeStamp }

   ---------------------------------------------------------------------------*/
int parseDebugEventData(
    LDAP		*ld,
    BerElement  *responseBer, 
    void		**eventData)
{
    ber_tag_t      tag = 0;
    ber_len_t      size = 0;
	EVT_DebugInfo  *data = NULL;
    ber_int_t      count = 0;
    DB_Parameter*  param = NULL;
    int            i = 0 , j = 0;
    int     err = LDAP_SUCCESS;
	
    //allocate memory for the event data.
    *eventData = data = (EVT_DebugInfo*)
      allocEventData(sizeof(EVT_DebugInfo), EVT_TAG_DEBUG_EVENT_DATA);

    if (NULL == data)
    {
		err = LDAP_NO_MEMORY;
		goto error;
    }

	/* initialize data */
   data->formatString = NULL;
	data->paramCount = 0;
   data->parameters = NULL;
   data->perpetratorDN = NULL;

    /* get dsTime, milliseconds*/
	if (LBER_ERROR == ber_scanf(responseBer, "ii", 
                                &data->dsTime, &data->milliseconds))
	{
		err = LDAP_DECODING_ERROR;
		goto error;
	} 

    /* get the perpetratorDN */
    err = parseLDAPString(responseBer, &data->perpetratorDN, 0);
    if (err != LDAP_SUCCESS)
        goto error;

    /* get the format string */
    err = parseLDAPString(responseBer, &data->formatString, 0);
    if (err != LDAP_SUCCESS)
        goto error;

    /* get verb and paramCount*/
	if (LBER_ERROR == ber_scanf(responseBer, "ii", 
                                &data->verb, &data->paramCount))
	{
	    err = LDAP_DECODING_ERROR;
		goto error;
	} 
    /* get the parameters */
    if (data->paramCount > 0)
    {
        data->parameters = malloc(data->paramCount * sizeof(DB_Parameter));
        /* start the sequence of parameters */
        if ( LBER_ERROR == ber_scanf(responseBer, "{") )
	        {
	            err = LDAP_DECODING_ERROR;
		        goto error;
	        } 
    }
    
    for (i = 0 ; i < data->paramCount; i++)
    {
        param = &data->parameters[i];
        tag = ber_peek_tag(responseBer, &size);
        switch(tag)
        {    
        case EVT_TAG_DB_ENTRYID:
	        if ( LBER_ERROR == 
                  ber_scanf(responseBer, "i", &param->value.integer) )
	        {
	            err = LDAP_DECODING_ERROR;
		        goto error;
	        } 
            param->type = DB_PARAM_TYPE_ENTRYID;
            break;
        case EVT_TAG_DB_INTEGER:
	        if ( LBER_ERROR == 
                  ber_scanf(responseBer, "i", &param->value.integer) )
	        {
	            err = LDAP_DECODING_ERROR;
		        goto error;
	        } 
            param->type = DB_PARAM_TYPE_INTEGER;
            break;
        case EVT_TAG_DB_RAWDATA:
	        if (0 != size)
	        {
                param->value.binary.size = (unsigned int)size;
	            size++; /*for the null character*/
	            param->value.binary.data = (unsigned char*)malloc(size);
		        if (NULL == param->value.binary.data)
		        {
		            err = LDAP_NO_MEMORY;
			        goto error;
		        }
		        if (LBER_ERROR == ber_scanf(responseBer, "s", 
		                                    param->value.binary.data, &size))
		        {
		            err = LDAP_DECODING_ERROR;
			        goto error;
		        } 
	        }
	        else
	        {
                param->value.binary.data = NULL;
                param->value.binary.size = 0;
                ber_skip_tag(responseBer, &size);
	        }
            param->type = DB_PARAM_TYPE_BINARY;
            break;

        case EVT_TAG_DB_STRING:
	        if (0 != size)
	        {
	            size++; /*for the null character*/
	            param->value.utf8Str = malloc(size);
		        if (NULL == param->value.utf8Str)
		        {
		            err = LDAP_NO_MEMORY;
			        goto error;
		        }
		        if (LBER_ERROR == ber_scanf(responseBer, "s", 
		                                    param->value.utf8Str, &size))
		        {
		            err = LDAP_DECODING_ERROR;
			        goto error;
		        } 
	        }
	        else
	        {
                param->value.utf8Str = NULL;
                ber_skip_tag(responseBer, &size);
	        }
            if (tag == EVT_TAG_DB_ENTRYID)
                param->type = DB_PARAM_TYPE_ENTRYID;
            else
                param->type = DB_PARAM_TYPE_STRING;
            break;

        case EVT_TAG_DB_TIMESTAMP:
	        if ( LBER_ERROR == 
                  ber_scanf(responseBer, "{iii}", 
                            &param->value.timeStamp.seconds,
                            &param->value.timeStamp.replicaNumber,
                            &param->value.timeStamp.event) )
	        {
	            err = LDAP_DECODING_ERROR;
		        goto error;
	        } 
            param->type = DB_PARAM_TYPE_TIMESTAMP;
            break;
        
        case EVT_TAG_DB_TIMEVECTOR:
            if (LBER_ERROR == ber_scanf(responseBer, "{i{", 
                                        &count))
            {
                err = LDAP_DECODING_ERROR;
                goto error;
            }
            else
            {
                param->value.timeStampVector.count = count;
                param->value.timeStampVector.timeStamps = 
                 malloc(count * sizeof(EVT_TimeStamp));
   		        if (NULL == param->value.binary.data)
		        {
		            err = LDAP_NO_MEMORY;
			        goto error;
		        }
                for (j=0; j<count; j++)
                {
                    if (LBER_ERROR == 
                        ber_scanf(responseBer, "{iii}",
                            &param->value.timeStampVector.timeStamps[j].seconds,
                            &param->value.timeStampVector.timeStamps[j].replicaNumber,
                            &param->value.timeStampVector.timeStamps[j].event) )
	                {
	                    err = LDAP_DECODING_ERROR;
		                goto error;
	                } 
                }
                if (LBER_ERROR == ber_scanf(responseBer, "}", 
                                            &count))
                {
                    err = LDAP_DECODING_ERROR;
                    goto error;
                }
                param->type = DB_PARAM_TYPE_TIMEVECTOR;
            }
            break;
        
        case EVT_TAG_DB_ADDRESS:
            /* get the type */
	        if (LBER_ERROR == ber_scanf(responseBer, "{i", 
                                        &param->value.netAddress.type))
	        {
	            err = LDAP_DECODING_ERROR;
		        goto error;
	        } 

            /* get the data and its size */
            tag = ber_peek_tag(responseBer, &size);
	        if (LBER_OCTETSTRING == tag)
            {
                if (0 != size) 
	            {   /* allocate memory for the event data plus enough for the 
                    * address data on the end plus the null terminator.
                    */
                    param->value.netAddress.length = (unsigned int)size;
                    size++;
                    param->value.netAddress.data = malloc(size);
		            if (NULL == data)
		            {
		                err = LDAP_NO_MEMORY;
			            goto error;
		            }
		            if (LBER_ERROR == 
                         ber_scanf(responseBer, "s}", 
		                           param->value.netAddress.data, &size))
		            {
		                err = LDAP_DECODING_ERROR;
			            goto error;
		            }
	            }
	            else
	            {
                    param->value.netAddress.length = 0;
		            param->value.netAddress.data = NULL;
                }
	        }
            else
	        {
	            err = LDAP_DECODING_ERROR;
		        goto error;
	        } 
            param->type = DB_PARAM_TYPE_ADDRESS;
            break;
        
        default:
	        err = LDAP_DECODING_ERROR;
		    goto error;
            break;
        }
    }

    return err;

error:
    if (NULL != data)
	{
	   ldap_event_free(data);
	   *eventData = NULL;
	}

	return err;
}

/* ---------------------------------------------------------------------------
    parseChangeServerAddressData

    This function completes the parsing of a berelement that contains a 
	response for an Entry event.

    responseBer  (IN) the ber element containing the response to be parsed.

    eventData   (OUT) a pointer to the EVT_ChangeServerAddr structure 
                      containing the parsed data.

		ChangeServerAddress ::= [APPLICATION 12]SEQUENCE {
			flags		INTEGER,
			proto		INTEGER,
			addrFamily  INTEGER,
			address		OCTET STRING,
			pstkname	LDAPSTRING,
			source		LDAPSTRING }
          
   ---------------------------------------------------------------------------*/
int parseChangeServerAddressData(
    LDAP		*ld,
    BerElement  *responseBer, 
    void		**eventData)
{
    ber_tag_t              tag = 0;
    ber_len_t              size = 0;
	EVT_ChangeServerAddr  *data = NULL;
	int     err = LDAP_SUCCESS;	

    //allocate memory for the event data.
    *eventData = data = (EVT_ChangeServerAddr*)
      allocEventData(sizeof(EVT_ChangeServerAddr), EVT_TAG_CHANGE_SERVER_ADDRESS);

    if (NULL == data)
    {
		err = LDAP_NO_MEMORY;
		goto error;
    }

	/* initialize data */
	data->addr = NULL;
	data->pstkname = NULL;

    /* get type, emuObjFlags, and security */
	if (LBER_ERROR == ber_scanf(responseBer, "iii", 
	                            &data->flags,
	                            &data->proto,
	                            &data->addrFamily))
	{
	    err = LDAP_DECODING_ERROR;
		goto error;
	} 

    /* get the addr */
    tag = ber_peek_tag(responseBer, &size);
	assert(LBER_OCTETSTRING == tag);
	if (LBER_OCTETSTRING != tag)
	{
	    err = LDAP_DECODING_ERROR;
	    goto error;
	}
	if (0 != size)
	{
	    size++; /*for the null character*/
	    data->addr = (unsigned char*)malloc(size);
		if (NULL == data->addr)
		{
		    err = LDAP_NO_MEMORY;
			goto error;
		}
		if (LBER_ERROR == ber_scanf(responseBer, "s", 
		                            data->addr, &data->addrSize))
		{
		    err = LDAP_DECODING_ERROR;
			goto error;
		} 
	}
	else
	{
        data->addrSize = 0;
        ber_skip_tag(responseBer, &size);
	}

    /* get the pstkname */
    err = parseLDAPString(responseBer, &data->pstkname, 0);
    if (err != LDAP_SUCCESS)
        goto error;

    /* get the sourceModule */
    err = parseLDAPString(responseBer, &data->sourceModule, 0);
    if (err != LDAP_SUCCESS)
        goto error;

    return err;

error:
    if (NULL != data)
	{
	   ldap_event_free(data);
	   *eventData = NULL;
	}

	return err;
}

/* ---------------------------------------------------------------------------
    parseChangeConfigParamData

    This function completes the parsing of a berelement that contains a 
	response for an Entry event.

    responseBer  (IN) the ber element containing the response to be parsed.

    eventData   (OUT) a pointer to the EVT_ChangeConfigParm structure 
                      containing the parsed data.

		ChangeConfigParameter ::= [APPLICATION 13]SEQUENCE {
			type			INTEGER,
			name			LDAPSTRING,
			value       CHOICE {
								booleanVal [1] BOOLEAN
								integerVal [2] INTEGER
								binaryVal  [3] OCTET STRING
								utf8Val    [4] OCTET STRING } OPTIONAL
							}
                           
   ---------------------------------------------------------------------------*/
int parseChangeConfigParamData(
    LDAP		*ld,
    BerElement  *responseBer, 
    void		**eventData)
{
    ber_tag_t              tag = 0;
    ber_len_t              size = 0;
	EVT_ChangeConfigParm  *data = NULL;
	int     err = LDAP_SUCCESS;	

    //allocate memory for the event data.
    *eventData = data = (EVT_ChangeConfigParm*)
      allocEventData(sizeof(EVT_ChangeConfigParm), EVT_TAG_CHANGE_CONFIG_PARAM);

    if (NULL == data)
    {
		err = LDAP_NO_MEMORY;
		goto error;
    }

	/* initialize data */
	data->name = NULL;
    data->value.binary.data = NULL;
    data->value.binary.size = 0;

    /* get type, emuObjFlags, and security */
	if (LBER_ERROR == ber_scanf(responseBer, "i", &data->type))
	{
	    err = LDAP_DECODING_ERROR;
		goto error;
	} 

    /* get the name */
    err = parseLDAPString(responseBer, &data->name, 0);
    if (err != LDAP_SUCCESS)
        goto error;

    /* get the data */
    tag = ber_peek_tag(responseBer, &size);
    switch(tag)
    {    
    case EVT_TAG_CFG_BOOLEAN:
        if ( (EVT_CFG_TYPE_BOOLEAN != data->type) ||
	         (LBER_ERROR == ber_scanf(responseBer, "i", &data->value.boolean)) )
	    {
	        err = LDAP_DECODING_ERROR;
		    goto error;
	    } 
        break;
    case EVT_TAG_CFG_INTEGER:
	    if ( (EVT_CFG_TYPE_INT != data->type) ||
	         (LBER_ERROR == ber_scanf(responseBer, "i", &data->value.integer)) )
	    {
	        err = LDAP_DECODING_ERROR;
		    goto error;
	    } 
        break;
    case EVT_TAG_CFG_BINARY:
        if (EVT_CFG_TYPE_BINARY != data->type)
	    {
	        err = LDAP_DECODING_ERROR;
		    goto error;
	    } 
	    if (0 != size)
	    {
	        size++; /*for the null character*/
	        data->value.binary.data = (unsigned char*)malloc(size);
		    if (NULL == data->value.binary.data)
		    {
		        err = LDAP_NO_MEMORY;
			    goto error;
		    }
		    if (LBER_ERROR == ber_scanf(responseBer, "s", 
		                                data->value.binary.data, &data->value.binary.size))
		    {
		        err = LDAP_DECODING_ERROR;
			    goto error;
		    } 
	    }
	    else
	    {
            ber_skip_tag(responseBer, &size);
	    }
        break;

    case EVT_TAG_CFG_UTF8:
        if (EVT_CFG_TYPE_STRING != data->type)
	    {
	        err = LDAP_DECODING_ERROR;
		    goto error;
	    } 
        err = parseLDAPString(responseBer, &data->value.utf8Str, 
                                       EVT_TAG_CFG_UTF8);
        if (err != LDAP_SUCCESS)
            goto error;
        break;
        
    case EVT_TAG_CFG_NULL:
        if (EVT_CFG_TYPE_NULL != data->type)
	    {
	        err = LDAP_DECODING_ERROR;
		    goto error;
	    } 
        break;
        
    default:
	        err = LDAP_DECODING_ERROR;
		    goto error;
        break;
    }

    return err;

error:
    if (NULL != data)
	{
	   ldap_event_free(data);
	   *eventData = NULL;
	}

	return err;
}

/* ---------------------------------------------------------------------------
    parseConnectionStateData

    This function completes the parsing of a berelement that contains a 
	response for an Entry event.

    ld           (IN) An LDAP session handle, as obtained from a call to
				  ldap_init().

    responseBer  (IN) the ber element containing the response to be parsed.

    eventData   (OUT) a pointer to the EVT_ChangeConnState structure containing
                      the parsed data.

    ConnectionState ::= [APPLICATION 9]SEQUENCE {
	       connectionDN     LDAPDN,
           oldFlags         INTEGER,
           newFlags			INTEGER,
	       sourceModule		LDAPSTRING }
   ---------------------------------------------------------------------------*/
int parseConnectionStateData(
    LDAP		*ld,
    BerElement  *responseBer, 
    void		**eventData)
{
	EVT_ChangeConnState	*data  = NULL;
	int     err = LDAP_SUCCESS;

    //allocate memory for the event data.
    *eventData = data = 
      (EVT_ChangeConnState*)allocEventData(sizeof(EVT_ChangeConnState), 
                                           EVT_TAG_CONNECTION_STATE);

    if (NULL == data)
    {
		err = LDAP_NO_MEMORY;
		goto error;
    }

    /* initialize string pointers */
    data->connectionDN = NULL;

    /* get the connectionDN */
    err = parseLDAPString(responseBer, &data->connectionDN, 0);
    if (err != LDAP_SUCCESS)
        goto error;

    /* get oldFlags, newFlags, source, psid */
	if (LBER_ERROR == ber_scanf(responseBer, "ii",
				                &data->oldFlags,
		    					&data->newFlags) )
	{
	    err = LDAP_DECODING_ERROR;
		goto error;
	} 

    /* get the connectionDN */
    err = parseLDAPString(responseBer, &data->sourceModule, 0);
    if (err != LDAP_SUCCESS)
        goto error;

    return err;

error:
    if (NULL != data)
	{
	    ldap_event_free(data);
	    *eventData = NULL;
	}

	return err;
}

/* ---------------------------------------------------------------------------
    parseNetAddrEventData

    This function completes the parsing of a berelement that contains a 
	response for an Entry event.

    ld           (IN) An LDAP session handle, as obtained from a call to
				  ldap_init().

    responseBer  (IN) the ber element containing the response to be parsed.

    eventData   (OUT) a pointer to the EVT_NetAddress structure containing the 
	                  parsed data.

	    NetworkAddress ::= [APPLICATION 8]SEQUENCE {
		    type	INTEGER,
		    data	OCTET STRING }
   ---------------------------------------------------------------------------*/
int parseNetAddrEventData(
    LDAP		*ld,
    BerElement  *responseBer, 
    void		**eventData)
{
    	ber_tag_t       tag = 0;
    	ber_len_t       size = 0;
	EVT_NetAddress	*data = NULL;
    	int             tempType = 0;
	int     err = LDAP_SUCCESS;	

    /* get the type */
	if (LBER_ERROR == ber_scanf(responseBer, "i", &tempType))
	{
	    err = LDAP_DECODING_ERROR;
	    goto error;
	} 

    /* get the data and its size */
    tag = ber_peek_tag(responseBer, &size);
	if (LBER_OCTETSTRING == tag)
    {
        if (0 != size) 
	    {   /* allocate memory for the event data plus enough for the 
             * address data on the end plus the null terminator.
             */
             size++;
            *eventData = data = 
              (EVT_NetAddress*)allocEventData(sizeof(EVT_NetAddress)+size,
                                              EVT_TAG_NETWORK_ADDRESS);
		    if (NULL == data)
		    {
		        err = LDAP_NO_MEMORY;
			    goto error;
		    }
            data->type = tempType;
		    if (LBER_ERROR == ber_scanf(responseBer, "s", 
		                                data->data, &size))
		    {
		        err = LDAP_DECODING_ERROR;
			    goto error;
		    }
            else
            {
                data->length = (int)size;
            }
	    }
	    else
	    {
            //allocate memory for the event data.
            *eventData = data = (EVT_NetAddress*)
            allocEventData(sizeof(EVT_NetAddress), EVT_TAG_NETWORK_ADDRESS);
		    if (NULL == data)
		    {
		        err = LDAP_NO_MEMORY;
			    goto error;
		    }
            data->type = tempType;
		    data->length = 0;
		    data->data[0] = 0;
	    }
    }
    else
	{
	    err = LDAP_DECODING_ERROR;
		goto error;
	} 

    return err;

error:
    if (NULL != data)
	{
	   	ldap_event_free(data);
	   	*eventData = NULL;
	}

	return err;
}
/* ---------------------------------------------------------------------------
    parseModuleStateEventData

    This function completes the parsing of a berelement that contains a 
	response for module state event.

    responseBer  (IN) the ber element containing the response to be parsed.

    eventData   (OUT) a pointer to the EVT_ModuleState structure containing the 
	                  parsed data.

	ModuleState ::= [APPLICATION 7]SEQUENCE {
		connectionDN	LDAPDN,
		flags          INTEGER,
		name           LDAPString (DSE_MAX_MODULE_NAME),
		description	   LDAPString (DSE_MAX_MODULE_DESCR),
		source         LDAPString }
   ---------------------------------------------------------------------------*/
int parseModuleStateEventData(
    LDAP		*ld,
    BerElement  *responseBer, 
    void		**eventData)
{
    ber_tag_t       tag = 0 ;
    ber_len_t       size = 0;
	EVT_ModuleState	*data = NULL;
	int     err = LDAP_SUCCESS;	

    //allocate memory for the event data.
    *eventData = data = (EVT_ModuleState*)
      allocEventData(sizeof(EVT_ModuleState), EVT_TAG_MODULE_STATE_DATA);

    if (NULL == data)
    {
		err = LDAP_NO_MEMORY;
		goto error;
    }

	/* initialize data */
    data->connectionDN = NULL;
    data->name[0] = '\0';
    data->description[0] = '\0';
    data->sourceModule = NULL;

    /* get the connectionDN */
    err = parseLDAPString(responseBer, &data->connectionDN, 0);
    if (err != LDAP_SUCCESS)
        goto error;

    /* get flags, module, source */
	if (LBER_ERROR == ber_scanf(responseBer, "i", &data->flags))
	{
	    err = LDAP_DECODING_ERROR;
		goto error;
	} 

    /* get name */
    tag = ber_peek_tag(responseBer, &size);
	if ( (LBER_OCTETSTRING == tag) && 
	     (0 != size) && (size < EVT_MAX_MODULE_NAME) )
	{
        size = sizeof(data->name);
		if (LBER_ERROR == ber_scanf(responseBer, "s", 
		                            data->name, &size))
		{
		    err = LDAP_DECODING_ERROR;
			goto error;
        }
	}
    else
	{
	    err = LDAP_DECODING_ERROR;
		goto error;
	} 

    /* get description */
    tag = ber_peek_tag(responseBer, &size);
	if ( (LBER_OCTETSTRING == tag) && 
	     (size < EVT_MAX_MODULE_DESCR) )
	{
        size = sizeof(data->description);
		if (LBER_ERROR == ber_scanf(responseBer, "s", 
		                            data->description, &size))
		{
		    err = LDAP_DECODING_ERROR;
			goto error;
        }
	}
    else
	{
	    err = LDAP_DECODING_ERROR;
		goto error;
	} 

    /* get the sourceModule */
    err = parseLDAPString(responseBer, &data->sourceModule, 0);
    if (err != LDAP_SUCCESS)
        goto error;

    return err;

error:
    if (NULL != data)
	{
	   ldap_event_free(data);
	   *eventData = NULL;
	}

	return err;
}
/* ---------------------------------------------------------------------------
    parseSEVEventData

    This function completes the parsing of a berelement that contains a 
	response for an SEV (security equivalence vector) event.

    responseBer  (IN) the ber element containing the response to be parsed.

    eventData   (OUT) a pointer to the EVT_SEVInfo structure containing the 
	                  parsed data.

	DSESEVInfo ::= [APPLICATION 6]SEQUENCE {
		entry		LDAPDN,
		retryCount	INTEGER,
		valueDN		LDAPDN,
		referral	ReferralList }

   referralList := SEQUENCE {
			count		INTEGER,
			referrals	SEQUENCE of ReferralAddress }

	ReferralAddress := SEQUENCE {
		addressType		INTEGER,
		address			OCTET STRING}


   ---------------------------------------------------------------------------*/
int parseSEVEventData(
    LDAP		*ld,
    BerElement  *responseBer, 
    void		**eventData)
{
	EVT_SEVInfo	 *data = NULL;
	int     err = LDAP_SUCCESS;	

    //allocate memory for the event data.
    *eventData = data = (EVT_SEVInfo*)
      allocEventData(sizeof(EVT_SEVInfo), EVT_TAG_DSESEV_INFO);

    if (NULL == data)
    {
		err = LDAP_NO_MEMORY;
		goto error;
    }

	/* initialize data */
	data->entryDN = NULL;
	data->valueDN = NULL;
    data->referrals = NULL;

    /* get the entryDN */
    err = parseLDAPString(responseBer, &data->entryDN, 0);
    if (err != LDAP_SUCCESS)
        goto error;

    /* get retryCount */
	if (LBER_ERROR == ber_scanf(responseBer, "i", &data->retryCount))
	{
	    err = LDAP_DECODING_ERROR;
		goto error;
	} 

    /* get the valueDN */
    err = parseLDAPString(responseBer, &data->valueDN, 0);
    if (err != LDAP_SUCCESS)
        goto error;

    /* get the referral */
	/* get the localReferral */
    err = parseReferralList(responseBer, &data->referralCount, 
                                     &data->referrals);
    if (err != LDAP_SUCCESS)
        goto error;

    return err;

error:
    if (NULL != data)
	{
	   	ldap_event_free(data);
	   	*eventData = NULL;
	}

	return err;
}
/* ---------------------------------------------------------------------------
    parseBinderyEventData

    This function completes the parsing of a berelement that contains a 
	response for an Entry event.

    responseBer  (IN) the ber element containing the response to be parsed.

    eventData   (OUT) a pointer to the EVT_BinderyObjectInfo structure 
                      containing the parsed data.

	BinderyObjectInfo [APPLICATION 5]::= SEQUENCE {
		entry		LDAPDN,
		type		INTEGER,
		emuObjFlags	INTEGER,
		security	INTEGER,
		name		LDAPString (48) }
   ---------------------------------------------------------------------------*/
int parseBinderyEventData(
    LDAP		*ld,
    BerElement  *responseBer, 
    void		**eventData)
{
	EVT_BinderyObjectInfo  *data = NULL;
	int     err = LDAP_SUCCESS;	

    //allocate memory for the event data.
    *eventData = data = (EVT_BinderyObjectInfo*)
      allocEventData(sizeof(EVT_BinderyObjectInfo), EVT_TAG_BINDERY_EVENT_DATA);

    if (NULL == data)
    {
		err = LDAP_NO_MEMORY;
		goto error;
    }

	/* initialize data */
	data->entryDN = NULL;
    data->name = NULL;

    /* get the entryDN */
    err = parseLDAPString(responseBer, &data->entryDN, 0);
    if (err != LDAP_SUCCESS)
        goto error;

    /* get type, emuObjFlags, and security */
	if (LBER_ERROR == ber_scanf(responseBer, "iii", 
	                            &data->type,
	                            &data->emuObjFlags,
	                            &data->security))
	{
	    err = LDAP_DECODING_ERROR;
		goto error;
	} 

    /* get the name */
    err = parseLDAPString(responseBer, &data->name, 0);
    if (err != LDAP_SUCCESS)
        goto error;

    return err;

error:
    if (NULL != data)
	{
	   	ldap_event_free(data);
	   	*eventData = NULL;
	}

	return err;
}

/* ---------------------------------------------------------------------------
    parseSkulkEventData

    This function completes the parsing of a berelement that contains a 
	response for an Entry event.

    responseBer  (IN) the ber element containing the response to be parsed.

    eventData   (OUT) a pointer to the EVT_SkulkData structure containing the 
	                  parsed data.
   ---------------------------------------------------------------------------*/
int parseSkulkEventData(
    LDAP		*ld,
    BerElement  *responseBer, 
    void		**eventData)
{
    	*eventData = NULL;
	int     err = LDAP_SUCCESS;

	return err = LDAP_SUCCESS;
}
/* ---------------------------------------------------------------------------
    parseGeneralEventData

    This function completes the parsing of a berelement that contains a 
	response for an Entry event.

    responseBer  (IN) the ber element containing the response to be parsed.

    eventData   (OUT) a pointer to the EVT_EventData structure containing the 
	                  parsed data.

	GeneralEventData ::= [APPLICATION 3]SEQUENCE {
		    dsTime		    [1] INTEGER,
		    milliseconds	[2] INTEGER,
		    verb		    [3] INTEGER,
            currentProcess  [4] INTEGER,
            perpetratorDN   [5] LDAPDN,
            integerValues   [7] SEQUENCE OF INTEGER,
            stringValues    [8] SEQUENCE OF LDAPSTRING}

   ---------------------------------------------------------------------------*/
int parseGeneralEventData(
    LDAP		*ld,
    BerElement  *responseBer, 
    void		**eventData)
{
    ber_tag_t       tag = 0;
    ber_len_t      size = 0;
	EVT_EventData  *data = NULL;
    int            index = 0;
	int     err = LDAP_SUCCESS;	

    //allocate memory for the event data.
    *eventData = data = (EVT_EventData*)
      allocEventData(sizeof(EVT_EventData), EVT_TAG_GENERAL_EVENT_DATA);

    if (NULL == data)
    {
		err = LDAP_NO_MEMORY;
		goto error;
    }

	/* initialize data */
	data->perpetratorDN = NULL;

    data->intValues[0] = 0;
    data->intValues[1] = 0;
    data->intValues[2] = 0;
    data->intValues[3] = 0;

    data->strValues[0] = NULL;
    data->strValues[1] = NULL;
    data->strValues[2] = NULL;
    data->strValues[3] = NULL;

    /* get dsTime, milliseconds, verb, currentProcess, connID */
	if (LBER_ERROR == ber_scanf(responseBer, "iiii", 
	                            &data->dstime,
	                            &data->milliseconds,
	                            &data->verb,
	                            &data->curProcess))
	{
	    err = LDAP_DECODING_ERROR;
		goto error;
	} 

    /* get the perpetratorDN */
    err = parseLDAPString(responseBer, &data->perpetratorDN,
                                   EVT_TAG_GEN_PERP);
    if (err != LDAP_SUCCESS)
        goto error;

    /* get integers if present */
    tag = ber_peek_tag(responseBer, &size);
	if (EVT_TAG_GEN_INTEGERS == tag) /*we have integers*/
	{
        ber_skip_tag(responseBer, &size); /* skip the sequence */
        index = 0;
	    while(LBER_INTEGER == ber_peek_tag(responseBer, &size))
		{
            if (index > 3)
		    {
		        err = LDAP_DECODING_ERROR;
			    goto error;
		    } 
            if (LBER_ERROR == ber_scanf(responseBer, "i", 
                                        &data->intValues[index]))
		    {
		        err = LDAP_DECODING_ERROR;
			    goto error;
		    } 
            index++;
        }
    }

    /* get strings if present */
    tag = ber_peek_tag(responseBer, &size);
	if (EVT_TAG_GEN_STRINGS == tag) /*we have strings*/
	{
        ber_skip_tag(responseBer, &size); /* skip the sequence */
        index = 0;
	    while(LBER_OCTETSTRING == ber_peek_tag(responseBer, &size))
		{
	        if (0 != size)
	        {
	            size++; /*for the null character*/
	            data->strValues[index] = (char*)malloc(size);
		        if (NULL == data->strValues[index])
		        {
		            err = LDAP_NO_MEMORY;
			        goto error;
		        }
		        if (LBER_ERROR == ber_scanf(responseBer, "s", 
		                                    data->strValues[index], &size))
		        {
		            err = LDAP_DECODING_ERROR;
			        goto error;
		        } 
	        }
	        else
	        {
                ber_skip_tag(responseBer, &size);
	        }
            index++;
            if (index > 3)
		    {
		        err = LDAP_DECODING_ERROR;
			    goto error;
		    } 
        }
    }

    return err;

error:
    if (NULL != data)
	{
	   	ldap_event_free(data);
	   	*eventData = NULL;
	}

	return err;
}

/* ---------------------------------------------------------------------------
    parseValueEventData

    This function completes the parsing of a berelement that contains a 
	response for an value event.

    ld           (IN) An LDAP session handle, as obtained from a call to
				  ldap_init().

    responseBer  (IN) the ber element containing the response to be parsed.

    eventData   (OUT) a pointer to the EVT_ValueInfo structure containing the 
	                  parsed data.

	ValueEventData ::= [APPLICATION 2]SEQUENCE {
             perpetratorDN   LDAPDN,
		     entry		     LDAPDN,
		     attribute       OCTET STRING,
             syntax          LDAPOID,
             class           OCTET STRING,
             timeStamp       DSETimestamp,
             data            OCTET STRING
             verb            INTEGER }

  ---------------------------------------------------------------------------*/
int parseValueEventData(
    LDAP		*ld,
    BerElement  *responseBer, 
    void		**eventData)
{
    ber_tag_t       tag = 0 ;
    ber_len_t       size = 0;
	EVT_ValueInfo	*data = NULL;
	int     err = LDAP_SUCCESS;	

    //allocate memory for the event data.
    *eventData = data = (EVT_ValueInfo*)
      allocEventData(sizeof(EVT_ValueInfo), EVT_TAG_VALUE_EVENT_DATA);

    if (NULL == data)
    {
		err = LDAP_NO_MEMORY;
		goto error;
    }

	/* initialize data */
    data->perpetratorDN = NULL;
	data->entryDN = NULL;
	data->attributeName = NULL;
    data->syntaxOID = NULL;
	data->className = NULL;
	data->value = NULL;
    data->size = 0;

    /* get the perpetratorDN */
    err = parseLDAPString(responseBer, &data->perpetratorDN, 0);
    if (err != LDAP_SUCCESS)
        goto error;

    /* get the entryDN */
    err = parseLDAPString(responseBer, &data->entryDN, 0);
    if (err != LDAP_SUCCESS)
        goto error;

    /* get the attribute name */
    err = parseLDAPString(responseBer, &data->attributeName, 0);
    if (err != LDAP_SUCCESS)
        goto error;

    /* get the syntax oid */
    err = parseLDAPString(responseBer, &data->syntaxOID, 0);
    if (err != LDAP_SUCCESS)
        goto error;

    /* get the className */
    err = parseLDAPString(responseBer, &data->className, 0);
    if (err != LDAP_SUCCESS)
        goto error;

    /* get the timestamp */
	if (LBER_ERROR == ber_scanf(responseBer, "{iii}", 
	                             &data->timeStamp.seconds,
	                             &data->timeStamp.replicaNumber,
	                             &data->timeStamp.event))
	{
	    err = LDAP_DECODING_ERROR;
		goto error;
	} 

    /* get the value and its size if present */
    tag = ber_peek_tag(responseBer, &size);
	if ( (LBER_OCTETSTRING == tag) && 
	     (0 != size) )
	{
	    size++; /*for the null character*/
	    data->value = (char*)malloc(size);
		if (NULL == data->value)
		{
		    err = LDAP_NO_MEMORY;
			goto error;
		}
		if (LBER_ERROR == ber_scanf(responseBer, "s", 
		                            data->value, &size))
		{
		    err = LDAP_DECODING_ERROR;
			goto error;
		}
        else
           data->size = (int)size;
	}

    /* get the verb value if it is present.
     * Falcon beta did not have the verb. After falcon FCS the 
     * ber_peek_tag call can go away.
     */
    tag = ber_peek_tag(responseBer, &size);
    if (LBER_INTEGER == tag)
    {
    	if (LBER_ERROR == ber_scanf(responseBer, "i", &data->verb))
	    {
	        err = LDAP_DECODING_ERROR;
		    goto error;
	    } 
    }
    else
        data->verb = 0;

    return err;

error:
    if (NULL != data)
	{
	   ldap_event_free(data);
	   *eventData = NULL;
	}

	return err;
}

/* ---------------------------------------------------------------------------
    parseEntryEventData

    This function completes the parsing of a berelement that contains a 
	response for an Entry event. The initial sequence tag has already been 
	parsed by the calling function.

    ld           (IN) An LDAP session handle, as obtained from a call to
				  ldap_init().

    responseBer  (IN) the ber element containing the response to be parsed.

    eventData   (OUT) a pointer to the EVT_EntryInfo structure containing the 
	                  parsed data.

		EntryEventData ::= SEQUENCE {
            perpetratorDN   LDAPDN,
			entry		    LDAPDN,
			class		    LDAPOID,
			creationTime    DSETimestamp,
			verb            INTEGER,
			flags           INTEGER,
			newDN		    LDAPOID OPTIONAL }

		DSETimeStamp ::= SEQUENCE {
			seconds		    INTEGER,
			replicaNumber	INTEGER,
			event		    INTEGER }

  ---------------------------------------------------------------------------*/
int parseEntryEventData(
    LDAP		*ld,
    BerElement  *responseBer, 
    void		**eventData)
{
    ber_tag_t       tag = 0;
    ber_len_t       size = 0;
	EVT_EntryInfo	*data = NULL;
	int     err = LDAP_SUCCESS;	

    //allocate memory for the event data.
    *eventData = data = (EVT_EntryInfo*)
      allocEventData(sizeof(EVT_EntryInfo), EVT_TAG_ENTRY_EVENT_DATA);

    if (NULL == data)
    {
		err = LDAP_NO_MEMORY;
		goto error;
    }

	/* initialize data */
    data->perpetratorDN = NULL;
	data->entryDN = NULL;
	data->className = NULL;
	data->newDN = NULL;

    /* get the perpetratorDN */
    err = parseLDAPString(responseBer, &data->perpetratorDN, 0);
    if (err != LDAP_SUCCESS)
        goto error;

    /* get the entryDN */
    err = parseLDAPString(responseBer, &data->entryDN, 0);
    if (err != LDAP_SUCCESS)
        goto error;

    /* get the className */
    err = parseLDAPString(responseBer, &data->className, 0);
    if (err != LDAP_SUCCESS)
        goto error;

    /* get the integer data */
	if (LBER_ERROR == ber_scanf(responseBer, "{iii}ii", 
	                             &data->creationTime.seconds,
	                             &data->creationTime.replicaNumber,
	                             &data->creationTime.event,
	                             &data->verb,
	                             &data->flags))
	{
	    err = LDAP_DECODING_ERROR;
		goto error;
	} 

    /* get the newDN if present */
    tag = ber_peek_tag(responseBer, &size);
	if ( (LBER_OCTETSTRING == tag) && 
	     (0 != size) )
	{
	    size++; /*for the null character*/
	    data->newDN = (char*)malloc(size);
		if (NULL == data->newDN)
		{
		    err = LDAP_NO_MEMORY;
			goto error;
		}
		if (LBER_ERROR == ber_scanf(responseBer, "s", 
		                            data->newDN, &size))
		{
		    err = LDAP_DECODING_ERROR;
			goto error;
		} 
	}

    return err;

error:
    if (NULL != data)
	{
	    ldap_event_free(data);
	    *eventData = NULL;
	}

	return err;
}

/* ---------------------------------------------------------------------------
    parseStatusLogEventData

    This function completes the parsing of a berelement that contains a 
	response for an Entry event. The initial sequence tag has already been 
	parsed by the calling function.

    ld           (IN) An LDAP session handle, as obtained from a call to
				  ldap_init().

    responseBer  (IN) the ber element containing the response to be parsed.

    eventData   (OUT) a pointer to the EVT_EntryInfo structure containing the 
	                  parsed data.
                      

	StatusLogEventData ::= [APPLICATION 15]SEQUENCE {
 		    logTime        INTEGER,
	        moduleName     LDAPSTRING,
		    messageLevel   LDAPSTRING,
			messageEvent   LDAPSTRING,
			messageCode    LDAPSTRING,
			messageDesc    LDAPSTRING,
			statusDoc      LDAPSTRING}

  ---------------------------------------------------------------------------*/
int parseStatusLogEventData(
    LDAP		*ld,
    BerElement  *responseBer, 
    void		**eventData)
{
	EVT_StatusLogInfo  *data = NULL;
	int     err = LDAP_SUCCESS;	

    //allocate memory for the event data.
    *eventData = data = (EVT_StatusLogInfo*)
      allocEventData(sizeof(EVT_StatusLogInfo), EVT_TAG_STATUS_LOG);

    if (NULL == data)
    {
		err = LDAP_NO_MEMORY;
		goto error;
    }

	/* initialize data */
	data->logTime = 0;
    data->moduleName = NULL;
    data->messageLevel = NULL;
    data->messageEvent = NULL;
    data->messageCode = NULL;
    data->messageDesc = NULL;
    data->statusDoc = NULL;
    
    if (LBER_ERROR == ber_scanf(responseBer, "i", &data->logTime))
    {
        err = LDAP_DECODING_ERROR;
        goto error;
    }

    /* get the moduleName */
    err = parseLDAPString(responseBer, &data->moduleName, 0);
    if (err != LDAP_SUCCESS)
        goto error;

    /* get the messageLevel */
    err = parseLDAPString(responseBer, &data->messageLevel, 0);
    if (err != LDAP_SUCCESS)
        goto error;

    /* get the messageEvent */
    err = parseLDAPString(responseBer, &data->messageEvent, 0);
    if (err != LDAP_SUCCESS)
        goto error;

    /* get the messageCode */
    err = parseLDAPString(responseBer, &data->messageCode, 0);
    if (err != LDAP_SUCCESS)
        goto error;

    /* get the messageDesc */
    err = parseLDAPString(responseBer, &data->messageDesc, 0);
    if (err != LDAP_SUCCESS)
        goto error;

    /* get the statusDoc */
    err = parseLDAPString(responseBer, &data->statusDoc, 0);
    if (err != LDAP_SUCCESS)
        goto error;

    return err;

error:
    if (NULL != data)
	{
	    ldap_event_free(data);
	    *eventData = NULL;
	}

	return err;
}

/* ---------------------------------------------------------------------------
    ldap_parse_ds_event

    After a successful call to ldap_monitor event(), server responses to the 
    EventMonitorRequest are retrieved by calling the ldap_result function. If 
    the return value of ldap_result is equal to LDAP_RES_INTERMEDIATE, the 
    result is parsed by calling ldap_parse_ds_event. 

    ld       (IN) An LDAP session handle, as obtained from a call to
				  ldap_init().

    eventMessage  (IN) a pointer to the LDAPMessage returned by ldap_result().

    eventType	(OUT) a pointer to the int to be set to the event type.

    eventResult  (OUT) a pointer to the int to be set to event's result.

    eventData	(OUT) a pointer to a pointer to a structure containing the data
                      returned by this particular event. The structure is 
                      allocated by this function. The type of the structure is
                      determined by the eventType. If the event does not have 
                      associated data the pointer will be set to NULL. 
                      When the application no longer needs the data it should 
                      free the data by calling the ldap_event_free() function.

    The response field will contain the LBER encoding of the 
    EventMonitorResponse whose ASN.1 definition is given below.

        EventMonitorResponse ::= SEQUENCE {
             EventType            INTEGER,
             ResultCode           INTEGER,
             EventData            OCTET STRING OPTIONAL}


  ---------------------------------------------------------------------------*/
LDAP_F( int )
ldap_parse_ds_event(
    LDAP		*ld,
    LDAPMessage	*eventMessage,
    int		    *eventType,
    int		    *eventResult,
    void		**eventData,
    int         freeIt)
{
	BerElement       *responseBer = NULL;
    	struct berval    *extRespData = NULL;
	ber_len_t        dataLength = 0;
	ber_tag_t        dataTag = 0;
	char             *retOID = NULL;
	int     	 err = LDAP_SUCCESS;
   /* result code and responseCode must be non-null */
    if ( (NULL == eventType) || (NULL == eventResult) )
	   return (err = LDAP_PARAM_ERROR);

    if ( NULL != eventData)
	  *eventData = NULL;

    err = ldap_parse_intermediate(ld, eventMessage, 
                                           &retOID, &extRespData, 
                                           NULL, freeIt);

    if (err != LDAP_SUCCESS)
	    goto cleanup;

    /* verifiy the OID of the result */
    if ( 0 != strcmp(retOID, NLDAP_EVENT_NOTIFICATION) )
	{
	    err = LDAP_DECODING_ERROR;
		goto cleanup;
	}

    /* Create a BerElement from the berval returned  */
    responseBer = ber_init(extRespData);
    if (responseBer == NULL)
    {
        err = LDAP_NO_MEMORY;
        goto cleanup;
    }
    
    /* Extract the response code. */
    if (LBER_ERROR == 
         ber_scanf(responseBer, "{ii" /*}*/, eventType, eventResult))
    {
    	err = LDAP_DECODING_ERROR;
        goto cleanup;
    }

    if (NULL != eventData)
    {
        dataTag = ber_skip_tag(responseBer, &dataLength);
		switch (dataTag)
		{
		case EVT_TAG_ENTRY_EVENT_DATA:
		    err = parseEntryEventData(ld, responseBer, eventData);
		    break; 
		case EVT_TAG_VALUE_EVENT_DATA:
		    err = parseValueEventData(ld, responseBer, eventData);
		    break; 
		case EVT_TAG_GENERAL_EVENT_DATA:
		    err = parseGeneralEventData(ld, responseBer, eventData);
		    break; 
		case EVT_TAG_SKULK_DATA:
		    err = parseSkulkEventData(ld, responseBer, eventData);
		    break; 
		case EVT_TAG_BINDERY_EVENT_DATA:
		    err = parseBinderyEventData(ld, responseBer, eventData);
		    break; 
		case EVT_TAG_DSESEV_INFO:
		    err = parseSEVEventData(ld, responseBer, eventData);
		    break; 
		case EVT_TAG_MODULE_STATE_DATA:
		    err = parseModuleStateEventData(ld, responseBer, eventData);
		    break; 
		case EVT_TAG_NETWORK_ADDRESS:
		    err = parseNetAddrEventData(ld, responseBer, eventData);
		    break; 
		case EVT_TAG_CONNECTION_STATE:
		    err = parseConnectionStateData(ld, responseBer, eventData);
		    break; 
		case EVT_TAG_CHANGE_SERVER_ADDRESS:
        	    err = parseChangeServerAddressData(ld, responseBer, eventData);
        	    break;
		case EVT_TAG_CHANGE_CONFIG_PARAM:
        	    err = parseChangeConfigParamData(ld, responseBer, eventData);
        	    break;
        	case EVT_TAG_STATUS_LOG:
            	    err = parseStatusLogEventData(ld, responseBer, eventData);
            	    break;
        	case EVT_TAG_DEBUG_EVENT_DATA:
            	    err = parseDebugEventData(ld, responseBer, eventData);
            	    break;
        	case EVT_TAG_NO_DATA:
		default:
		    *eventData = NULL;
		    break;
		}
	}

cleanup:
	/* Free memory that we allocated when building the request ber and berval */
	if (extRespData)
		ber_bvfree(extRespData);

	if (responseBer)
		ber_free(responseBer, 1); 

	if (retOID)
		ldap_memfree(retOID);

	/* if we failed then free the eventData */
    if ( (err != LDAP_SUCCESS) &&
         (NULL != eventData) )
    {
		if (NULL != *eventData)
			ldap_event_free(*eventData);
		*eventData = NULL;
	}

    return err;
}

/* ---------------------------------------------------------------------------
    freeReferralList

    This function frees memory allocated for a relical list.

    count      (IN) the number of referrals in the referral array

	replicas   (IN)	the array of relpicas
  ---------------------------------------------------------------------------*/
void freeReferralList(
    int                   count,
    EVT_ReferralAddress*  replicas)
{
    int i = 0;
    
    if (NULL != replicas)
    {
        for (i=1; i<count; i++)
        {
            if (NULL != replicas[i].address)
                free(replicas[i].address);
        }
        free(replicas);
    }
}
/* ---------------------------------------------------------------------------
    ldap_event_free

    This function frees memory allocated by the 
    ldap_parse_ds_event_intermediate or the ldap_parse_ds_event_extended
	function.

    ld       (IN) An LDAP session handle, as obtained from a call to
				  ldap_init().

	eventData  (IN)	A pointer to the event data to be freed.
  ---------------------------------------------------------------------------*/
LDAP_F( void )
ldap_event_free(
    void  *eventData)
{
    ber_tag_t   tag = 0;
    int i = 0;

	if (eventData != NULL)
	{
	    tag = *(((ber_tag_t*)eventData)-1);
        switch (tag)
		{
		case EVT_TAG_ENTRY_EVENT_DATA:
		    if (NULL != ((EVT_EntryInfo*)eventData)->perpetratorDN)
			    free(((EVT_EntryInfo*)eventData)->perpetratorDN);
		    if (NULL != ((EVT_EntryInfo*)eventData)->entryDN)
			    free(((EVT_EntryInfo*)eventData)->entryDN);
		    if (NULL != ((EVT_EntryInfo*)eventData)->className)
			    free(((EVT_EntryInfo*)eventData)->className);
		    if (NULL != ((EVT_EntryInfo*)eventData)->newDN)
			    free(((EVT_EntryInfo*)eventData)->newDN);
		    break; 
            
		case EVT_TAG_VALUE_EVENT_DATA:
		    if (NULL != ((EVT_ValueInfo*)eventData)->perpetratorDN)
			    free(((EVT_ValueInfo*)eventData)->perpetratorDN);
		    if (NULL != ((EVT_ValueInfo*)eventData)->entryDN)
			    free(((EVT_ValueInfo*)eventData)->entryDN);
		    if (NULL != ((EVT_ValueInfo*)eventData)->attributeName)
			    free(((EVT_ValueInfo*)eventData)->attributeName);
		    if (NULL != ((EVT_ValueInfo*)eventData)->syntaxOID)
			    free(((EVT_ValueInfo*)eventData)->syntaxOID);
		    if (NULL != ((EVT_ValueInfo*)eventData)->className)
			    free(((EVT_ValueInfo*)eventData)->className);
		    if (NULL != ((EVT_ValueInfo*)eventData)->value)
			    free(((EVT_ValueInfo*)eventData)->value);
		    break; 
            
		case EVT_TAG_GENERAL_EVENT_DATA:
		    if (NULL != ((EVT_EventData*)eventData)->perpetratorDN)
			    free(((EVT_EventData*)eventData)->perpetratorDN);
		    if (NULL != ((EVT_EventData*)eventData)->strValues[0])
			    free(((EVT_EventData*)eventData)->strValues[0]);
		    if (NULL != ((EVT_EventData*)eventData)->strValues[1])
			    free(((EVT_EventData*)eventData)->strValues[1]);
		    if (NULL != ((EVT_EventData*)eventData)->strValues[2])
			    free(((EVT_EventData*)eventData)->strValues[2]);
		    if (NULL != ((EVT_EventData*)eventData)->strValues[3])
			    free(((EVT_EventData*)eventData)->strValues[3]);
		    break; 
            
		case EVT_TAG_SKULK_DATA:
		    break; 
            
		case EVT_TAG_BINDERY_EVENT_DATA:
		    if (NULL != ((EVT_BinderyObjectInfo*)eventData)->entryDN)
			    free(((EVT_BinderyObjectInfo*)eventData)->entryDN);
		    if (NULL != ((EVT_BinderyObjectInfo*)eventData)->name)
			    free(((EVT_BinderyObjectInfo*)eventData)->name);
		    break; 
            
		case EVT_TAG_DSESEV_INFO:
		    if (NULL != ((EVT_SEVInfo*)eventData)->entryDN)
			    free(((EVT_SEVInfo*)eventData)->entryDN);
		    if (NULL != ((EVT_SEVInfo*)eventData)->valueDN)
			    free(((EVT_SEVInfo*)eventData)->valueDN);
            freeReferralList(((EVT_SEVInfo*)eventData)->referralCount,
                             ((EVT_SEVInfo*)eventData)->referrals);
		    break; 
            
		case EVT_TAG_MODULE_STATE_DATA:
		    if (NULL != ((EVT_ModuleState*)eventData)->connectionDN)
			    free(((EVT_ModuleState*)eventData)->connectionDN);
		    if (NULL != ((EVT_ModuleState*)eventData)->sourceModule)
			    free(((EVT_ModuleState*)eventData)->sourceModule);
		    break; 
            
		case EVT_TAG_NETWORK_ADDRESS:
		    break; 
            
		case EVT_TAG_CONNECTION_STATE:
		    if (NULL != ((EVT_ChangeConnState*)eventData)->connectionDN)
			    free(((EVT_ChangeConnState*)eventData)->connectionDN);
		    if (NULL != ((EVT_ChangeConnState*)eventData)->sourceModule)
			    free(((EVT_ChangeConnState*)eventData)->sourceModule);
		    break; 
            
        case EVT_TAG_CHANGE_SERVER_ADDRESS:
        	if (NULL != ((EVT_ChangeServerAddr*)eventData)->addr)
            	free(((EVT_ChangeServerAddr*)eventData)->addr);
        	if (NULL != ((EVT_ChangeServerAddr*)eventData)->pstkname)
            	free(((EVT_ChangeServerAddr*)eventData)->pstkname);
        	if (NULL != ((EVT_ChangeServerAddr*)eventData)->sourceModule)
            	free(((EVT_ChangeServerAddr*)eventData)->sourceModule);
        	break;
            
        case EVT_TAG_CHANGE_CONFIG_PARAM:
        	if (NULL != ((EVT_ChangeConfigParm*)eventData)->name)
            	free(((EVT_ChangeConfigParm*)eventData)->name);
            if ( (EVT_CFG_TYPE_BINARY == 
                    ((EVT_ChangeConfigParm*)eventData)->type) &&
                 (NULL != ((EVT_ChangeConfigParm*)eventData)->value.binary.data) )
                free(((EVT_ChangeConfigParm*)eventData)->value.binary.data);
            if ( (EVT_CFG_TYPE_STRING == 
                    ((EVT_ChangeConfigParm*)eventData)->type) &&
                 (NULL != ((EVT_ChangeConfigParm*)eventData)->value.utf8Str) )
                free(((EVT_ChangeConfigParm*)eventData)->value.utf8Str);
            break;
            
        case EVT_TAG_STATUS_LOG:
            if (NULL != ((EVT_StatusLogInfo*)eventData)->moduleName)
                free(((EVT_StatusLogInfo*)eventData)->moduleName);    
            if (NULL != ((EVT_StatusLogInfo*)eventData)->messageLevel)
                free(((EVT_StatusLogInfo*)eventData)->messageLevel);    
            if (NULL != ((EVT_StatusLogInfo*)eventData)->messageEvent)
                free(((EVT_StatusLogInfo*)eventData)->messageEvent);    
            if (NULL != ((EVT_StatusLogInfo*)eventData)->messageCode)
                free(((EVT_StatusLogInfo*)eventData)->messageCode);    
            if (NULL != ((EVT_StatusLogInfo*)eventData)->messageDesc)
                free(((EVT_StatusLogInfo*)eventData)->messageDesc);    
            if (NULL != ((EVT_StatusLogInfo*)eventData)->statusDoc)
                free(((EVT_StatusLogInfo*)eventData)->statusDoc);
            break;
            
        case EVT_TAG_DEBUG_EVENT_DATA:
        {
            DB_Parameter*  param;
            EVT_DebugInfo   *dbInfo = (EVT_DebugInfo*)eventData;
            
            if (NULL != dbInfo->formatString)
                free(dbInfo->formatString);
					 
				if (NULL != dbInfo->perpetratorDN)
					free(dbInfo->perpetratorDN);
					
            for (i = 0; i < dbInfo->paramCount; i++)
            {
                param = &dbInfo->parameters[i];
                switch (param->type)
                {
                case DB_PARAM_TYPE_STRING:
                    if (param->value.utf8Str != NULL)
                        free(param->value.utf8Str);
                    break;
                case DB_PARAM_TYPE_BINARY:
                    if (param->value.binary.data != NULL)
                        free(param->value.binary.data);
                    break;
                case DB_PARAM_TYPE_ADDRESS:
                    if (param->value.netAddress.data != NULL)
                        free(param->value.netAddress.data);
                    break;
                case DB_PARAM_TYPE_TIMEVECTOR:
                    if (param->value.timeStampVector.timeStamps != NULL)
                        free(param->value.timeStampVector.timeStamps);
                    break;
                case DB_PARAM_TYPE_ENTRYID:
                case DB_PARAM_TYPE_INTEGER:
                case DB_PARAM_TYPE_TIMESTAMP:
                default:
                    break;
                }
            }
            if (dbInfo->parameters != NULL)
                free(dbInfo->parameters);
            break;
        }

		default:
		    break;
		}
	}
	
	free(((ber_tag_t*)eventData)-1);
}



