/*
 * $Novell: lburp.c,v 1.0 2003/08/04 17:41:39 $
 ******************************************************************************
 * Copyright (c) 2006 Novell, Inc.
 * All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; version 2.1 of the license.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, contact Novell, Inc.
 *
 * To contact Novell about this file by physical or electronic mail,
 * you may find current contact information at www.novell.com
 *
 ****************************************************************************
 *
 * Copyright (C) 2003 Novell, Inc. All Rights Reserved.
 *
 * THIS WORK IS SUBJECT TO U.S. AND INTERNATIONAL COPYRIGHT LAWS AND
 * TREATIES. USE, MODIFICATION, AND REDISTRIBUTION OF THIS WORK IS SUBJECT
 * TO VERSION 2.0.1 OF THE OPENLDAP PUBLIC LICENSE, A COPY OF WHICH IS
 * AVAILABLE AT HTTP://WWW.OPENLDAP.ORG/LICENSE.HTML OR IN THE FILE "LICENSE"
 * IN THE TOP-LEVEL DIRECTORY OF THE DISTRIBUTION. ANY USE OR EXPLOITATION
 * OF THIS WORK OTHER THAN AS AUTHORIZED IN VERSION 2.0.1 OF THE OPENLDAP
 * PUBLIC LICENSE, OR OTHER PRIOR WRITTEN CONSENT FROM NOVELL, COULD SUBJECT
 * TREATIES. USE, MODIFICATION, AND REDISTRIBUTION OF THIS WORK IS SUBJECT
 * TO VERSION 2.0.1 OF THE OPENLDAP PUBLIC LICENSE, A COPY OF WHICH IS
 * AVAILABLE AT HTTP://WWW.OPENLDAP.ORG/LICENSE.HTML OR IN THE FILE "LICENSE"
 * IN THE TOP-LEVEL DIRECTORY OF THE DISTRIBUTION. ANY USE OR EXPLOITATION
 * OF THIS WORK OTHER THAN AS AUTHORIZED IN VERSION 2.0.1 OF THE OPENLDAP
 * PUBLIC LICENSE, OR OTHER PRIOR WRITTEN CONSENT FROM NOVELL, COULD SUBJECT
 * THE PERPETRATOR TO CRIMINAL AND CIVIL LIABILITY.
 ******************************************************************************
*/

#include <stdio.h>
#include <stdlib.h>
                                                                                                  
#ifdef HAVE_STRING_H
#include <string.h>
#endif
 
#include "ldap.h"
#include <ldapx.h>

#define	BUFFER_SIZE	512
LDAP_F(int)
ldap_backup_object (
	LDAP 	*ld,
	const char 	*dn,
	const char    *passwd,
	char	**objectState,
	char 	**objectInfo,
	char    **chunckSize,
	int 	*size)
{
	BerElement*		requestber = NULL;
	BerElement*		returnedber = NULL;
	struct berval*	requestbv = NULL;
	struct berval*	returnedbv = NULL;
	int	i = 0, arrayCount = 0, arrayElement = 0, err = LDAP_SUCCESS;
	long	mts=0, revision = 0;
	char *returnedOID = NULL;
	char *arrayBuffer = NULL;
	char *returnedBuffer = NULL;
	char *temp = NULL;

	if (!dn)
		return LDAP_PARAM_ERROR;

	requestber = ber_alloc();
	if(requestber == NULL)
		return LDAP_NO_MEMORY;

	if (!*objectState) 
		mts = revision = 0;
	else
	{
		mts = atol(strtok(*objectState, "+"));

		if (!mts)
			return LDAP_ENCODING_ERROR;

		revision = atoi(strtok(NULL, "+"));
		if (!revision)
			return LDAP_ENCODING_ERROR;
	}

	/* stuff in the object dn, mts and revision */
	err = ber_printf(requestber, "siis", dn, mts, revision,((passwd)? passwd: "\0"));
	if (err == LBER_ERROR) {
		err = LDAP_ENCODING_ERROR;
		goto cleanup;
	}

	/* Converts the BER we just built to a berval that we will send out
	in the extended request */
	if (ber_flatten(requestber, &requestbv) == LBER_ERROR) {
		err = LDAP_NO_MEMORY;
		goto cleanup;
	}

	/* Call the ldap_extended_operation_s (synchronously) */
	err = ldap_extended_operation_s(ld,
									NLDAP_LDAP_BACKUP_REQUEST,
									requestbv,
									NULL,
									NULL,
									&returnedOID,
									&returnedbv);

    /* Make sure there is a returned OID */
	if (!returnedOID) {
		err = LDAP_NOT_SUPPORTED;
		goto cleanup;
	}
											  
	/* Is this what we were expecting to get back */
	if (strcmp(returnedOID, NLDAP_LDAP_BACKUP_RESPONSE)) {
		err = LDAP_NOT_SUPPORTED;
		goto cleanup;
	}

	if (err != LDAP_SUCCESS)
		goto cleanup;

	/* Do we have a good returned berval? */
	if (!returnedbv) { 
		/* No returned berval means rather drastic error .... operations error*/
		err = LDAP_OPERATIONS_ERROR;
		goto cleanup;
	}

	/* Build a berElement from the returned berval */
	returnedber = ber_init(returnedbv);
	if (returnedber == NULL) {
		err = LDAP_NO_MEMORY;
		goto cleanup;
	}

    err = ber_scanf(returnedber, "i", size);
	if (err == LBER_ERROR) {
		err  = LDAP_DECODING_ERROR;
		goto cleanup;
	}

	err  = ber_scanf(returnedber, "i", &mts);
	if (err == LBER_ERROR) {
		err = LDAP_DECODING_ERROR;
		goto cleanup;
	}

	err = ber_scanf(returnedber, "i", &revision);
	if (err == LBER_ERROR) {
		err = LDAP_DECODING_ERROR;
		goto cleanup;
	}

	err = ber_scanf(returnedber, "a", &returnedBuffer);
	if (err == LBER_ERROR) {
		err = LDAP_DECODING_ERROR;
		goto cleanup;
	}

	err = ber_scanf(returnedber, "{i[", &arrayCount);
	if (err == LBER_ERROR) {
		err = LDAP_DECODING_ERROR;
		goto cleanup;
	}

	temp = (char *)malloc(BUFFER_SIZE);
        if(temp == NULL) {
                err = LDAP_NO_MEMORY;
                goto cleanup;
        }
        strcpy(temp, "");
	sprintf(temp,"%d",arrayCount);
	arrayBuffer = (char *)malloc(strlen(temp)+2);
        if(arrayBuffer == NULL) {
                err = LDAP_NO_MEMORY;
                goto cleanup;
        }
        strcpy(arrayBuffer, temp);
	strcat(arrayBuffer, (const char*)";");

	for (i=0; i<arrayCount; i++)
	{
		err = ber_scanf(returnedber, "{i}", &arrayElement);
		if (err == LBER_ERROR) {
			err = LDAP_DECODING_ERROR;
			goto cleanup;
		}
		strcpy(temp, "");
		sprintf(temp,"%d",arrayElement);
		arrayBuffer = (char *)realloc(arrayBuffer, strlen(arrayBuffer)+strlen(temp)+2);
                if(arrayBuffer == NULL) {
                        err = LDAP_NO_MEMORY;
                        goto cleanup;
                }
	        strcat(arrayBuffer, temp);
		strcat(arrayBuffer, (const char*)";");
	}

	*objectInfo = (char *)malloc(*size);
        if(*objectInfo == NULL) {
                err = LDAP_NO_MEMORY;
                goto cleanup;
        }
	memcpy(*objectInfo, returnedBuffer, *size);	
	*chunckSize = (char *)malloc(strlen(arrayBuffer));
        if(*chunckSize == NULL) {
                err = LDAP_NO_MEMORY;
                goto cleanup;
        }
	memcpy(*chunckSize, arrayBuffer, strlen(arrayBuffer));

	if (*objectState)  
		free(*objectState);

	sprintf(temp, "%d+%d", mts, revision);
	*objectState = strdup(temp);

	err = LDAP_SUCCESS;

cleanup:

	/* Free memory that we allocated when building the request ber and berval */
	if (requestber)
		ber_free(requestber, 1);
	if (requestbv)
		ber_bvfree(requestbv);
    if (returnedber)
		ber_free(returnedber, 1);
	if (returnedbv)
		ber_bvfree(returnedbv);

	/* Free the returned OID string if one was returned */
	if (returnedOID)
			ldap_memfree(returnedOID);
	if (arrayBuffer)
                        ldap_memfree(arrayBuffer);
	if (returnedBuffer)
			ldap_memfree(returnedBuffer);

	if (temp)
			ldap_memfree(temp);


	/* return the appropriate error/success code */
	return err;
}

LDAP_F(int)
ldap_restore_object (
	LDAP 	*ld,
	const char	*dn,
	const char	*passwd,
	char	*objectInfo,
	char    *chunckSize,
	int	size)
{
	char *returnedOID = NULL;
	struct berval *requestbv = NULL;
	struct berval *returnedbv = NULL;
	BerElement*    requestber = NULL;
	int	 err = LDAP_SUCCESS;
	char *temp = NULL, *newbuffer = NULL, *blob = NULL;

	if (!dn|| !objectInfo)
		return LDAP_PARAM_ERROR;

   	requestber = ber_alloc();
	if(requestber == NULL)
		return LDAP_NO_MEMORY;

	/* stuff in the object dn, the trusteeDN and attrname parameters */
	err = ber_printf(requestber, "ssio", dn, ((passwd)? passwd: "\0"), size, objectInfo, size);
	if (err == LBER_ERROR) {
		err = LDAP_ENCODING_ERROR;
		goto cleanup;
	}

	temp = strtok(chunckSize, ";");
	ber_printf(requestber, "{i[", atoi(temp));
	while ((temp = strtok(NULL, ";"))!=NULL)
	{
		err = ber_printf(requestber, "{i}", atoi(temp));
		if (err == LBER_ERROR) {
			err = LDAP_ENCODING_ERROR;
			goto cleanup;
		}
		blob = temp;
	}
	err = ber_printf(requestber, "]}");

	newbuffer = (char *)malloc(size);
        if(newbuffer == NULL) {
                err = LDAP_NO_MEMORY;
                goto cleanup;
        }
	memcpy(newbuffer, blob+1+strlen(blob), size);	

	/* stuff in the object dn, the trusteeDN and attrname parameters
	err = ber_printf(requestber, "ssio", dn, passwd, size, objectInfo, size);
	if (err == LBER_ERROR) {
		err = LDAP_ENCODING_ERROR;
		goto cleanup;
	}
	*/

	/* Converts the BER we just built to a berval that we will send out
	in the extended request */
	if (ber_flatten(requestber, &requestbv) == LBER_ERROR) {
		goto cleanup;
	}

	/* Call the ldap_extended_operation (synchronously)  */
	err = ldap_extended_operation_s(ld,
									NLDAP_LDAP_RESTORE_REQUEST,
									requestbv,
									NULL,
									NULL,
									&returnedOID,
									&returnedbv);

	/* Make sure there is a returned OID */
	if (!returnedOID) {
		err = LDAP_NOT_SUPPORTED;
		goto cleanup;
	}
																		
	/* Is this what we were expecting to get back */
	if (strcmp(returnedOID, (const char *)NLDAP_LDAP_RESTORE_RESPONSE)) {
		err = LDAP_NOT_SUPPORTED;
		goto cleanup;
	}

	if (err != LDAP_SUCCESS)
		goto cleanup;

cleanup:
	/* Free memory that we allocated when building the request ber and berval */
	if (requestber)
		ber_free(requestber, 1);

	if (requestbv)
		ber_bvfree(requestbv);

	if (returnedbv)
		ber_bvfree(returnedbv);

	/* Free the returned OID string if one was returned */
	if (returnedOID)
		ldap_memfree(returnedOID);

	if (newbuffer)
		ldap_memfree(newbuffer);

	/* return the appropriate error/success code */
	return (err);
}
