/*
 * $Novell: lburp.c,v 1.0 2003/08/04 17:41:39 $
 ******************************************************************************
 * Copyright (c) 2006 Novell, Inc.
 * All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; version 2.1 of the license.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, contact Novell, Inc.
 *
 * To contact Novell about this file by physical or electronic mail,
 * you may find current contact information at www.novell.com
 *
 ****************************************************************************
 *
 * Copyright (C) 2003 Novell, Inc. All Rights Reserved.
 *
 * THIS WORK IS SUBJECT TO U.S. AND INTERNATIONAL COPYRIGHT LAWS AND
 * TREATIES. USE, MODIFICATION, AND REDISTRIBUTION OF THIS WORK IS SUBJECT
 * TO VERSION 2.0.1 OF THE OPENLDAP PUBLIC LICENSE, A COPY OF WHICH IS
 * AVAILABLE AT HTTP://WWW.OPENLDAP.ORG/LICENSE.HTML OR IN THE FILE "LICENSE"
 * IN THE TOP-LEVEL DIRECTORY OF THE DISTRIBUTION. ANY USE OR EXPLOITATION
 * OF THIS WORK OTHER THAN AS AUTHORIZED IN VERSION 2.0.1 OF THE OPENLDAP
 * PUBLIC LICENSE, OR OTHER PRIOR WRITTEN CONSENT FROM NOVELL, COULD SUBJECT
 * TREATIES. USE, MODIFICATION, AND REDISTRIBUTION OF THIS WORK IS SUBJECT
 * TO VERSION 2.0.1 OF THE OPENLDAP PUBLIC LICENSE, A COPY OF WHICH IS
 * AVAILABLE AT HTTP://WWW.OPENLDAP.ORG/LICENSE.HTML OR IN THE FILE "LICENSE"
 * IN THE TOP-LEVEL DIRECTORY OF THE DISTRIBUTION. ANY USE OR EXPLOITATION
 * OF THIS WORK OTHER THAN AS AUTHORIZED IN VERSION 2.0.1 OF THE OPENLDAP
 * PUBLIC LICENSE, OR OTHER PRIOR WRITTEN CONSENT FROM NOVELL, COULD SUBJECT
 * THE PERPETRATOR TO CRIMINAL AND CIVIL LIABILITY.
 ******************************************************************************
*/

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
                                                                                                  
#ifdef HAVE_STRING_H
#include <string.h>
#endif
 
//#include "ldap-int.h"
#include "ldap.h"
#include <ldapx.h>


/* Helper functions to decode user supplied data */
int
LBURPEncodeAdd(
		 LDAP *ld,
		 LDAP_CONST char *dn,
		 LDAPMod **attrs,
		 LDAPControl **controls,
		 BerElement * ber )
{
	int	i = 0, err=LDAP_SUCCESS; 

	// TODO - Check with somebody from ServerTeam
	assert( ld != NULL );
	assert( dn != NULL );
	
	err = ber_printf( ber, "{t{s{", /* '}}}' */
			LDAP_REQ_ADD, dn );
	
	if ( err == -1 ) {
		err = LDAP_ENCODING_ERROR;
		ber_free( ber, 1 );
		return err;
	}
	
	/* for each attribute in the entry... */
	for ( i = 0; attrs[i] != NULL; i++ ) {
		if ( ( attrs[i]->mod_op & LDAP_MOD_BVALUES) != 0 ) {
			err = ber_printf( ber, "{s[V]}", attrs[i]->mod_type,
				attrs[i]->mod_values );
		} else {
			err = ber_printf( ber, "{s[v]}", attrs[i]->mod_type,
				attrs[i]->mod_values );
		}
		if ( err == -1 ) {
			err = LDAP_ENCODING_ERROR;
			ber_free( ber, 1 );
			return err;
		}
	}
	
	if ( ber_printf( ber, /*{{*/ "}}" ) == -1 ) {
		err = LDAP_ENCODING_ERROR;
		ber_free( ber, 1 );
		return err;
	}
	
	if (controls != NULL)
	{
		if ( ber_printf( ber, "t{", LDAP_TAG_CONTROLS ) == -1 ) {
			err = LDAP_ENCODING_ERROR;
			ber_free( ber, 1 );
			return( err );
		}
		for (i=0; controls[i] != NULL; i++) 
		{
			if ( ber_printf( ber, "{s", controls[i]->ldctl_oid ) == -1 ) {
				err = LDAP_ENCODING_ERROR;
				ber_free( ber, 1 );
				return( err );
			}
			
			if (controls[i]->ldctl_iscritical)
				if ( ber_printf( ber, "b", (ber_int_t) controls[i]->ldctl_iscritical ) == -1 ) {
					err = LDAP_ENCODING_ERROR;
					ber_free( ber, 1 );
					return( err );
				}

			if (controls[i]->ldctl_value.bv_val != NULL)
				if ( ber_printf( ber, "O", &(controls[i]->ldctl_value) ) == -1 ) {
					err = LDAP_ENCODING_ERROR;
					ber_free( ber, 1 );
					return( err );
				}

			if ( ber_printf( ber, "N}" ) == -1 ) {
				err = LDAP_ENCODING_ERROR;
				ber_free( ber, 1 );
				return( err );
			}
		}

		if ( ber_printf( ber, "}" ) == -1 ) {
			err = LDAP_ENCODING_ERROR;
			ber_free( ber, 1 );
			return( err );
		}

	}

	if ( ber_printf( ber, /*{*/ "N}" ) == -1 ) {
		err = LDAP_ENCODING_ERROR;
		ber_free( ber, 1 );
		return err;
	}

	return LDAP_SUCCESS;
}

int
LBURPEncodeDelete(
			LDAP *ld,
			LDAP_CONST char* dn,
			LDAPControl **controls,
			BerElement * ber)
{
	int i = 0, err = LDAP_SUCCESS;

	assert( ber != NULL);
	assert( ld != NULL );
//	assert( LDAP_VALID( ld ) );
	assert( dn != NULL );
	
	if ( ber_printf( ber, "{ts", /* '}' */
		LDAP_REQ_DELETE, dn ) == -1 )

	{
		err = LDAP_ENCODING_ERROR;
		ber_free( ber, 1 );
		return( err );
	}
	
	if (controls != NULL)
	{
		if ( ber_printf( ber, "t{", LDAP_TAG_CONTROLS ) == -1 ) {
			err = LDAP_ENCODING_ERROR;
			ber_free( ber, 1 );
			return( err );
		}
		for (i=0; controls[i] != NULL; i++) 
		{
			if ( ber_printf( ber, "{s", controls[i]->ldctl_oid ) == -1 ) {
				err = LDAP_ENCODING_ERROR;
				ber_free( ber, 1 );
				return( err );
			}
			
			if (controls[i]->ldctl_iscritical)
				if ( ber_printf( ber, "b", (ber_int_t) controls[i]->ldctl_iscritical ) == -1 ) {
					err = LDAP_ENCODING_ERROR;
					ber_free( ber, 1 );
					return( err );
				}

			if (controls[i]->ldctl_value.bv_val != NULL)
				if ( ber_printf( ber, "O", &(controls[i]->ldctl_value) ) == -1 ) {
					err = LDAP_ENCODING_ERROR;
					ber_free( ber, 1 );
					return( err );
				}

			if ( ber_printf( ber, "}" ) == -1 ) {
				err = LDAP_ENCODING_ERROR;
				ber_free( ber, 1 );
				return( err );
			}
		}

		if ( ber_printf( ber, "N}" ) == -1 ) {
			err = LDAP_ENCODING_ERROR;
			ber_free( ber, 1 );
			return( err );
		}

	}

	if ( ber_printf( ber, /*{*/ "N}" ) == -1 ) {
		err = LDAP_ENCODING_ERROR;
		ber_free( ber, 1 );
		return( err );
	}

	return LDAP_SUCCESS;
}


/*
* LBURPMod - to encode a mod operation into the ber
*
* Parameters:
*
*	ld		LDAP descriptor
*	dn		DN of the object to modify
*	mods		List of modifications to make.  This is null-terminated
*			array of struct LDAPMod's, specifying the modifications
*			to perform.
*
* Example:
*	LDAPMod	*mods[] = { 
*			{ LDAP_MOD_ADD, "cn", { "babs jensen", "babs", 0 } },
*			{ LDAP_MOD_REPLACE, "sn", { "jensen", 0 } },
*			0
*		}
*/
int
LBURPEncodeMod( LDAP *ld,
		 LDAP_CONST char *dn,
		 LDAPMod **mods,
		 LDAPControl **controls,
		 BerElement * ber)
{
	int	i = 0, err=LDAP_SUCCESS;
	
	/*
	* A modify request looks like this:
	*	ModifyRequet ::= SEQUENCE {
	*		object		DistinguishedName,
	*		modifications	SEQUENCE OF SEQUENCE {
	*			operation	ENUMERATED {
	*				add	(0),
	*				delete	(1),
	*				replace	(2)
	*			},
	*			modification	SEQUENCE {
	*				type	AttributeType,
	*				values	SET OF AttributeValue
	*			}
	*		}
	*	}
	*/
	
	if ( ber_printf( ber, "{t{s{" /*}}}*/, LDAP_REQ_MODIFY, dn ) == -1 ) {
		err = LDAP_ENCODING_ERROR;
		ber_free( ber, 1 );
		return( err );
	}
	
	/* for each modification to be performed... */
	for ( i = 0; mods[i] != NULL; i++ ) {
		if (( mods[i]->mod_op & LDAP_MOD_BVALUES) != 0 ) {
			err = ber_printf( ber, "{e{s[V]}}",
				(ber_int_t) ( mods[i]->mod_op & ~LDAP_MOD_BVALUES ),
				mods[i]->mod_type, mods[i]->mod_bvalues );
		} else {
			err = ber_printf( ber, "{e{s[v]}}",
				(ber_int_t) mods[i]->mod_op,
				mods[i]->mod_type, mods[i]->mod_values );
		}
		
		if ( err == -1 ) {
			err = LDAP_ENCODING_ERROR;
			ber_free( ber, 1 );
			return( err );
		}
	}
	
	if ( ber_printf( ber, /*{{*/ "}}" ) == -1 ) {
		err = LDAP_ENCODING_ERROR;
		ber_free( ber, 1 );
		return( err );
	}

	if (controls != NULL)
	{
		if ( ber_printf( ber, "t{", LDAP_TAG_CONTROLS ) == -1 ) {
			err = LDAP_ENCODING_ERROR;
			ber_free( ber, 1 );
			return( err );
		}
		for (i=0; controls[i]!=NULL; i++)	
		{
			if ( ber_printf( ber, "{s", controls[i]->ldctl_oid ) == -1 ) {
				err = LDAP_ENCODING_ERROR;
				ber_free( ber, 1 );
				return( err );
			}
			
			if (controls[i]->ldctl_iscritical)
				if ( ber_printf( ber, "b", (ber_int_t) controls[i]->ldctl_iscritical ) == -1 ) {
					err = LDAP_ENCODING_ERROR;
					ber_free( ber, 1 );
					return( err );
				}

			if (controls[i]->ldctl_value.bv_val != NULL)
				if ( ber_printf( ber, "O", &(controls[i]->ldctl_value) ) == -1 ) {
					err = LDAP_ENCODING_ERROR;
					ber_free( ber, 1 );
					return( err );
				}

			if ( ber_printf( ber, "N}" ) == -1 ) {
				err = LDAP_ENCODING_ERROR;
				ber_free( ber, 1 );
				return( err );
			}
		}

		if ( ber_printf( ber, "}" ) == -1 ) {
			err = LDAP_ENCODING_ERROR;
			ber_free( ber, 1 );
			return( err );
		}

	}

	if ( ber_printf( ber, /*{*/ "}" ) == -1 ) {
		err = LDAP_ENCODING_ERROR;
		ber_free( ber, 1 );
		return( err );
	}
	
	return LDAP_SUCCESS;
}

/*
*	LBURPModRDN - to encode a modrdn into an exisiting ber
*
* Parameters:
*	ld				LDAP descriptor
*	dn				DN of the object to modify
*	newrdn			RDN to give the object
*	deleteoldrdn	nonzero means to delete old rdn values from the entry
*  ber             berelment into which this modrdn will be encoded
*
* Returns the LDAP error code.
*/

int
LBURPEncodeModRDN(
			LDAP *ld,
			LDAP_CONST char *dn,
			LDAP_CONST char *newrdn,
			LDAP_CONST char *newSuperior,
			int deleteoldrdn,
			LDAPControl **controls,
			BerElement	*ber )
{
	
	int i = 0, err=LDAP_SUCCESS;
	
	err = ber_printf( ber, "{t{ssb", /* '}' */ 
		LDAP_REQ_MODDN,
		dn, newrdn, (ber_int_t) deleteoldrdn );
	if ( err < 0 ) {
		err = LDAP_ENCODING_ERROR;
		ber_free( ber, 1 );
		return( err );
	}

	if(newSuperior)
	{
		err = ber_printf( ber, "s", newSuperior );
	}

	if ( err < 0 ) {
		err = LDAP_ENCODING_ERROR;
		ber_free( ber, 1 );
		return( err );
	}
	
	err = ber_printf( ber, /*{{*/ "}" );
	if ( err < 0 ) {
		err = LDAP_ENCODING_ERROR;
		ber_free( ber, 1 );
		return( err );
	}

	if (controls != NULL) 
	{
		if ( ber_printf( ber, "t{", LDAP_TAG_CONTROLS ) == -1 ) {
			err = LDAP_ENCODING_ERROR;
			ber_free( ber, 1 );
			return( err );
		}
		for (i=0; controls[i]!=NULL; i++)
		{
			if ( ber_printf( ber, "{s", controls[i]->ldctl_oid ) == -1 ) {
				err = LDAP_ENCODING_ERROR;
				ber_free( ber, 1 );
				return( err );
			}
			
			if (controls[i]->ldctl_iscritical)
				if ( ber_printf( ber, "b", (ber_int_t) controls[i]->ldctl_iscritical ) == -1 ) {
					err = LDAP_ENCODING_ERROR;
					ber_free( ber, 1 );
					return( err );
				}

			if (controls[i]->ldctl_value.bv_val != NULL)
				if ( ber_printf( ber, "O", &(controls[i]->ldctl_value) ) == -1 ) {
					err = LDAP_ENCODING_ERROR;
					ber_free( ber, 1 );
					return( err );
				}

			if ( ber_printf( ber, "N}" ) == -1 ) {
				err = LDAP_ENCODING_ERROR;
				ber_free( ber, 1 );
				return( err );
			}
		}

		if ( ber_printf( ber, "}" ) == -1 ) {
			err = LDAP_ENCODING_ERROR;
			ber_free( ber, 1 );
			return( err );
		}

	}

	err = ber_printf( ber, /*{{*/ "}" );
	if ( err < 0 ) {
		err = LDAP_ENCODING_ERROR;
		ber_free( ber, 1 );
		return( err );
	}
	
	return LDAP_SUCCESS;
}
LDAP_F(int)
ldap_lburp_start_request (
		LDAP *ld,
		int *msgID)
{
	BerElement*    requestber = NULL;
	struct berval* requestbv = NULL;
	int            err = LDAP_SUCCESS;

	/* Allocate a BerElement for stuffing in the function IN parameters */
	requestber = ber_alloc();
	if(requestber == NULL)
		return LDAP_NO_MEMORY;

	/* stuff in the StartLBURPRequest OID */
    err = ber_printf(requestber, "{s}", (char *) NLDAP_LBURP_INCREMENTAL_UPDATE);
	if (err == LBER_ERROR) {
		err = LDAP_ENCODING_ERROR;
		goto cleanup;
	}

	/* Convert the BER we just built to a berval that we will send out
	   in the extended request */
	if (ber_flatten(requestber, &requestbv) == LBER_ERROR) {
		err = LDAP_NO_MEMORY;
		goto cleanup;
	}
	
	/* Call the ldap_extended_operation (synchronous) */
	err = ldap_extended_operation(
			ld, 
			NLDAP_LBURP_START_REQUEST, 
			requestbv, 
			NULL,  /* no server controls */
			NULL,  /* no client controls */
			msgID);

cleanup:

	/* Free memory that we allocated when building the request ber and berval */
	if (requestbv)
		ber_bvfree(requestbv);

	if (requestber)
		ber_free(requestber, 1);

	//ld->ld_errno = err;

	/* return the appropriate error/success code */
	return (err);
}

LDAP_F(int)
ldap_parse_lburp_start_response(
				LDAP 	*ld,
				LDAPMessage	*lburpStartMessage,
				int	*resultCode,
				char	**errorMsg,
				int	*tranSize,
				int	freeIt
				)
{
	char *returnedOID = NULL;
	struct berval *returnedbv = NULL;
	BerElement 	*returnedber = NULL;
	int	err = LDAP_SUCCESS;

	err = ldap_parse_result (	ld,
								lburpStartMessage,
								resultCode,
								NULL,
								errorMsg,
								NULL,
								NULL,
								0);
	if (err != LDAP_SUCCESS)
		goto cleanup;

	err = ldap_parse_extended_result (	ld,
										lburpStartMessage,
										&returnedOID,
										&returnedbv,
										freeIt);

	if (err != LDAP_SUCCESS)
		goto cleanup;

	/* Make sure there is a returned OID */
	if (!returnedOID) {
		err = LDAP_NOT_SUPPORTED;
		goto cleanup;
	}

	/* Is this what we were expecting to get back */
	if (strcmp(returnedOID, NLDAP_LBURP_START_RESPONSE))
		err = LDAP_NOT_SUPPORTED;

	if (returnedbv == NULL) {
			goto cleanup;
	}

	/* Parse the returned entries  */
	returnedber = ber_init(returnedbv);
	if (returnedber == NULL) {
		err = LDAP_NO_MEMORY;
		goto cleanup;
	}

	/* The returned data contains the transaction size - scan it  */
	err = ber_scanf(returnedber, "{i}", tranSize);
	if (err == LBER_ERROR) {
		err = LDAP_DECODING_ERROR;
		goto cleanup;
	}

	err = LDAP_SUCCESS;

cleanup:
	/* Free the returned OID string if one was returned */
	if (returnedOID)
		ldap_memfree(returnedOID);

	/* free any memory the extended_operation API must have allocated */
	if (returnedbv)
		ber_bvfree(returnedbv);

	if (returnedber)
			ber_free(returnedber, 1);

	//ld->ld_errno = err;

	/* return the appropriate error/success code */
	return (err);
}

LDAP_F(int)
ldap_lburp_operation_request(
				LDAP *ld,
				int sequenceNumber, 
				LBURPUpdateOperationList **op,
				int *msgID)
{
	BerElement*		requestber = NULL;
	struct berval*  requestbv = NULL;
	int  i = 0, err = LDAP_SUCCESS;

    /* Allocate a BerElement for stuffing in the function IN parameters */
	requestber = ber_alloc();
	if(requestber == NULL)
		return LDAP_NO_MEMORY;

	/* print sequence Number in the and the starting Bracket */
	if (-1 == ber_printf(requestber,"{i", sequenceNumber))
	{
		err = LDAP_ENCODING_ERROR;
		goto cleanup;
	}

	/* for each record in the operation ... */
	for ( i = 0; op[i] != NULL; i++ ) {
		if(op[i]->operation == LDAP_REQ_ADD) {
			err = LBURPEncodeAdd(ld, op[i]->dn, (LDAPMod **)op[i]->value.attrs, (LDAPControl **)op[i]->ServerControls, requestber);
			if ( err != LDAP_SUCCESS )
			goto cleanup;
		}
		else if(op[i]->operation == LDAP_REQ_DELETE) {
			err =  LBURPEncodeDelete(ld, op[i]->dn, (LDAPControl **)op[i]->ServerControls, requestber);
			if ( LDAP_SUCCESS != err )
				goto cleanup;
		}
		else if(op[i]->operation == LDAP_REQ_MODIFY) {
			err = LBURPEncodeMod(ld, op[i]->dn, (LDAPMod **)op[i]->value.attrs, (LDAPControl**)op[i]->ServerControls, requestber);
			if (LDAP_SUCCESS != err)
				goto cleanup;
		}
		else if(op[i]->operation == LDAP_REQ_MODRDN) {
			err = LBURPEncodeModRDN(ld,op[i]->dn, op[i]->value.newRDN, op[i]->value.newSuperior, op[i]->value.deleteOldRDN, (LDAPControl **)op[i]->ServerControls, requestber);
			if (LDAP_SUCCESS != err)
			goto cleanup;
		}
		else {
		err = LDAP_NOT_SUPPORTED;
		goto cleanup;
		}
	}
	
	if ( ber_printf(requestber, "}") == -1 ) {
		err = LDAP_ENCODING_ERROR;
		goto cleanup;
	}

	/* Convert the BER we just built to a berval that we will send out
	   in the extended request */
	if (ber_flatten(requestber, &requestbv) == LBER_ERROR) {
		err = LDAP_NO_MEMORY;
		goto cleanup;
	}

	err = ldap_extended_operation(	ld, 
									NLDAP_LBURP_OPERATION_REQUEST, 
									requestbv, 
									NULL, 
									NULL, 
									msgID);
cleanup:

	/* Free memory that we allocated when building the request ber and berval */
	if (requestbv)
		ber_bvfree(requestbv);

	if (requestber)
		ber_free(requestber, 1);

	//ld->ld_errno = err;

	/* return the appropriate error/success code */
	return (err);
}

LDAP_F(int) 
ldap_parse_lburp_operation_response( 
		LDAP *ld,
		LDAPMessage *lburpMessage,
		int	*resultCode,
		char	**errorMsg,
		int	*failedOpsCount,
		LBURPUpdateResult **failedOperations, 
		int freeIt)
{

	BerElement*	   returnedber = NULL;
	struct berval	*returnedbv= NULL; 
	LBURPUpdateResult *current=NULL;
	char *returnedOID = NULL;
	int	err = LDAP_SUCCESS;
	ber_tag_t	tag = 0;
	ber_len_t	tlen = 0;
	char *temp=NULL;

	/* initialize failedOperatons list to NULL*/ 
	if ( failedOperations != NULL )
		*failedOperations = NULL; 

	if (failedOpsCount != NULL)
			*failedOpsCount=0;

	err = ldap_parse_result (ld, lburpMessage, resultCode, 
									NULL, errorMsg, NULL, NULL, 0);

	if ( err != LDAP_SUCCESS )
		goto cleanup;

	err = ldap_parse_extended_result (ld, lburpMessage, &returnedOID, 
												&returnedbv, freeIt);

	if (err != LDAP_SUCCESS) 
		goto cleanup;
	
	/* Make sure there is a returned OID */
	if (!returnedOID) {
		err = LDAP_NOT_SUPPORTED;
		goto cleanup;
	}

	/* Is this what we were expecting to get back */
	if ( 0 != strcmp(returnedOID, NLDAP_LBURP_OPERATION_RESPONSE)) {
		err = LDAP_NOT_SUPPORTED;
		goto cleanup;
	}
	/* Meaning all updates gone through fine, no error reported */
	if ( returnedbv == NULL ) {
			err = LDAP_SUCCESS;
			goto cleanup;
	}

	returnedber = ber_init(returnedbv);
	if (returnedber == NULL) {
		err = LDAP_NO_MEMORY;
		goto cleanup;
	}

	*failedOperations = (LBURPUpdateResult*)malloc(1 * sizeof(LBURPUpdateResult)); 

	if ( *failedOperations == NULL ) {
		err = LDAP_NO_MEMORY;
		goto cleanup;
	} 
	else 
		current = *failedOperations;	
    
    while (1) {

	if (( tag = ber_peek_tag(returnedber,&tlen)) == LBER_DEFAULT) {
		err = LDAP_SUCCESS;	
		goto cleanup;
	}

	if ((err = ber_scanf(returnedber,"{i", &(current->packageID))) == LBER_ERROR) {
		err = LDAP_DECODING_ERROR;
		goto cleanup;
	}

	if((tag = ber_peek_tag(returnedber ,&tlen)) == LBER_ERROR) {
		err = LDAP_SUCCESS;
		goto cleanup;
	}

	if ((err = ber_scanf(returnedber,"{iaa" /*}*/, &(current->resultCode),
			&temp, &(current->errorMsg))) == LBER_ERROR) {
		err = LDAP_DECODING_ERROR;
		goto cleanup;
	}
	free(temp);

	*failedOpsCount = *failedOpsCount + 1;
	*failedOperations = (LBURPUpdateResult*)
					 realloc( *failedOperations, (*failedOpsCount + 1)*sizeof(LBURPUpdateResult));
	current = *failedOperations + *failedOpsCount;
	} //of while
cleanup:

    /* Free memory that we allocated when building the request ber and berval */
	if (returnedbv)
		ber_bvfree(returnedbv);
					
	if (returnedber)
		ber_free(returnedber, 1);

	/* Free the returned OID string if one was returned */
	if (returnedOID)
		ldap_memfree(returnedOID);

	if ((err!= LDAP_SUCCESS) && (err!= LBER_DEFAULT) && (failedOperations!=NULL) && (*failedOperations != NULL)) {
		free (*failedOperations);
		*failedOperations = NULL;
		*failedOpsCount = 0;
	}

	if ( err == 80 )
	err = 0;

	//ld->ld_errno = err;

	/* return the appropriate error/success code */
	return (err);
}


LDAP_F(int)
ldap_lburp_end_request (
		LDAP *ld,
		int sequenceNumber,
		int	*msgID
		)
{
	BerElement*		requestber = NULL;
	struct berval*  requestbv = NULL;
	int				err = LDAP_SUCCESS;

	/* Allocate a BerElement for stuffing in the function IN parameters */
	requestber = ber_alloc_t(LBER_USE_DER);
	if(requestber == NULL)
		return LDAP_NO_MEMORY;

	/* stuff in the StartLBURPRequest OID */
    err = ber_printf(requestber, "{i}", sequenceNumber);
	if (err == LBER_ERROR) {
		err = LDAP_ENCODING_ERROR;
		goto cleanup;
	}

	/* Convert the BER we just built to a berval that we will send out
	   in the extended request */
	if (ber_flatten(requestber, &requestbv) == LBER_ERROR) {
		err = LDAP_NO_MEMORY;
		goto cleanup;
	}

	err = ldap_extended_operation(
			ld, 
			NLDAP_LBURP_END_REQUEST, 
			requestbv, 
			NULL,  /* no server controls */
			NULL,  /* no client controls */
			msgID);

cleanup:

	/* Free memory that we allocated when building the request ber and berval */
	if (requestbv)
		ber_bvfree(requestbv);

	if (requestber)
		ber_free(requestber, 1);

	//ld->ld_errno = err;

	/* return the appropriate error/success code */
	return (err);
}

LDAP_F(int) 
ldap_parse_lburp_end_response (
		LDAP *ld,
		LDAPMessage *lburpEndMessage,
		int	*resultCode,
		char **errorMsg,
		int	freeIt
		)
{		
	BerElement 	*returnedber = NULL;
	struct berval *returnedbv = NULL;
	char *returnedOID = NULL;
	int	err = LDAP_SUCCESS;

	err = ldap_parse_result (	ld,
								lburpEndMessage,
								resultCode,
								NULL,
								errorMsg,
								NULL,
								NULL,
								0);
	if (err != LDAP_SUCCESS)
		goto cleanup;

	err = ldap_parse_extended_result (	ld,
										lburpEndMessage,
										&returnedOID,
										&returnedbv,
										freeIt);

	if (err != LDAP_SUCCESS)
		goto cleanup;

	/* Make sure there is a returned OID */
	if (!returnedOID) {
		err = LDAP_NOT_SUPPORTED;
		goto cleanup;
	}

	/* Is this what we were expecting to get back */
	if (strcmp(returnedOID, NLDAP_LBURP_END_RESPONSE))
		err = LDAP_NOT_SUPPORTED;

cleanup:
	/* Free the returned OID string if one was returned */
	if (returnedOID)
		ldap_memfree(returnedOID);

	/* free any memory the extended_operation API must have allocated */
	if (returnedbv)
		ber_bvfree(returnedbv);

	//ld->ld_errno = err;

	/* return the appropriate error/success code */
	return (err);
}

