/*
 * $Novell: lcother.c,v 1.8 2002/01/25 00:12:23 $
 ******************************************************************************
 * Copyright (c) 2006 Novell, Inc.
 * All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; version 2.1 of the license.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, contact Novell, Inc.
 *
 * To contact Novell about this file by physical or electronic mail,
 * you may find current contact information at www.novell.com
 *
 ****************************************************************************
 *
 * Copyright (C) 1999, 2000 Novell, Inc. All Rights Reserved.
 * 
 * THIS WORK IS SUBJECT TO U.S. AND INTERNATIONAL COPYRIGHT LAWS AND
 * TREATIES. USE, MODIFICATION, AND REDISTRIBUTION OF THIS WORK IS SUBJECT
 * TO VERSION 2.0.1 OF THE OPENLDAP PUBLIC LICENSE, A COPY OF WHICH IS
 * AVAILABLE AT HTTP://WWW.OPENLDAP.ORG/LICENSE.HTML OR IN THE FILE "LICENSE"
 * IN THE TOP-LEVEL DIRECTORY OF THE DISTRIBUTION. ANY USE OR EXPLOITATION
 * OF THIS WORK OTHER THAN AS AUTHORIZED IN VERSION 2.0.1 OF THE OPENLDAP
 * PUBLIC LICENSE, OR OTHER PRIOR WRITTEN CONSENT FROM NOVELL, COULD SUBJECT
 * THE PERPETRATOR TO CRIMINAL AND CIVIL LIABILITY. 
 ******************************************************************************
 */

#if defined(WIN32)
#include<windows.h>
#define uint16 unsigned short
#define uint8 unsigned char

#elif defined(UNIX)
#define uint16 unsigned short
#define uint8  unsigned char

#elif defined(MODESTO)
#include <stdlib.h>
#define uint16 unsigned short
#define uint8  unsigned char

#endif

//#include "portable.h"

#ifdef HAVE_STRING_H
	#include <string.h>
#endif

/* Core LDAP header */
#include "ldap.h"

/* The SDK header file										*/
#include "ldapx.h"

/*wrapper functions header */
#include "lcother.h"

/* ------------------------------------------------------------------
ldap_trigger_back_process
==============================

RequestBer
	NULL

ResponseBer
	NULL

Description
	This API is used to trigger the specified background process on the 
	NDS server
-------------------------------------------------------------------- */
LDAP_F(int) ldap_trigger_back_process(LDAP* ld, int processID)
{

	switch(processID) {

	case LDAP_BK_PROCESS_BKLINKER:
		return wrapper_ldap_trigger_backlink(ld);
		break;

	case LDAP_BK_PROCESS_JANITOR:
		return wrapper_ldap_trigger_janitor(ld);
		break;

	case LDAP_BK_PROCESS_LIMBER:
		return wrapper_ldap_trigger_limber(ld);
		break;

	case LDAP_BK_PROCESS_SKULKER:
		return wrapper_ldap_trigger_skulker(ld);
		break;

	case LDAP_BK_PROCESS_SCHEMA_SYNC:
		return wrapper_ldap_trigger_schema_sync(ld);
		break;

	case LDAP_BK_PROCESS_PART_PURGE:
		return wrapper_ldap_trigger_partition_purge(ld);
		break;

	default:
		return LDAP_PARAM_ERROR;
	}

}




/* ------------------------------------------------------------------
wrapper_ldap_trigger_backlink
==============================

RequestBer
	NULL

ResponseBer
	NULL

Description
	This API is called indirectly and is used to trigger the backling 
	process on the specified server
-------------------------------------------------------------------- */
int wrapper_ldap_trigger_backlink(LDAP* ld)
{
   int            err = LDAP_SUCCESS;

   char*       returnedOID = NULL;
   struct berval* returnedbv = NULL;
   
   /* Call the ldap_extended_operation (synchronously) */
   if ((err = ldap_extended_operation_s(ld, NLDAP_TRIGGER_BKLINKER_EXTENDED_REQUEST, NULL, NULL, NULL, &returnedOID, &returnedbv)) != LDAP_SUCCESS)
      goto done;
   
   /* Make sure there is a returned OID */
   if (!returnedOID) {
      err = LDAP_NOT_SUPPORTED;
      goto done;
   }
   
   /* Is this what we were expecting to get back */
   if (strcmp(returnedOID, NLDAP_TRIGGER_BKLINKER_EXTENDED_REPLY)) {
      err = LDAP_NOT_SUPPORTED;
      goto done;
   }

done:
   /* Free the returned OID string if one was returned */
   if (returnedOID)
	   ldap_memfree(returnedOID);

   /* free any memory the extended_operation API must have allocated */
	if (returnedbv)
      ber_bvfree(returnedbv);

   /* return the appropriate error/success code */
   return (err);   
}


/* ------------------------------------------------------------------
wrapper_ldap_trigger_janitor
==============================

RequestBer
	NULL

ResponseBer
	NULL

Description
	This API is called indirectly and is used to trigger the 
	janitor process on the specified server
-------------------------------------------------------------------- */
int wrapper_ldap_trigger_janitor(LDAP* ld)
{
   int            err = LDAP_SUCCESS;

   char*       returnedOID = NULL;
   struct berval* returnedbv = NULL;
   
   /* Call the ldap_extended_operation (synchronously) */
   if ((err = ldap_extended_operation_s(ld, NLDAP_TRIGGER_JANITOR_EXTENDED_REQUEST, NULL, NULL, NULL, &returnedOID, &returnedbv)) != LDAP_SUCCESS)
      goto done;
   
   /* Make sure there is a returned OID */
   if (!returnedOID) {
      err = LDAP_NOT_SUPPORTED;
      goto done;
   }
   
   /* Is this what we were expecting to get back */
   if (strcmp(returnedOID, NLDAP_TRIGGER_JANITOR_EXTENDED_REPLY)) {
      err = LDAP_NOT_SUPPORTED;
      goto done;
   }

done:
   /* Free the returned OID string if one was returned */
   if (returnedOID)
	   ldap_memfree(returnedOID);

   /* free any memory the extended_operation API must have allocated */
	if (returnedbv)
      ber_bvfree(returnedbv);

   /* return the appropriate error/success code */
   return (err);   
}

/* ------------------------------------------------------------------
wrapper_ldap_trigger_limber
==============================

RequestBer
	NULL

ResponseBer
	NULL

Description
	This API is called indirectly and is used to trigger the 
	limber process on the specified server
-------------------------------------------------------------------- */
int wrapper_ldap_trigger_limber(LDAP* ld)
{
   int            err = LDAP_SUCCESS;

   char*       returnedOID = NULL;
   struct berval* returnedbv = NULL;
   
   /* Call the ldap_extended_operation (synchronously) */
   if ((err = ldap_extended_operation_s(ld, NLDAP_TRIGGER_LIMBER_EXTENDED_REQUEST, NULL, NULL, NULL, &returnedOID, &returnedbv)) != LDAP_SUCCESS)
      goto done;
   
   /* Make sure there is a returned OID */
   if (!returnedOID) {
      err = LDAP_NOT_SUPPORTED;
      goto done;
   }
   
   /* Is this what we were expecting to get back */
   if (strcmp(returnedOID, NLDAP_TRIGGER_LIMBER_EXTENDED_REPLY)) {
      err = LDAP_NOT_SUPPORTED;
      goto done;
   }

done:
   /* Free the returned OID string if one was returned */
   if (returnedOID)
	   ldap_memfree(returnedOID);

   /* free any memory the extended_operation API must have allocated */
	if (returnedbv)
      ber_bvfree(returnedbv);

   /* return the appropriate error/success code */
   return (err);   
}


/* ------------------------------------------------------------------
wrapper_ldap_trigger_skulker
==============================

RequestBer
	NULL

ResponseBer
	NULL

Description
	This API is called indirectly and is used to trigger the skulker 
	process on the specified server
-------------------------------------------------------------------- */
int wrapper_ldap_trigger_skulker(LDAP* ld)
{
   int            err = LDAP_SUCCESS;

   char*       returnedOID = NULL;
   struct berval* returnedbv = NULL;
   
   /* Call the ldap_extended_operation (synchronously) */
   if ((err = ldap_extended_operation_s(ld, NLDAP_TRIGGER_SKULKER_EXTENDED_REQUEST, NULL, NULL, NULL, &returnedOID, &returnedbv)) != LDAP_SUCCESS)
      goto done;
   
   /* Make sure there is a returned OID */
   if (!returnedOID) {
      err = LDAP_NOT_SUPPORTED;
      goto done;
   }
   
   /* Is this what we were expecting to get back */
   if (strcmp(returnedOID, NLDAP_TRIGGER_SKULKER_EXTENDED_REPLY)) {
      err = LDAP_NOT_SUPPORTED;
      goto done;
   }

done:
   /* Free the returned OID string if one was returned */
   if (returnedOID)
	   ldap_memfree(returnedOID);

   /* free any memory the extended_operation API must have allocated */
	if (returnedbv)
      ber_bvfree(returnedbv);

   /* return the appropriate error/success code */
   return (err);   
}

/* ------------------------------------------------------------------
wrapper_ldap_trigger_schema_sync
==============================

RequestBer
	NULL

ResponseBer
	NULL

Description
	This API is called indirectly and is used to trigger the schema 
	sync process on the specified server
-------------------------------------------------------------------- */
int wrapper_ldap_trigger_schema_sync(LDAP* ld)
{
   int            err = LDAP_SUCCESS;

   char*       returnedOID = NULL;
   struct berval* returnedbv = NULL;
   
   /* Call the ldap_extended_operation (synchronously) */
   if ((err = ldap_extended_operation_s(ld, NLDAP_TRIGGER_SCHEMA_SYNC_EXTENDED_REQUEST, NULL, NULL, NULL, &returnedOID, &returnedbv)) != LDAP_SUCCESS)
      goto done;
   
   /* Make sure there is a returned OID */
   if (!returnedOID) {
      err = LDAP_NOT_SUPPORTED;
      goto done;
   }
   
   /* Is this what we were expecting to get back */
   if (strcmp(returnedOID, NLDAP_TRIGGER_SCHEMA_SYNC_EXTENDED_REPLY)) {
      err = LDAP_NOT_SUPPORTED;
      goto done;
   }

done:
   /* Free the returned OID string if one was returned */
   if (returnedOID)
	   ldap_memfree(returnedOID);

   /* free any memory the extended_operation API must have allocated */
	if (returnedbv)
      ber_bvfree(returnedbv);

   /* return the appropriate error/success code */
   return (err);   
}

/* ------------------------------------------------------------------
wrapper_ldap_trigger_partition_purge
==============================

RequestBer
	NULL

ResponseBer
	NULL

Description
	This API is called indirectly and is used to trigger the 
	partition purge process on the specified server
-------------------------------------------------------------------- */
int wrapper_ldap_trigger_partition_purge(LDAP* ld)
{
   int            err = LDAP_SUCCESS;

   char*       returnedOID = NULL;
   struct berval* returnedbv = NULL;
   
   /* Call the ldap_extended_operation (synchronously) */
   if ((err = ldap_extended_operation_s(ld, NLDAP_TRIGGER_PART_PURGE_EXTENDED_REQUEST, NULL, NULL, NULL, &returnedOID, &returnedbv)) != LDAP_SUCCESS)
      goto done;
   
   /* Make sure there is a returned OID */
   if (!returnedOID) {
      err = LDAP_NOT_SUPPORTED;
      goto done;
   }
   
   /* Is this what we were expecting to get back */
   if (strcmp(returnedOID, NLDAP_TRIGGER_PART_PURGE_EXTENDED_REPLY)) {
      err = LDAP_NOT_SUPPORTED;
      goto done;
   }

done:
   /* Free the returned OID string if one was returned */
   if (returnedOID)
	   ldap_memfree(returnedOID);

   /* free any memory the extended_operation API must have allocated */
	if (returnedbv)
      ber_bvfree(returnedbv);

   /* return the appropriate error/success code */
   return (err);   
}

