/*
 * $Novell: lcpart.c,v 1.26 2002/03/11 17:41:39 $
 ******************************************************************************
 * Copyright (c) 2006 Novell, Inc.
 * All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; version 2.1 of the license.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, contact Novell, Inc.
 *
 * To contact Novell about this file by physical or electronic mail,
 * you may find current contact information at www.novell.com
 *
 ****************************************************************************
 *
 * Copyright (C) 1999, 2000 Novell, Inc. All Rights Reserved.
 * 
 * THIS WORK IS SUBJECT TO U.S. AND INTERNATIONAL COPYRIGHT LAWS AND
 * TREATIES. USE, MODIFICATION, AND REDISTRIBUTION OF THIS WORK IS SUBJECT
 * TO VERSION 2.0.1 OF THE OPENLDAP PUBLIC LICENSE, A COPY OF WHICH IS
 * AVAILABLE AT HTTP://WWW.OPENLDAP.ORG/LICENSE.HTML OR IN THE FILE "LICENSE"
 * IN THE TOP-LEVEL DIRECTORY OF THE DISTRIBUTION. ANY USE OR EXPLOITATION
 * OF THIS WORK OTHER THAN AS AUTHORIZED IN VERSION 2.0.1 OF THE OPENLDAP
 * PUBLIC LICENSE, OR OTHER PRIOR WRITTEN CONSENT FROM NOVELL, COULD SUBJECT
 * THE PERPETRATOR TO CRIMINAL AND CIVIL LIABILITY. 
 ******************************************************************************
 */

#if defined(WIN32)
#include<windows.h>
#elif defined(MODESTO)
#include <stdlib.h>
#elif defined(UNIX)
#endif

//#include "portable.h"

#include "ldap.h"

/* The SDK definitions are here*/
#include "ldapx.h"

#ifdef HAVE_STRING_H
#include <string.h>
#endif


/*---------------------------------------------------------------------------
 * These flags need to be kept in sync manually.  They are copied as is from
   dswire.h
 */

#define DSP_OUTPUT_FIELDS           0x00000001L
#define DSP_PARTITION_ID            0x00000002L
#define DSP_REPLICA_STATE           0x00000004L
#define DSP_MODIFICATION_TIMESTAMP  0x00000008L
#define DSP_PURGE_TIME              0x00000010L
#define DSP_LOCAL_PARTITION_ID      0x00000020L
#define DSP_PARTITION_DN            0x00000040L
#define DSP_REPLICA_TYPE            0x00000080L
#define DSP_PARTITION_BUSY          0x00000100L

/* ------------------------------------------------------------------
ldap_split_partition
==============================

RequestBer
    Verb        INTEGER
    flags       INTEGER
    dn          OCTET STRING

ResponseBer
    NULL

Description
    This API splits a partition rooted at dn
-------------------------------------------------------------------- */
LDAP_F(int) ldap_split_partition(LDAP* ld, char* dn, int flags)
{
   
    BerElement*    requestber = NULL;
    int            err = LDAP_SUCCESS;
    struct berval*  requestbv = NULL;

    char*       returnedOID = NULL;
    struct berval* returnedbv = NULL;

   
    /* Allocate a BerElement for stuffing in the function IN parameters */
    requestber = ber_alloc();
    if(requestber == NULL)
        return LDAP_NO_MEMORY;

    /* stuff in the Verb, the flags and the dn that identify the API */
    err = ber_printf(requestber, "is", flags, dn);
    if (err == LBER_ERROR) {
        err = LDAP_ENCODING_ERROR;
        goto cleanup;
    }

    /* Converts the BER we just built to a berval that we will send out
      in the extended request */
    if (ber_flatten(requestber, &requestbv) == LBER_ERROR) {
        err = LDAP_NO_MEMORY;
        goto cleanup;
    }

   /* Call the ldap_extended_operation (synchronously) */
   if ((err = ldap_extended_operation_s(ld, NLDAP_SPLIT_PARTITION_EXTENDED_REQUEST, requestbv, NULL, NULL, &returnedOID, &returnedbv)) != LDAP_SUCCESS)
       goto cleanup;

   /* Make sure there is a returned OID */
   if (!returnedOID) {
       err = LDAP_NOT_SUPPORTED;
       goto cleanup;
   }
   
   /* Is this what we were expecting to get back */
   if (strcmp(returnedOID, NLDAP_SPLIT_PARTITION_EXTENDED_REPLY))
       err = LDAP_NOT_SUPPORTED;

   /* NO RETURNED BERVAL FOR THIS API */

cleanup:
    /* Free the returned OID string if one was returned */
    if (returnedOID)
        ldap_memfree(returnedOID);

      /* free any memory the extended_operation API must have allocated */
    if (returnedbv)
        ber_bvfree(returnedbv);

    /* Free memory that we allocated when building the request ber and berval */
    if (requestbv)
        ber_bvfree(requestbv);

    if (requestber)
        ber_free(requestber, 1); 

    /* return the appropriate error/success code */
    return (err);     
}

LDAP_F(int) ldap_create_naming_context(LDAP* ld, char* dn, int flags)
{  
    return ldap_split_partition(ld, dn, flags);
}

/* ------------------------------------------------------------------
ldap_merge_partitions
==============================

RequestBer
    Verb        INTEGER
    flags       INTEGER
    dn          OCTET STRING

ResponseBer
    NULL

Description
    This API merges the partition rooted at dn with the parent
    partition.
-------------------------------------------------------------------- */
LDAP_F(int) ldap_merge_partitions(LDAP* ld, char*dn, int flags)
{
    BerElement*    requestber = NULL;
    int            err = LDAP_SUCCESS;
    struct berval*  requestbv = NULL;

    char*       returnedOID = NULL;
    struct berval* returnedbv = NULL;

    /* Allocate a BerElement for stuffing in the function IN parameters */
    requestber = ber_alloc();
    if(requestber == NULL)
        return LDAP_NO_MEMORY;

    /* stuff in the Verb, the flags and the dn that identify the API */
    err = ber_printf(requestber, "is", flags, dn);
    if (err == LBER_ERROR) {
        err = LDAP_ENCODING_ERROR;
        goto cleanup;
    }

    /* Converts the BER we just built to a berval that we will send out
       in the extended request */
    if (ber_flatten(requestber, &requestbv) == LBER_ERROR) {
        err = LDAP_NO_MEMORY;
        goto cleanup;
    }

    /* Call the ldap_extended_operation (synchronously) */
    if ((err = ldap_extended_operation_s(ld, NLDAP_MERGE_PARTITION_EXTENDED_REQUEST, requestbv, NULL, NULL, &returnedOID, &returnedbv)) != LDAP_SUCCESS)
        goto cleanup;
   
    /* Make sure there is a returned OID */
    if (!returnedOID) {
        err = LDAP_NOT_SUPPORTED;
        goto cleanup;
    }
   
    /* Is this what we were expecting to get back */
    if (strcmp(returnedOID, NLDAP_MERGE_PARTITION_EXTENDED_REPLY))
        err = LDAP_NOT_SUPPORTED;
   
    /* NO RETURNED BERVAL FOR THIS API */

cleanup:
   /* Free the returned OID string if one was returned */
    if (returnedOID)
        ldap_memfree(returnedOID);

    /* free any memory the extended_operation API must have allocated */
    if (returnedbv)
        ber_bvfree(returnedbv);

    /* Free memory that we allocated when building the request ber and berval */
    if (requestbv)
        ber_bvfree(requestbv);

    if (requestber)
        ber_free(requestber, 1);

    /* return the appropriate error/success code */
    return (err);       
}

LDAP_F(int) ldap_merge_naming_contexts(LDAP* ld, char*dn, int flags)
{
    return ldap_merge_partitions(ld, dn, flags);
}

/* ------------------------------------------------------------------
ldap_add_replica
==============================

RequestBer
    Verb        INTEGER
    flags       INTEGER
    replicaType INTEGER
    serverName  OCTET STRING
    dn          OCTET STRING

ResponseBer
    NULL

Description
    This API adds a replica of type replicaType to the server
    identified by serverName.  The partition is specified by
    the dn parameter.
-------------------------------------------------------------------- */
LDAP_F(int) ldap_add_replica(LDAP* ld, char* dn, char* serverDN, LDAP_REPLICA_TYPE replicaType, int flags)
{
    BerElement*    requestber = NULL;
    int            err = LDAP_SUCCESS;
    struct berval*  requestbv = NULL;

    char*       returnedOID = NULL;
    struct berval* returnedbv = NULL;
   
    /* Allocate a BerElement for stuffing in the function IN parameters */
    requestber = ber_alloc();
    if(requestber == NULL)
        return LDAP_NO_MEMORY;

    /* stuff in the Verb, the replica type, flags and the dn's of the server and replica*/
    err = ber_printf(requestber, "iiss", flags, replicaType, serverDN, dn);
    if (err == LBER_ERROR) {
        err = LDAP_ENCODING_ERROR;
        goto cleanup;
    }

    /* Converts the BER we just built to a berval that we will send out
      in the extended request */
    if (ber_flatten(requestber, &requestbv) == LBER_ERROR) {
        err = LDAP_NO_MEMORY;
        goto cleanup;
    }

    /* Call the ldap_extended_operation (synchronously) */
    if ((err = ldap_extended_operation_s(ld, NLDAP_ADD_REPLICA_EXTENDED_REQUEST, requestbv, NULL, NULL, &returnedOID, &returnedbv)) != LDAP_SUCCESS)
        goto cleanup;

    /* Make sure there is a returned OID */
    if (!returnedOID) {
        err = LDAP_NOT_SUPPORTED;
        goto cleanup;
    }

    /* Is this what we were expecting to get back */
    if (strcmp(returnedOID, NLDAP_ADD_REPLICA_EXTENDED_REPLY))
        err = LDAP_NOT_SUPPORTED;

    /* NO RETURNED BERVAL FOR THIS API */

cleanup:
    /* Free the returned OID string if one was returned */
    if (returnedOID)
        ldap_memfree(returnedOID);

    /* free any memory the extended_operation API must have allocated */
    if (returnedbv)
        ber_bvfree(returnedbv);

   /* Free memory that we allocated when building the request ber and berval */
    if (requestbv)
        ber_bvfree(requestbv);

    if (requestber)
        ber_free(requestber, 1); 

    /* return the appropriate error/success code */
    return (err);
}

/* ------------------------------------------------------------------
ldap_remove_replica
==============================

RequestBer
    Verb        INTEGER
    flags       INTEGER
    serverName  OCTET STRING
    dn          OCTET STRING

ResponseBer
    NULL

Description
    This API removes the replica specified by dn from the server
    specified by the serverName.
-------------------------------------------------------------------- */
LDAP_F(int) ldap_remove_replica(LDAP* ld, char* dn, char* serverDN, int flags)
{
    BerElement*    requestber = NULL;
    int            err = LDAP_SUCCESS;
    struct berval*  requestbv = NULL;

    char*       returnedOID = NULL;
    struct berval* returnedbv = NULL;
 
    /* Allocate a BerElement for stuffing in the function IN parameters */
    requestber = ber_alloc();
    if(requestber == NULL)
        return LDAP_NO_MEMORY;

    /* stuff in the Verb, the flags and the dn's of the server and replica*/
    err = ber_printf(requestber, "iss", flags, serverDN, dn);
    if (err == LBER_ERROR) {
        err = LDAP_ENCODING_ERROR;
        goto cleanup;
    }

    /* Converts the BER we just built to a berval that we will send out
       in the extended request */
    if (ber_flatten(requestber, &requestbv) == LBER_ERROR) {
        err = LDAP_NO_MEMORY;
        goto cleanup;
    }

    /* Call the ldap_extended_operation (synchronously) */
    if ((err = ldap_extended_operation_s(ld, NLDAP_REMOVE_REPLICA_EXTENDED_REQUEST, requestbv, NULL, NULL, &returnedOID, &returnedbv)) != LDAP_SUCCESS)
        goto cleanup;
   
    /* Make sure there is a returned OID */
    if (!returnedOID) {
        err = LDAP_NOT_SUPPORTED;
        goto cleanup;
    }
   
    /* Is this what we were expecting to get back */
    if (strcmp(returnedOID, NLDAP_REMOVE_REPLICA_EXTENDED_REPLY))
        err = LDAP_NOT_SUPPORTED;

    /* NO RETURNED BERVAL FOR THIS API */

cleanup:
    /* Free the returned OID string if one was returned */
    if (returnedOID)
        ldap_memfree(returnedOID);

    /* free any memory the extended_operation API must have allocated */
    if (returnedbv)
        ber_bvfree(returnedbv);

    /* Free memory that we allocated when building the request ber and berval */
    if (requestbv)
        ber_bvfree(requestbv);

    if (requestber)
        ber_free(requestber, 1); 

    /* return the appropriate error/success code */
    return (err);
}

/* ------------------------------------------------------------------
ldap_partition_entry_count
==============================

RequestBer
    Verb        INTEGER
    dn          OCTET STRING

ResponseBer
    count       INTEGER

Description
    This API returns a count of the number of objects in the specified
    partition.
-------------------------------------------------------------------- */
LDAP_F(int) ldap_partition_entry_count(LDAP* ld, char* dn, unsigned long* count)
{
    BerElement*    requestber = NULL;
    BerElement*     returnedber = NULL;
    int            err = LDAP_SUCCESS;
    struct berval*  requestbv = NULL;

    char*       returnedOID = NULL;
    struct berval* returnedbv = NULL;

    /* Allocate a BerElement for stuffing in the function IN parameters */
    requestber = ber_alloc();
    if(requestber == NULL)
        return LDAP_NO_MEMORY;

    /* stuff in the Verb, and the dn identifying the partition */
    err = ber_printf(requestber, "s", dn);
    if (err == LBER_ERROR) {
        err = LDAP_ENCODING_ERROR;
        goto cleanup;
    }

    /* Converts the BER we just built to a berval that we will send out
       in the extended request */
    if (ber_flatten(requestber, &requestbv) == LBER_ERROR) {
        err = LDAP_NO_MEMORY;
        goto cleanup;
    }

    /* Call the ldap_extended_operation (synchronously) */
    if ((err = ldap_extended_operation_s(ld, NLDAP_PARTITION_ENTRY_COUNT_EXTENDED_REQUEST, requestbv, NULL, NULL, &returnedOID, &returnedbv)) != LDAP_SUCCESS)
        goto cleanup;
   
    /* Make sure there is a returned OID */
    if (!returnedOID) {
        err = LDAP_NOT_SUPPORTED;
        goto cleanup;
    }
   
    /* Is this what we were expecting to get back */
    if (strcmp(returnedOID, NLDAP_PARTITION_ENTRY_COUNT_EXTENDED_REPLY)) {
        err = LDAP_NOT_SUPPORTED;
        goto cleanup;
    }
   
    /* Do we have a good returned berval */
    if (!returnedbv) {
        /* No returned berval means rather drastic error. Return operations error */
        err = LDAP_OPERATIONS_ERROR;
        goto cleanup;
    }

    /* Build a berElement from the returned berval */
    returnedber = ber_init(returnedbv);
    if (returnedber == NULL) {
        err = LDAP_NO_MEMORY;
        goto cleanup;
    }

    /* The returned data contains the count - scan it */
    err = ber_scanf(returnedber, "i", count);
    if (err == LBER_ERROR) {
        err = LDAP_ENCODING_ERROR;
        goto cleanup;  
    }
      else
         /* Extended request completed successfully */
         err = LDAP_SUCCESS;
 
cleanup:
    if (returnedber)
        ber_free(returnedber, 1);

    /* Free the returned OID string if one was returned */
    if (returnedOID)
        ldap_memfree(returnedOID);

    /* free any memory the extended_operation API must have allocated */
    if (returnedbv)
        ber_bvfree(returnedbv);

    /* Free memory that we allocated when building the request ber and berval */
    if (requestbv)
        ber_bvfree(requestbv);

    if (requestber)
        ber_free(requestber, 1);  

    /* return the appropriate error/success code */
    return (err);     
}

LDAP_F(int) ldap_naming_context_entry_count(LDAP* ld, char* dn, unsigned long* count)
{
    return ldap_partition_entry_count(ld, dn, count);
}

/* ------------------------------------------------------------------
ldap_abort_partition_operation
==============================

RequestBer
    Verb        INTEGER
    flags       INTEGER
    dn          OCTET STRING

ResponseBer
    NULL

Description
    This API aborts the partition operation currently in progress
    on the specified replica.
-------------------------------------------------------------------- */
LDAP_F(int) ldap_abort_partition_operation(LDAP* ld, char* dn, int flags)
{

    BerElement*    requestber = NULL;
    int            err = LDAP_SUCCESS;
    struct berval*  requestbv = NULL;

    char*       returnedOID = NULL;
    struct berval* returnedbv = NULL;

   
    /* Allocate a BerElement for stuffing in the function IN parameters */
    requestber = ber_alloc();
    if(requestber == NULL)
        return LDAP_NO_MEMORY;

    /* stuff in the Verb, and the dn identifying the partition */
    err = ber_printf(requestber, "is", flags, dn);
    if (err == LBER_ERROR) {
        err = LDAP_ENCODING_ERROR;
        goto cleanup;
    }

    /* Converts the BER we just built to a berval that we will send out
      in the extended request */
    if (ber_flatten(requestber, &requestbv) == LBER_ERROR) {
        err = LDAP_NO_MEMORY;
        goto cleanup;
    }

    /* Call the ldap_extended_operation (synchronously) */
    if ((err = ldap_extended_operation_s(ld, NLDAP_ABORT_PARTITION_OPERATION_EXTENDED_REQUEST, requestbv, NULL, NULL, &returnedOID, &returnedbv)) != LDAP_SUCCESS)
        goto cleanup;
   
    /* Make sure there is a returned OID */
    if (!returnedOID) {
        err = LDAP_NOT_SUPPORTED;
        goto cleanup;
    }
   
    /* Is this what we were expecting to get back */
    if (strcmp(returnedOID, NLDAP_ABORT_PARTITION_OPERATION_EXTENDED_REPLY))
        err = LDAP_NOT_SUPPORTED;

      
    /* NO RETURNED BERVAL FOR THIS API */

cleanup:
    /* Free the returned OID string if one was returned */
    if (returnedOID)
        ldap_memfree(returnedOID);

    /* free any memory the extended_operation API must have allocated */
    if (returnedbv)
        ber_bvfree(returnedbv);

    /* Free memory that we allocated when building the request ber and berval */
    if (requestbv)
        ber_bvfree(requestbv);

    if (requestber)
        ber_free(requestber, 1);

    /* return the appropriate error/success code */
    return (err);     
}

LDAP_F(int) ldap_abort_naming_context_operation(LDAP* ld, char* dn, int flags)
{
    return ldap_abort_partition_operation(ld, dn, flags);
}

/* ------------------------------------------------------------------
ldap_change_replica_type
==============================

RequestBer
    Verb        INTEGER
    flags       INTEGER
    replicaType INTEGER
    serverName  OCTET STRING
    dn          OCTET STRING

ResponseBer
    NULL

Description
    This API changes the type of the repica identified by dn
    to the new type specified by replicaType.
-------------------------------------------------------------------- */
LDAP_F(int) ldap_change_replica_type(LDAP* ld, char* dn, char* serverName, LDAP_REPLICA_TYPE replicaType, int flags)
{

    BerElement*    requestber = NULL;
    int            err = LDAP_SUCCESS;
    struct berval*  requestbv = NULL;

    char*       returnedOID = NULL;
    struct berval* returnedbv = NULL;

   
    /* Allocate a BerElement for stuffing in the function IN parameters */
    requestber = ber_alloc();
    if(requestber == NULL)
        return LDAP_NO_MEMORY;

    /* stuff in the Verb, and the dn identifying the partition */
    err = ber_printf(requestber, "iiss", flags, replicaType, serverName, dn);
    if (err == LBER_ERROR) {
        err = LDAP_ENCODING_ERROR;
        goto cleanup;
    }

    /* Converts the BER we just built to a berval that we will send out
       in the extended request */
    if (ber_flatten(requestber, &requestbv) == LBER_ERROR) {
        err = LDAP_NO_MEMORY;
        goto cleanup;
    }

    /* Call the ldap_extended_operation (synchronously) */
    if ((err = ldap_extended_operation_s(ld, NLDAP_CHANGE_REPLICA_TYPE_EXTENDED_REQUEST, requestbv, NULL, NULL, &returnedOID, &returnedbv)) != LDAP_SUCCESS)
        goto cleanup;
   
    /* Make sure there is a returned OID */
    if (!returnedOID) {
        err = LDAP_NOT_SUPPORTED;
        goto cleanup;
    }
   
    /* Is this what we were expecting to get back */
    if (strcmp(returnedOID, NLDAP_CHANGE_REPLICA_TYPE_EXTENDED_REPLY))
        err = LDAP_NOT_SUPPORTED;

        /* NO RETURNED BERVAL FOR THIS API */

cleanup:
    /* Free the returned OID string if one was returned */
    if (returnedOID)
        ldap_memfree(returnedOID);

    /* free any memory the extended_operation API must have allocated */
    if (returnedbv)
        ber_bvfree(returnedbv);

    /* Free memory that we allocated when building the request ber and berval */
    if (requestbv)
        ber_bvfree(requestbv);

    if (requestber)
        ber_free(requestber, 1); 

    /* return the appropriate error/success code */
    return (err);  
}


/* ------------------------------------------------------------------
ldap_get_replica_info
==============================

RequestBer
    Verb            INTEGER
    infoFlags       INTEGER
    serverName      OCTET STRING
    dn              OCTET STRING

ResponseBer
    partitionID,      INTEGER
    replicaState      INTEGER
    modificationTime  INTEGER
    purgeTime         INTEGER
    localPartitionID  INTEGER
    partitionDN       OCTET STRING
    replicaType       INTEGER
    flags            INTEGER

Description
    This API returns other interesting information about the replica
    identified by dn and residing on the server identified by 
    serverName.
-------------------------------------------------------------------- */
LDAP_F(int) ldap_get_replica_info(LDAP* ld, char* dn, char* serverDN, LDAPReplicaInfo* replicaInfo)
{
    BerElement*    requestber = NULL;
    BerElement*     returnedber = NULL;
    int            err = LDAP_SUCCESS;
    struct berval*  requestbv = NULL;
    char*       partDN = NULL;

    char*       returnedOID = NULL;
    struct berval* returnedbv = NULL;

    /* Allocate a BerElement for stuffing in the function IN parameters */
    requestber = ber_alloc();
    if(requestber == NULL)
        return LDAP_NO_MEMORY;

    /* stuff in the Verb, the serverName and the dn identifying the partition */
    err = ber_printf(requestber, "ss", serverDN, dn);
    if (err == LBER_ERROR) {
        err = LDAP_ENCODING_ERROR;
        goto cleanup;
    }

    /* Converts the BER we just built to a berval that we will send out
       in the extended request */
    if (ber_flatten(requestber, &requestbv) == LBER_ERROR) {
        err = LDAP_NO_MEMORY;
        goto cleanup;
    }

    /* Call the ldap_extended_operation (synchronously) */
    if ((err = ldap_extended_operation_s(ld, NLDAP_GET_REPLICA_INFO_EXTENDED_REQUEST, requestbv, NULL, NULL, &returnedOID, &returnedbv)) != LDAP_SUCCESS)
        goto cleanup;
   
    /* Make sure there is a returned OID */
    if (!returnedOID) {
        err = LDAP_NOT_SUPPORTED;
        goto cleanup;
    }
   
    /* Is this what we were expecting to get back */
    if (strcmp(returnedOID, NLDAP_GET_REPLICA_INFO_EXTENDED_REPLY)) {
        err = LDAP_NOT_SUPPORTED;
        goto cleanup;
    }

    /* Do we have a good returned berval */
    if (!returnedbv) {
        /* No returned berval means rather drastic error. Return operations error */
        err = LDAP_OPERATIONS_ERROR;
        goto cleanup;
    }


    /* Build a berElement from the returned berval */
    returnedber = ber_init(returnedbv);
    if (returnedber == NULL) {
        err = LDAP_NO_MEMORY;
        goto cleanup;
    }

    /* extract the returned information into the structure that will hold
     the replica info*/
    err = ber_scanf(returnedber, "iiiiiaii",
        &(replicaInfo->rootID),
        &(replicaInfo->state),
        &(replicaInfo->modificationTime),
        &(replicaInfo->purgeTime),
        &(replicaInfo->localReplicaID),
        &partDN,
        &(replicaInfo->replicaType),
        &(replicaInfo->flags));

    if (err == LBER_ERROR) {
        err = LDAP_ENCODING_ERROR;
        goto cleanup;  
    }
    else  {
        /* Extended request completed successfully */
        err = LDAP_SUCCESS;

        /*Copy the dn of the partition into the structure as well */
        if (partDN)
           strcpy(replicaInfo->namingContextDN, partDN);
      
    }

cleanup:

    if (returnedber)
        ber_free(returnedber, 1);

    if (partDN)
        ldap_memfree(partDN);

    /* Free the returned OID string if one was returned */
    if (returnedOID)
        ldap_memfree(returnedOID);

    /* free any memory the extended_operation API must have allocated */
    if (returnedbv)
        ber_bvfree(returnedbv);

   /* Free memory that we allocated when building the request ber and berval */
   if (requestbv)
       ber_bvfree(requestbv);

    if (requestber)
        ber_free(requestber, 1); 

    /* return the appropriate error/success code */
    return (err);
}

/* ------------------------------------------------------------------
ldap_receive_all_updates
==============================

RequestBer
    Verb        INTEGER
    partitionRoot  OCTET STRING
    toServerDN     OCTET STRING
    fromServerDN   OCTET STRING

ResponseBer
     NULL

Description
    This API schedules an update from server "toServerND" to "fromServerDN".
-------------------------------------------------------------------- */
LDAP_F(int) ldap_receive_all_updates(LDAP *ld, char* partitionRoot, char* toServerDN, char* fromServerDN)
{
    BerElement*    requestber = NULL;
    int            err = LDAP_SUCCESS;
    struct berval*  requestbv = NULL;

    char*       returnedOID = NULL;
    struct berval* returnedbv = NULL;

    /* Required parameters */
    if (!partitionRoot || !toServerDN)
        return LDAP_PARAM_ERROR;

    /* This parameter is not currently used, but can't be NULL for ber_printf. */
    if (!fromServerDN)
        fromServerDN = "unusedParam";

    /* Allocate a BerElement for stuffing in the function IN parameters */
    requestber = ber_alloc();
    if(requestber == NULL)
        return LDAP_NO_MEMORY;
        
    /* Stuff in the Verb, the serverName and the dns identifying the servers */
    err = ber_printf(requestber, "sss", partitionRoot, toServerDN, fromServerDN);

    if (err == LBER_ERROR) {
        err = LDAP_ENCODING_ERROR;
        goto cleanup;
    }

    /* Converts the BER we just built to a berval that we will send out
      in the extended request */
    if (ber_flatten(requestber, &requestbv) == LBER_ERROR) {
        err = LDAP_NO_MEMORY;
        goto cleanup;
    }

    /* Call the ldap_extended_operation (synchronously) */
    if ((err = ldap_extended_operation_s(ld, NLDAP_RECEIVE_ALL_UPDATES_EXTENDED_REQUEST, requestbv, NULL, NULL, &returnedOID, &returnedbv)) != LDAP_SUCCESS)
        goto cleanup;

    /* Make sure there is a returned OID */
    if (!returnedOID) {
        err = LDAP_NOT_SUPPORTED;
        goto cleanup;
    }
   
    /* Is this what we were expecting to get back */
    if (strcmp(returnedOID, NLDAP_RECEIVE_ALL_UPDATES_EXTENDED_REPLY))
        err = LDAP_NOT_SUPPORTED;
   
    /* NO RETURNED BERVAL FOR THIS API */

cleanup:
    /* Free the returned OID string if one was returned */
    if (returnedOID)
        ldap_memfree(returnedOID);

    /* free any memory the extended_operation API must have allocated */
    if (returnedbv)
        ber_bvfree(returnedbv);

    /* Free memory that we allocated when building the request ber and berval */
    if (requestbv)
        ber_bvfree(requestbv);

    if(requestber)
        ber_free(requestber, 1);

    /* return the appropriate error/success code */
    return (err);
}

/* ------------------------------------------------------------------
ldap_send_all_updates
==============================

RequestBer
    Verb        INTEGER
    partitionRoot  OCTET STRING
    origServerDN   OCTET STRING

ResponseBer
    NULL

Description
    This API schedules an update for the replica ring rooted at
    partitionRoot.
-------------------------------------------------------------------- */
LDAP_F(int) ldap_send_all_updates(LDAP *ld, char* partitionRoot, char* origServerDN)
{
    BerElement*    requestber = NULL;
    int            err = LDAP_SUCCESS;
    struct berval*  requestbv = NULL;

    char*       returnedOID = NULL;
    struct berval* returnedbv = NULL;

   
    /* Allocate a BerElement for stuffing in the function IN parameters */
    requestber = ber_alloc();
    if(requestber == NULL)
        return LDAP_NO_MEMORY;

    /* Parameter is not currently used.  Allow NULL */
    if (origServerDN == NULL)
        origServerDN = "";

    /* stuff in the Verb, the serverName and the dn identifying the partition Root */
    err = ber_printf(requestber, "ss", partitionRoot, origServerDN);
    if (err == LBER_ERROR) {
        err = LDAP_ENCODING_ERROR;
        goto cleanup;
    }

    /* Converts the BER we just built to a berval that we will send out
      in the extended request */
    if (ber_flatten(requestber, &requestbv) == LBER_ERROR) {
        err = LDAP_NO_MEMORY;
        goto cleanup;
    }

    /* Call the ldap_extended_operation (synchronously) */
    if ((err = ldap_extended_operation_s(ld, NLDAP_SEND_ALL_UPDATES_EXTENDED_REQUEST, requestbv, NULL, NULL, &returnedOID, &returnedbv)) != LDAP_SUCCESS)
        goto cleanup;
   
    /* Make sure there is a returned OID */
    if (!returnedOID) {
        err = LDAP_NOT_SUPPORTED;
        goto cleanup;
    }
   
    /* Is this what we were expecting to get back */
    if (strcmp(returnedOID, NLDAP_SEND_ALL_UPDATES_EXTENDED_REPLY))
        err = LDAP_NOT_SUPPORTED;

       
    /* NO RETURNED BERVAL FOR THIS API */

cleanup:
    /* Free the returned OID string if one was returned */
    if (returnedOID)
        ldap_memfree(returnedOID);
 
    /* free any memory the extended_operation API must have allocated */
    if (returnedbv)
        ber_bvfree(returnedbv);

    /* Free memory that we allocated when building the request ber and berval */
    if (requestbv)
        ber_bvfree(requestbv);

    if(requestber)
        ber_free(requestber, 1);

    /* return the appropriate error/success code */
    return (err); 
}


/* ------------------------------------------------------------------
ldap_request_partition_sync
==============================

RequestBer
    Verb            INTEGER
    serverName      OCTET STRING
    partitionRoot   OCTET STRING
    delay           INTEGER

ResponseBer
    NULL

Description
    This API schedules an update for the replica ring rooted at
    partitionRoot.
-------------------------------------------------------------------- */

LDAP_F(int) ldap_request_partition_sync(LDAP *ld, char* serverName, char* partitionRoot, int delay)
{

    BerElement*    requestber = NULL;
    int            err = LDAP_SUCCESS;
    struct berval*  requestbv = NULL;

    char*       returnedOID = NULL;
    struct berval* returnedbv = NULL;

    /* Allocate a BerElement for stuffing in the function IN parameters */
    requestber = ber_alloc();
    if(requestber == NULL)
        return LDAP_NO_MEMORY;

    /* stuff in the Verb, the serverName and the dn identifying the partition Root */
    err = ber_printf(requestber, "ssi", serverName, partitionRoot, delay);
    if (err == LBER_ERROR) {
        err = LDAP_ENCODING_ERROR;
        goto cleanup;
    }

    /* Converts the BER we just built to a berval that we will send out
      in the extended request */
    if (ber_flatten(requestber, &requestbv) == LBER_ERROR) {
        err = LDAP_NO_MEMORY;
        goto cleanup;
    }

    /* Call the ldap_extended_operation (synchronously) */
    if ((err = ldap_extended_operation_s(ld, NLDAP_REQUEST_PARTITION_SYNC_EXTENDED_REQUEST, requestbv, NULL, NULL, &returnedOID, &returnedbv)) != LDAP_SUCCESS)
        goto cleanup;
   
    /* Make sure there is a returned OID */
    if (!returnedOID) {
        err = LDAP_NOT_SUPPORTED;
        goto cleanup;
    }
   
    /* Is this what we were expecting to get back */
    if (strcmp(returnedOID, NLDAP_REQUEST_PARTITION_SYNC_EXTENDED_REPLY))
        err = LDAP_NOT_SUPPORTED;

       /* NO RETURNED BERVAL FOR THIS API */

cleanup:
   /* Free the returned OID string if one was returned */
    if (returnedOID)
        ldap_memfree(returnedOID);

    /* free any memory the extended_operation API must have allocated */
    if (returnedbv)
        ber_bvfree(returnedbv);

    /* Free memory that we allocated when building the request ber and berval */
    if (requestbv)
        ber_bvfree(requestbv);

    if(requestber)
        ber_free(requestber, 1); 

    /* return the appropriate error/success code */
    return (err); 
}

LDAP_F(int) ldap_request_naming_context_sync(LDAP *ld, char* serverName, char* partitionRoot, int delay)
{
    return ldap_request_partition_sync(ld, serverName, partitionRoot, delay);
}

/* ------------------------------------------------------------------
ldap_request_schema_sync
==============================

RequestBer
    Verb            INTEGER
    partitionDN     OCTET STRING
    delay       INTEGER

ResponseBer
    NULL

Description
    This API schedules an schema update to occur between the servers
    in the specified parititon ring after a period of delay seconds.
-------------------------------------------------------------------- */
LDAP_F(int) ldap_request_schema_sync(LDAP*ld, char* partitionDN, int delay)
{

   
    BerElement*    requestber = NULL;
    int            err = LDAP_SUCCESS;
    struct berval*  requestbv = NULL;

    char*       returnedOID = NULL;
    struct berval* returnedbv = NULL;

    /* Allocate a BerElement for stuffing in the function IN parameters */
    requestber = ber_alloc();
    if(requestber == NULL)
        return LDAP_NO_MEMORY;

    /* stuff in the Verb, the partition DN and the delay*/
    err = ber_printf(requestber, "si", partitionDN, delay);
    if (err == LBER_ERROR) {
        err = LDAP_ENCODING_ERROR;
        goto cleanup;
    }

    /* Converts the BER we just built to a berval that we will send out
       in the extended request */
    if (ber_flatten(requestber, &requestbv) == LBER_ERROR) {
        err = LDAP_NO_MEMORY;
        goto cleanup;
    }

    /* Call the ldap_extended_operation (synchronously) */
    if ((err = ldap_extended_operation_s(ld, NLDAP_REQUEST_SCHEMA_SYNC_EXTENDED_REQUEST, requestbv, NULL, NULL, &returnedOID, &returnedbv)) != LDAP_SUCCESS)
        goto cleanup;
   
    /* Make sure there is a returned OID */
    if (!returnedOID) {
        err = LDAP_NOT_SUPPORTED;
        goto cleanup;
    }
   
    /* Is this what we were expecting to get back */
    if (strcmp(returnedOID, NLDAP_REQUEST_SCHEMA_SYNC_EXTENDED_REPLY))
        err = LDAP_NOT_SUPPORTED;
   
       /* NO RETURNED BERVAL FOR THIS API */

cleanup:
    /* Free the returned OID string if one was returned */
    if (returnedOID)
        ldap_memfree(returnedOID);

    /* free any memory the extended_operation API must have allocated */
    if (returnedbv)
        ber_bvfree(returnedbv);

    /* Free memory that we allocated when building the request ber and berval */
    if (requestbv)
        ber_bvfree(requestbv);

    if(requestber)
        ber_free(requestber, 1);      // Free data in the ber as we caused it to be allocated

    /* return the appropriate error/success code */
    return (err);
}


/* ------------------------------------------------------------------
ldap_list_replicas
==============================

RequestBer
    Verb            INTEGER
    serverDN    OCTET STRING

ResponseBer
    replicaList    SEQUENCE OF OCTET STRINGS

Description
    This API returns a list of replicas that reside on the specified
    server.
-------------------------------------------------------------------- */
LDAP_F(int) ldap_list_replicas(LDAP *ld, char *serverDN, char*** replicaList)
{
    BerElement*    requestber = NULL;
    BerElement*     returnedber = NULL;
    int            err = LDAP_SUCCESS;
    struct berval*  requestbv = NULL;

    char*       returnedOID = NULL ;
    struct berval* returnedbv = NULL;

   
    /* Allocate a BerElement for stuffing in the function IN parameters */
    requestber = ber_alloc();
    if(requestber == NULL)
        return LDAP_NO_MEMORY;

    /* stuff in the Verb, the server DN */
    err = ber_printf(requestber, "s", serverDN);
    if (err == LBER_ERROR) {
        err = LDAP_ENCODING_ERROR;
        goto cleanup;
    }

    /* Converts the BER we just built to a berval that we will send out
       in the extended request */
    if (ber_flatten(requestber, &requestbv) == LBER_ERROR) {
        err = LDAP_NO_MEMORY;
        goto cleanup;
    }

    /* Call the ldap_extended_operation (synchronously) */
    if ((err = ldap_extended_operation_s(ld, NLDAP_LIST_REPLICAS_EXTENDED_REQUEST, requestbv, NULL, NULL, &returnedOID, &returnedbv)) != LDAP_SUCCESS)
        goto cleanup;
   
    /* Make sure there is a returned OID */
    if (!returnedOID) {
        err = LDAP_NOT_SUPPORTED;
        goto cleanup;
    }
   
    /* Is this what we were expecting to get back */
    if (strcmp(returnedOID, NLDAP_LIST_REPLICAS_EXTENDED_REPLY)) {
        err = LDAP_NOT_SUPPORTED;
        goto cleanup;
    }

    /* Do we have a good returned berval */
    if (!returnedbv) {
        /* No returned berval means rather drastic error. Return operations error */
        err = LDAP_OPERATIONS_ERROR;
        goto cleanup;
    }

    /* Create a BerElement from the berval returned by the extended call */
    returnedber = ber_init(returnedbv);
    if (returnedber == NULL) {
        err = LDAP_NO_MEMORY;
        goto cleanup;
     }

     /* Extract the replica list as a sequence of string vectors */
     err = ber_scanf(returnedber, "v", replicaList);
     if (err == LBER_ERROR) {
        err = LDAP_ENCODING_ERROR;
        goto cleanup;
     }
     else 
        /* Extended request completed successfully */
        err = LDAP_SUCCESS;
 
cleanup:
    if (returnedber)
        ber_free(returnedber, 1);

    /* Free the returned OID string if one was returned */
    if (returnedOID)
        ldap_memfree(returnedOID);

    /* free any memory the extended_operation API must have allocated */
    if (returnedbv)
        ber_bvfree(returnedbv);

    /* Free memory that we allocated when building the request ber and berval */
    if (requestbv)
        ber_bvfree(requestbv);

   if(requestber)
       ber_free(requestber, 1);      // Free data in the ber as we caused it to be allocated

    /* return the appropriate error/success code */
    return (err); 
}

/* ------------------------------------------------------------------
ldap_set_replication_filter
==============================

RequestBer
    serverName  OCTET STRING
    SEQUENCE of SEQUENCE {
        classname   OCTET STRING
        SEQUENCE of ATTRIBUTES
        }

where
  ATTRIBUTES:: OCTET STRING

ResponseBer
        NULL
Description

    This API is used to set the attribute and class filter on NDS 
    Virtual Replica.  The serverDN identifies the server on which
    the filter needs to be set.  With NDS 8.5 these filters are set
    on a per server basis.  The filter identifies the classes and
    attributes that comprise the filter and has the following format:
     
    Example Filter:
    class1$attr1$attr2$attr3$$class2$attr1$$$
-------------------------------------------------------------------- */
LDAP_F(int) ldap_set_replication_filter(LDAP* ld, const char* serverDN, const char* filter)
{

    BerElement*    requestber = NULL;
    int            err = LDAP_SUCCESS;
    struct berval*  requestbv = NULL;

    char*       returnedOID = NULL;
    struct berval* returnedbv = NULL;

    char *freePtr = NULL;
    char *currentPtr = NULL;
    char *startPtr = NULL;

    // Crashes ber_printf if NULL
    if ( (!serverDN) || (!filter) )
        return LDAP_PARAM_ERROR;

    /* Allocate a BerElement for stuffing in the function IN parameters */
    requestber = ber_alloc_t(0);
    if(requestber == NULL)
        return LDAP_NO_MEMORY;

    /* Add the server name to the ber encoding */
    err = ber_printf(requestber, "s{", serverDN);
    if (err == LBER_ERROR) {
        err = LDAP_ENCODING_ERROR;
        goto cleanup;
    }
    
    freePtr = (char*) malloc (strlen(filter) + 1);
    
    if (!freePtr)
        goto cleanup;

    strcpy(freePtr, filter);
    currentPtr = freePtr;

    // At the beginning of this loop currentPtr should be pointed
    // to the start of a class name
    while ( (currentPtr) && (*currentPtr != FILTER_SEP) ){
        // Point to the beginning of the class name
        startPtr = currentPtr;

        // Find the end of the class name
        currentPtr = strchr( currentPtr, FILTER_SEP);
        
        if (!currentPtr) {
            err = LDAP_PARAM_ERROR;
            goto cleanup;
        }

        // Null terminate the classname
        *currentPtr = '\0'; 
        
        // startPtr now points to a null terminated class name
        // ADd classname to the ber structure
        err = ber_printf(requestber, "{s{", startPtr);
        if (err == LBER_ERROR) {
            err = LDAP_ENCODING_ERROR;
            goto cleanup;
        }

        // This is where we parse the attributes
        currentPtr++;

        while (*currentPtr != FILTER_SEP) {
            // Point to the beginning of the attribute name     
            startPtr = currentPtr;

            // Find the end of the attribute name
            currentPtr = strchr( currentPtr, FILTER_SEP);
            if (!currentPtr) {
                err = LDAP_PARAM_ERROR;
                goto cleanup;
            }
            
            // Null terminate the classname
            *currentPtr = '\0';

            // startPtr now points to a null terminated attribute name
            // Add the attribute name to the ber
            err = ber_printf(requestber, "s", startPtr);
            if (err == LBER_ERROR) {
                err = LDAP_ENCODING_ERROR;
                goto cleanup;
            }
            // Point to the start of the next attribute.  If this is the
            // last attribute we are pointed to a seperator and the while
            // loop will quit before the next iteration
            currentPtr++;
        }


        // Done parsing all the attributes so close the sequence
        err = ber_printf(requestber, "}}");
        if (err == LBER_ERROR) {
            err = LDAP_ENCODING_ERROR;
            goto cleanup;
        }

        // make sure currentPtr is pointed to the beginning of the next class
        currentPtr++;

    }

    // All done, close the sequence
    err = ber_printf(requestber, "}");
    if (err == LBER_ERROR) {
        err = LDAP_ENCODING_ERROR;
        goto cleanup;
    }

    /* Converts the BER we just built to a berval that we will send out
       in the extended request */
    if (ber_flatten(requestber, &requestbv) == LBER_ERROR) {
        err = LDAP_NO_MEMORY;
        goto cleanup;
    }

    /* Call the ldap_extended_operation (synchronously) */
    if ((err = ldap_extended_operation_s(ld, 
                    NLDAP_SET_REPLICA_FILTER_EXTENDED_REQUEST, 
                    requestbv, 
                    NULL, 
                    NULL, 
                    &returnedOID, 
                    &returnedbv)) != LDAP_SUCCESS)
        goto cleanup;
   
    /* Make sure there is a returned OID */
    if (!returnedOID) {
        err = LDAP_NOT_SUPPORTED;
        goto cleanup;
    }
   
    /* Is this what we were expecting to get back */
    if (strcmp(returnedOID, NLDAP_SET_REPLICA_FILTER_EXTENDED_REPLY))
        err = LDAP_NOT_SUPPORTED;

    /* NO RETURNED BERVAL FOR THIS API */

cleanup:
    // If we duplicate the filter string free it
    if (freePtr)
        free(freePtr);

    /* Free the returned OID string if one was returned */
    if (returnedOID)
        ldap_memfree(returnedOID);

    /* free any memory the extended_operation API must have allocated */
    if (returnedbv)
        ber_bvfree(returnedbv);

    /* Free memory that we allocated when building the request ber and berval */
    if (requestbv)
        ber_bvfree(requestbv);

    if (requestber)
        ber_free(requestber, 1); 

    /* return the appropriate error/success code */
    return (err);  
}


/* ------------------------------------------------------------------
ldap_get_replication_filter
==============================

RequestBer
    serverName  OCTET STRING

ResponseBer 
  SEQUENCE of SEQUENCE {
        classname   OCTET STRING
        SEQUENCE of ATTRIBUTES
        }
where
  ATTRIBUTES:: OCTET STRING

Description
    This API is used to get the attribute and class filter on an 
    existing NDS Server.  The filter has the same format as that in 
    the set_replication_filter call. The caller must free filter by 
    calling ldapx_memfree
-------------------------------------------------------------------- */
LDAP_F(int) ldap_get_replication_filter(LDAP* ld, const char* serverDN, char** filter)
{

    BerElement      *requestber = NULL;
    BerElement      *returnedber = NULL;
    int             err = LDAP_SUCCESS;
    struct berval   *requestbv = NULL;
    ber_len_t       dummyLen = 0;

    char            *returnedOID = NULL;
    struct berval   *returnedbv = NULL;
    struct berval   *filterbv = NULL;
    char            **attrList = NULL;
    char            *className = NULL;
    const char      *SEPERATOR_CHAR = "$";

    int             count = 0;

    // Crashes ber_printf if NULL
    if ( (!serverDN) || (!filter) )
        return LDAP_PARAM_ERROR;

    /* Allocate a BerElement for stuffing in the function IN parameters */
    requestber = ber_alloc_t(0);
    if(requestber == NULL)
        return LDAP_NO_MEMORY;

    /* Add the server name to the ber structure */
    err = ber_printf(requestber, "s", serverDN);
    if (err == LBER_ERROR) {
        err = LDAP_ENCODING_ERROR;
        goto cleanup;
    }

    /* Converts the BER we just built to a berval that we will send out
       in the extended request */
    if (ber_flatten(requestber, &requestbv) == LBER_ERROR) {
        err = LDAP_NO_MEMORY;
        goto cleanup;
    }

    /* Call the ldap_extended_operation (synchronously) */
    if ((err = ldap_extended_operation_s(ld, 
                    NLDAP_GET_REPLICA_FILTER_EXTENDED_REQUEST, 
                    requestbv, 
                    NULL, 
                    NULL, 
                    &returnedOID, 
                    &returnedbv)) != LDAP_SUCCESS)
        goto cleanup;
   
    /* Make sure there is a returned OID */
    if (!returnedOID) {
        err = LDAP_NOT_SUPPORTED;
        goto cleanup;
    }
   
    /* Is this what we were expecting to get back */
    if (strcmp(returnedOID, NLDAP_GET_REPLICA_FILTER_EXTENDED_REPLY))
        err = LDAP_NOT_SUPPORTED;

    /* Do we have a good returned berval */
    if (!returnedbv) {
        
        /* No returned berval means rather drastic error. Return operations error */
        err = LDAP_OPERATIONS_ERROR;
        goto cleanup;   
    }

    /* Build a berElement from the returned berval */
    returnedber = ber_init(returnedbv);
    if (returnedber == NULL) {
        err = LDAP_DECODING_ERROR;
        goto cleanup;
    }

    // Allocate memory to hold the filter - Note that the resulting
    // filter cannot be any larger than the returned berval
    *filter = (char *)malloc( returnedbv->bv_len + 2);
    strcpy(*filter, "");

    // Extract the opening sequence
    err = ber_scanf(returnedber, "{");
    if (err == LBER_ERROR) {
        err = LDAP_DECODING_ERROR;
        goto cleanup;     
    }
        
    while (1) {

        // Extract the first sequence of classnames followed by attrname vector
        err = ber_scanf(returnedber, "{a{v}}", &className, &attrList);
        if (err == LBER_ERROR) {
            if ( ber_peek_tag(returnedber, &dummyLen) == LBER_DEFAULT) {
                // Was this error because we reached the end of the buffer.
                // In that case we break normally and $ terminate filter
                break;
            }
            else {
                // Otherwise we had a real error condition
                err = LDAP_DECODING_ERROR;
                goto cleanup;
            }
        }
            
        // copy the classname and terminate with $
        strcat(*filter, className);
        strcat(*filter, SEPERATOR_CHAR);
        
        if (NULL != attrList)
        {
            count = 0;
            while (attrList[count] != NULL) {
                strcat(*filter, attrList[count]);
                strcat(*filter, SEPERATOR_CHAR);
                count++;
            }
        }

        // close the class sequence with $
        strcat(*filter, SEPERATOR_CHAR);

        if (attrList)
            ldap_value_free(attrList);
         
        if (className)
            ldap_memfree(className);

    
    }
        
    // close the filter with $
    strcat(*filter, SEPERATOR_CHAR);
    
    err = LDAP_SUCCESS;

cleanup:
    /* Free the returned OID string if one was returned */
    if (returnedOID)
        ldap_memfree(returnedOID);

    /* free any memory the extended_operation API must have allocated */
    if (returnedbv)
        ber_bvfree(returnedbv);

    /* Free memory that we allocated when building the request ber and berval */
    if (requestbv)
        ber_bvfree(requestbv);

    if (requestber)
        ber_free(requestber, 1); 

    if (returnedber)
        ber_free(returnedber, 1);

    /* return the appropriate error/success code */
    return (err);  
}




/* ------------------------------------------------------------------
ldap_split_orphan_partition
==============================

RequestBer
    serverDN    OCTET STRING
    contextName OCTET STRING

ResponseBer
    NULL

Description
  This API is used to split an orphan partition on the specified server
-------------------------------------------------------------------- */
LDAP_F(int) ldap_split_orphan_partition(LDAP* ld, char* serverDN, char* contextName)
{
    BerElement*    requestber = NULL;
    int            err = LDAP_SUCCESS;
    struct berval*  requestbv = NULL;

    char*       returnedOID = NULL;
    struct berval* returnedbv = NULL;
   
    /* Allocate a BerElement for stuffing in the function IN parameters */
    requestber = ber_alloc();
    if(requestber == NULL)
        return LDAP_NO_MEMORY;

    /* stuff in the Verb, the replica type, flags and the dn's of the server and replica*/
    err = ber_printf(requestber, "ss", serverDN, contextName);
    if (err == LBER_ERROR) {
        err = LDAP_ENCODING_ERROR;
        goto cleanup;
    }

    /* Converts the BER we just built to a berval that we will send out
      in the extended request */
    if (ber_flatten(requestber, &requestbv) == LBER_ERROR) {
        err = LDAP_NO_MEMORY;
        goto cleanup;
    }

    /* Call the ldap_extended_operation (synchronously) */
    if ((err = ldap_extended_operation_s(ld, NLDAP_CREATE_ORPHAN_PARTITION_EXTENDED_REQUEST, requestbv, NULL, NULL, &returnedOID, &returnedbv)) != LDAP_SUCCESS)
        goto cleanup;

    /* Make sure there is a returned OID */
    if (!returnedOID) {
        err = LDAP_NOT_SUPPORTED;
        goto cleanup;
    }

    /* Is this what we were expecting to get back */
    if (strcmp(returnedOID, NLDAP_CREATE_ORPHAN_PARTITION_EXTENDED_REPLY))
        err = LDAP_NOT_SUPPORTED;

    /* NO RETURNED BERVAL FOR THIS API */

cleanup:
    /* Free the returned OID string if one was returned */
    if (returnedOID)
        ldap_memfree(returnedOID);

    /* free any memory the extended_operation API must have allocated */
    if (returnedbv)
        ber_bvfree(returnedbv);

   /* Free memory that we allocated when building the request ber and berval */
    if (requestbv)
        ber_bvfree(requestbv);

    if (requestber)
        ber_free(requestber, 1); 

    /* return the appropriate error/success code */
    return (err);
}

LDAP_F(int) ldap_create_orphan_naming_context(LDAP* ld, char* serverDN, char* contextName)
{
    return ldap_split_orphan_partition(ld, serverDN, contextName);
}
/* ------------------------------------------------------------------
ldap_remove_orphan_partition
==============================

RequestBer
    serverDN    OCTET STRING
    contextName OCTET STRING

ResponseBer
    NULL

Description
  This API is used to remove an orphan partition from the specified server
-------------------------------------------------------------------- */
LDAP_F(int) ldap_remove_orphan_partition(LDAP* ld, char* serverDN, char* contextName)
{
    BerElement*    requestber = NULL;
    int            err = LDAP_SUCCESS;
    struct berval*  requestbv = NULL;

    char*       returnedOID = NULL;
    struct berval* returnedbv = NULL;
   
    /* Allocate a BerElement for stuffing in the function IN parameters */
    requestber = ber_alloc();
    if(requestber == NULL)
        return LDAP_NO_MEMORY;

    /* stuff in the Verb, the replica type, flags and the dn's of the server and replica*/
    err = ber_printf(requestber, "ss", serverDN, contextName);
    if (err == LBER_ERROR) {
        err = LDAP_ENCODING_ERROR;
        goto cleanup;
    }

    /* Converts the BER we just built to a berval that we will send out
      in the extended request */
    if (ber_flatten(requestber, &requestbv) == LBER_ERROR) {
        err = LDAP_NO_MEMORY;
        goto cleanup;
    }

    /* Call the ldap_extended_operation (synchronously) */
    if ((err = ldap_extended_operation_s(ld, NLDAP_REMOVE_ORPHAN_PARTITION_EXTENDED_REQUEST, requestbv, NULL, NULL, &returnedOID, &returnedbv)) != LDAP_SUCCESS)
        goto cleanup;

    /* Make sure there is a returned OID */
    if (!returnedOID) {
        err = LDAP_NOT_SUPPORTED;
        goto cleanup;
    }

    /* Is this what we were expecting to get back */
    if (strcmp(returnedOID, NLDAP_REMOVE_ORPHAN_PARTITION_EXTENDED_REPLY))
        err = LDAP_NOT_SUPPORTED;

    /* NO RETURNED BERVAL FOR THIS API */

cleanup:
    /* Free the returned OID string if one was returned */
    if (returnedOID)
        ldap_memfree(returnedOID);

    /* free any memory the extended_operation API must have allocated */
    if (returnedbv)
        ber_bvfree(returnedbv);

   /* Free memory that we allocated when building the request ber and berval */
    if (requestbv)
        ber_bvfree(requestbv);

    if (requestber)
        ber_free(requestber, 1); 

    /* return the appropriate error/success code */
    return (err);
}

LDAP_F(int) ldap_remove_orphan_naming_context(LDAP* ld, char* serverDN, char* contextName)
{
    return ldap_remove_orphan_partition(ld, serverDN, contextName);
}
