/* $Novell: globals.h,v 1.40 2003/05/13 13:39:03 $ 
 ******************************************************************************
 * Copyright (c) 2006 Novell, Inc.
 * All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; version 2.1 of the license.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, contact Novell, Inc.
 *
 * To contact Novell about this file by physical or electronic mail,
 * you may find current contact information at www.novell.com
 *
 *****************************************************************************
 * Copyright (C) 1999, 2000 Novell, Inc. All Rights Reserved.
 *
 * THIS WORK IS SUBJECT TO U.S. AND INTERNATIONAL COPYRIGHT LAWS AND
 * TREATIES. USE, MODIFICATION, AND REDISTRIBUTION OF THIS WORK IS SUBJECT
 * TO VERSION 2.0.1 OF THE OPENLDAP PUBLIC LICENSE, A COPY OF WHICH IS
 * AVAILABLE AT HTTP://WWW.OPENLDAP.ORG/LICENSE.HTML OR IN THE FILE "LICENSE"
 * IN THE TOP-LEVEL DIRECTORY OF THE DISTRIBUTION. ANY USE OR EXPLOITATION
 * OF THIS WORK OTHER THAN AS AUTHORIZED IN VERSION 2.0.1 OF THE OPENLDAP
 * PUBLIC LICENSE, OR OTHER PRIOR WRITTEN CONSENT FROM NOVELL, COULD SUBJECT
 * THE PERPETRATOR TO CRIMINAL AND CIVIL LIABILITY.
 *****************************************************************************/

#if !defined(GLOBALS_H)
#define GLOBALS_H

/***************************************************************************
* SECTION 1
*
*  This section contains typedef(s) used by the global variables
*
***************************************************************************/

   /* When using OpenSSL, we need an abstraction layer to combine the
    * actual socket and the SSL handle.  This ldapssl_Socket structure
    * does the abstraction.
    */
typedef struct _ldapssl_Socket
{
   ber_socket_t   fd;
   SSL           *sslHandle;
   int            inUse;

} ldapssl_Socket, *pldapssl_Socket;

#define SOCKET_EXPAND_RATE   16


#ifdef __cplusplus
extern "C" {
#endif


/***************************************************************************
* SECTION 2
*
*  Global variables 
*
***************************************************************************/


extern struct ldap_io_funcs   g_SSL_fns;
extern struct ldap_io_funcs   g_SSL_fns_no_prepare;


/********************************************************************/
/*  Windows platform specific globals                               */
/********************************************************************/
#if defined(N_PLAT_WIN32) || defined(N_PLAT_WIN64)

   /* master crypto context handle
   */
   extern SSL_CTX             *g_CryptCtx;
   extern unsigned int         g_MaxSockets;
   extern ldapssl_Socket      *g_pSockets;
   extern HANDLE              *g_Locks;
   extern void                *g_pCertCallback;
   extern int                  g_SSLReg;

   /* critical section to synchronize access to the globals
   */
   extern CRITICAL_SECTION     g_Sync;

#endif /* N_PLAT_WIN32 */


/********************************************************************/
/*  UNIX platform specific globals                                   */
/********************************************************************/
#ifdef UNIX  

//#include "portable.h"
#include <pthread.h>

   extern SSL_CTX             *g_CryptCtx;
   extern unsigned int         g_MaxSockets;
   extern ldapssl_Socket      *g_pSockets;
   extern pthread_mutex_t      g_Locks[];
   extern void                *g_pCertCallback;

   extern int                     g_SSLReg;
   extern pthread_mutex_t         g_Sync;

#endif  /* ifdef UNIX*/


/********************************************************************/
/*  Modesto platform specific globals                               */
/********************************************************************/
#if defined(MODESTO)

   extern SSL_CTX             *g_CryptCtx;
   extern unsigned int         g_MaxSockets;
   extern ldapssl_Socket      *g_pSockets;
   extern NXMutex_t          **g_Locks;
   extern void                *g_pCertCallback;
   
   extern int                  g_SSLReg;
   extern NXMutex_t           *g_Sync;

#endif /* MODESTO */


/***************************************************************************
* SECTION 3
*
*  Global function prototypes
*
***************************************************************************/


#if defined(UNIX)

  extern int g_SSLReg;

  #define SOCKET_ERROR -1
  #define INVALID_SOCKET -1
  #define SOCKET int
  #define closesocket(param) close(param)
  #define GET_SSL_REGISTERED 	g_SSLReg
  #define SET_SSL_REGISTERED     g_SSLReg = 1
  #define SET_SSL_UNREGISTERED   g_SSLReg = 0
  #define GetMasterCryptCtx()	g_CryptCtx 
  #define SetMasterCryptCtx(ctx)	g_CryptCtx = ctx
  #define GetSocketTable()	g_pSockets
  #define SetSocketTable(x)	g_pSockets = x
  #define GetMaxSockets()	g_MaxSockets
  #define SetMaxSockets(x)	g_MaxSockets = x
  #define EnterSync()       pthread_mutex_lock(&g_Sync)
  #define ExitSync()        pthread_mutex_unlock(&g_Sync)
  
  #ifdef HAVE_LINUX 
  extern void __attribute__ ((__constructor__))LockInit(void);
  extern void __attribute__ ((__destructor__))LockCleanUp(void);
  extern void LockCallback(int mode, int type, const char *file, int line);
  #endif 
  
  #ifdef HAVE_SOLARIS
  extern void LockInit(void);
  extern void LockCleanUp(void);
  extern void LockCallback(int mode, int type, const char *file, int line);
  #endif 

  #ifdef HAVE_AIX
  extern void LockInit(void);
  extern void LockCleanUp(void);
  extern void LockCallback(int mode, int type, const char *file, int line);
  #endif 

  #define GetCertCallback()        g_pCertCallback
  #define SetCertCallback(x)       g_pCertCallback = x
  unsigned char *getRandomData(int size);
  unsigned char *randbuf;

#elif defined(N_PLAT_WIN32) || defined(N_PLAT_WIN64)

   #define EnterSync()              EnterCriticalSection( &g_Sync )
   #define ExitSync()               LeaveCriticalSection( &g_Sync )
   #define GET_SSL_REGISTERED       g_SSLReg
   #define SET_SSL_REGISTERED       g_SSLReg = 1
   #define SET_SSL_UNREGISTERED     g_SSLReg = 0

   #define GetMasterCryptCtx()      g_CryptCtx
   #define SetMasterCryptCtx(ctx)   g_CryptCtx = ctx

   #define GetSocketTable()         g_pSockets
   #define SetSocketTable(x)        g_pSockets = x
   #define GetMaxSockets()          g_MaxSockets
   #define SetMaxSockets(x)         g_MaxSockets = x

   extern int LockInit(void);
   extern void LockCleanUp(void);
   extern void LockCallback(int mode, int type, const char *file, int line);
   #define GetCertCallback()        g_pCertCallback
   #define SetCertCallback(x)       g_pCertCallback = x

#elif defined(MODESTO)

   #define EnterSync()              NXLock( g_Sync )
   #define ExitSync()               NXUnlock( g_Sync )
   #define GET_SSL_REGISTERED       g_SSLReg
   #define SET_SSL_REGISTERED       g_SSLReg = 1
   #define SET_SSL_UNREGISTERED     g_SSLReg = 0

   #define GetMasterCryptCtx()      g_CryptCtx
   #define SetMasterCryptCtx(ctx)   g_CryptCtx = ctx

   #define GetSocketTable()         g_pSockets
   #define SetSocketTable(x)        g_pSockets = x
   #define GetMaxSockets()          g_MaxSockets
   #define SetMaxSockets(x)         g_MaxSockets = x

   extern int LockInit(void);
   extern void LockCleanUp(void);
   extern void LockCallback(int mode, int type, const char *file, int line);
   #define GetCertCallback()        g_pCertCallback
   #define SetCertCallback(x)       g_pCertCallback = x

#endif /* defined(UNIX) */


#ifdef __cplusplus
}
#endif

#endif /* !defined(GLOBALS_H) */


