/* $Novell: pkcs12x.c,v 1.5 2002/09/18 21:40:14 $ 
 ******************************************************************************
 * Copyright (c) 2006 Novell, Inc.
 * All Rights Reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; version 2.1 of the license.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, contact Novell, Inc.
 *
 * To contact Novell about this file by physical or electronic mail,
 * you may find current contact information at www.novell.com
 *
 *****************************************************************************
 * Copyright (C) 1999, 2000 Novell, Inc. All Rights Reserved.
 * 
 * THIS WORK IS SUBJECT TO U.S. AND INTERNATIONAL COPYRIGHT LAWS AND
 * TREATIES. USE, MODIFICATION, AND REDISTRIBUTION OF THIS WORK IS SUBJECT
 * TO VERSION 2.0.1 OF THE OPENLDAP PUBLIC LICENSE, A COPY OF WHICH IS
 * AVAILABLE AT HTTP://WWW.OPENLDAP.ORG/LICENSE.HTML OR IN THE FILE "LICENSE"
 * IN THE TOP-LEVEL DIRECTORY OF THE DISTRIBUTION. ANY USE OR EXPLOITATION
 * OF THIS WORK OTHER THAN AS AUTHORIZED IN VERSION 2.0.1 OF THE OPENLDAP
 * PUBLIC LICENSE, OR OTHER PRIOR WRITTEN CONSENT FROM NOVELL, COULD SUBJECT
 * THE PERPETRATOR TO CRIMINAL AND CIVIL LIABILITY. 
  *****************************************************************************/

/****************************************************************************
*
* Description:   This file contains pkcs12 helper functions. 
*
****************************************************************************/

#include "ldap_cdefs.h"
#include "openssl/pem.h"
#include "openssl/pkcs12.h"

/* function prototypes needed by dump_certs_keys_buf
*/
int dump_certs_keys_p12
(
    BIO *out,
    PKCS12 *p12,
    char *pass
);
int dump_certs_pkeys_bags(
    BIO *out,
    STACK_OF(PKCS12_SAFEBAG) *bags,
    char *pass
);
int dump_certs_pkeys_bag
(
    BIO *out,
    PKCS12_SAFEBAG *bags,
    char *pass
);
int print_attribs
(
    BIO *out,
    STACK_OF(X509_ATTRIBUTE) *attrlst,
    char *name
);
int dump_cert_text
(
    BIO *out,
    X509 *x
);
void hex_prin
(
    BIO *out,
    unsigned char *buf,
    int len
);


/****************************************************************************
*
* Description:
*
*  dump_certs_keys_buf - Converts a pkcs12 buffer into a clear text buffer.
****************************************************************************/
LDAP_F(int) dump_certs_keys_buf(BIO *out, BIO *in, char *pass)
{
    PKCS12     *p12     = NULL;
    int         ret     = 0; /* error status */
    
    OpenSSL_add_all_algorithms();
    if (!(p12 = d2i_PKCS12_bio (in, NULL)))
    {
        goto end;
    }
    if (!dump_certs_keys_p12 (out, p12, pass))
    {
	     goto end;
    }
    ret = 1; /* success */
end:   
    if (p12) PKCS12_free(p12);
   
    return ret;
}


/****************************************************************************
*
* Description:
*
*  dump_certs_keys_p12 - Converts a pkcs12 structure into a clear text.
****************************************************************************/
int dump_certs_keys_p12 (BIO *out, PKCS12 *p12, char *pass)
{
	STACK_OF(PKCS7) *asafes;
	STACK_OF(PKCS12_SAFEBAG) *bags;
	int i = 0, bagnid = 0;
	PKCS7 *p7;

	if (!( asafes = M_PKCS12_unpack_authsafes (p12))) return 0;
	for (i = 0; i < sk_PKCS7_num (asafes); i++) {
		p7 = sk_PKCS7_value (asafes, i);
		bagnid = OBJ_obj2nid (p7->type);
		if (bagnid == NID_pkcs7_data) {
			bags = M_PKCS12_unpack_p7data (p7);
		} else if (bagnid == NID_pkcs7_encrypted) {
			bags = M_PKCS12_unpack_p7encdata (p7, pass, -1);
		} else continue;
		if (!bags) return 0;
	    	if (!dump_certs_pkeys_bags (out, bags, pass)) {
			sk_PKCS12_SAFEBAG_pop_free (bags, PKCS12_SAFEBAG_free);
			return 0;
		}
		sk_PKCS12_SAFEBAG_pop_free (bags, PKCS12_SAFEBAG_free);
	}
	sk_PKCS7_pop_free (asafes, PKCS7_free);
	return 1;
}


/****************************************************************************
*
* Description:
*
*  dump_certs_keys_bags - Converts an array of pkcs12 safe bags into a clear
*  text.
****************************************************************************/
int dump_certs_pkeys_bags (BIO *out, STACK_OF(PKCS12_SAFEBAG) *bags,
			   char *pass)
{
	int i = 0;
	for (i = 0; i < sk_PKCS12_SAFEBAG_num (bags); i++) {
		if (!dump_certs_pkeys_bag (out,sk_PKCS12_SAFEBAG_value (bags, i),pass))
		    return 0;
	}
	return 1;
}

/****************************************************************************
*
* Description:
*
*  dump_certs_keys_bag - Converts a pkcs12 safe bag into a clear text.
****************************************************************************/
int dump_certs_pkeys_bag (BIO *out, PKCS12_SAFEBAG *bag, char *pass)
{
	EVP_PKEY *pkey;
	PKCS8_PRIV_KEY_INFO *p8;
	X509 *x509;
	
	switch (M_PKCS12_bag_type(bag))
	{
	case NID_keyBag:
		print_attribs (out, bag->attrib, "Bag Attributes");
		p8 = bag->value.keybag;
		if (!(pkey = EVP_PKCS82PKEY (p8))) return 0;
		print_attribs (out, p8->attributes, "Key Attributes");
		PEM_write_bio_PrivateKey (out, pkey, NULL, NULL, 0, NULL, NULL);
		EVP_PKEY_free(pkey);
	break;

	case NID_pkcs8ShroudedKeyBag:
		print_attribs (out, bag->attrib, "Bag Attributes");
		if (!(p8 = M_PKCS12_decrypt_skey (bag, pass, -1)))
				return 0;
		if (!(pkey = EVP_PKCS82PKEY (p8))) return 0;
		print_attribs (out, p8->attributes, "Key Attributes");
		PKCS8_PRIV_KEY_INFO_free(p8);
		PEM_write_bio_PrivateKey (out, pkey, NULL, NULL, 0, NULL, NULL);
		EVP_PKEY_free(pkey);
	break;

	case NID_certBag:
		print_attribs (out, bag->attrib, "Bag Attributes");
		if (M_PKCS12_cert_bag_type(bag) != NID_x509Certificate )
								 return 1;
		if (!(x509 = M_PKCS12_certbag2x509(bag))) return 0;
		dump_cert_text (out, x509);
		PEM_write_bio_X509 (out, x509);
		X509_free(x509);
	break;

	case NID_safeContentsBag:
		print_attribs (out, bag->attrib, "Bag Attributes");
		return dump_certs_pkeys_bags (out, bag->value.safes, pass);
					
	default:
		return 1;
	break;
	}
	return 1;
}


/****************************************************************************
*
* Description:
*
*  print_attribs - Generalised attribute print. Handle PKCS#8 and bag
*  attributes
****************************************************************************/
int print_attribs (BIO *out, STACK_OF(X509_ATTRIBUTE) *attrlst, char *name)
{
	X509_ATTRIBUTE *attr;
	ASN1_TYPE *av;
	char *value = NULL;
	int i = 0, attr_nid = 0;
	if(!attrlst) {
		BIO_printf(out, "%s: <No Attributes>\n", name);
		return 1;
	}
	if(!sk_X509_ATTRIBUTE_num(attrlst)) {
		BIO_printf(out, "%s: <Empty Attributes>\n", name);
		return 1;
	}
	BIO_printf(out, "%s\n", name);
	for(i = 0; i < sk_X509_ATTRIBUTE_num(attrlst); i++) {
		attr = sk_X509_ATTRIBUTE_value(attrlst, i);
		attr_nid = OBJ_obj2nid(attr->object);
		BIO_printf(out, "    ");
		if(attr_nid == NID_undef) {
			i2a_ASN1_OBJECT (out, attr->object);
			BIO_printf(out, ": ");
		} else BIO_printf(out, "%s: ", OBJ_nid2ln(attr_nid));

		if(sk_ASN1_TYPE_num(attr->value.set)) {
			av = sk_ASN1_TYPE_value(attr->value.set, 0);
			switch(av->type) {
				case V_ASN1_BMPSTRING:
        			value = uni2asc(av->value.bmpstring->data,
                                	       av->value.bmpstring->length);
				BIO_printf(out, "%s\n", value);
				OPENSSL_free(value);
				break;

				case V_ASN1_OCTET_STRING:
				hex_prin(out, av->value.octet_string->data,
					av->value.octet_string->length);
				BIO_printf(out, "\n");	
				break;

				case V_ASN1_BIT_STRING:
				hex_prin(out, av->value.bit_string->data,
					av->value.bit_string->length);
				BIO_printf(out, "\n");	
				break;

				default:
					BIO_printf(out, "<Unsupported tag %d>\n", av->type);
				break;
			}
		} else BIO_printf(out, "<No Values>\n");
	}
	return 1;
}

/****************************************************************************
*
* Description:
*
*  dump_cert_text - Converts an X509 structure to a buffer.
****************************************************************************/
int dump_cert_text (BIO *out, X509 *x)
{
	char buf[256];
	X509_NAME_oneline(X509_get_subject_name(x),buf,256);
	BIO_puts(out,"subject=");
	BIO_puts(out,buf);

	X509_NAME_oneline(X509_get_issuer_name(x),buf,256);
	BIO_puts(out,"\nissuer= ");
	BIO_puts(out,buf);
	BIO_puts(out,"\n");
        return 0;
}

/****************************************************************************
*
* Description:
*
*  hex_prin - Converts hex values to hex string.
****************************************************************************/
void hex_prin(BIO *out, unsigned char *buf, int len)
{
	int i = 0;
	for (i = 0; i < len; i++) BIO_printf (out, "%02X ", buf[i]);
}
